/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or */
/* (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */



#include "../headers/stack.h"
#include "../headers/RNA_structure.h"

#ifndef RNA_MULTIPLE_STRUCTURE_H_
#define RNA_MULTIPLE_STRUCTURE_H_

/*******************************************/
/* RNA multiple sequence                   */
/*******************************************/

/**
 * RNA_sequence_m : list of RNA_sequence
 * - eq : eq factors
 * - super_pairing : type of the pairing for the super structure
 * - position of the partner in the super structure
 **/
typedef struct{
  int nb_seq; 
  char *eq; 
  int *super_pairing; 
  int *super_partner; 
  RNA_sequence *s;
}RNA_sequence_m;


int sequence_density(RNA_sequence_m ** list_rna, int nb_seq);

/*******************************************/
/* multiple post-ordre tree                */
/*******************************************/

typedef struct{
  int size;
  int *paired;
  char *labelX, *labelY;
  int positionX, positionY;
}node_m;

typedef struct{
  int size;
  int nb_seq;
  node_m *n;
}postorder_tree_m;

/* display a RNA_sequence_m  */ 
void display_RNA_sequence_m(RNA_sequence_m *list);

/* free a postorder_tree_m */
void free_postorder_tree_m(postorder_tree_m *tree);

/* RNA_sequence_m allocation function */
RNA_sequence_m *alloc_RNA_sequence_m(int size, int nb_seq);

/* copy the RNA_sequence_m temp in the RNA_sequence_m test */
void copy_sequence(RNA_sequence_m *test, RNA_sequence_m *temp);


/*
 * free a RNA_sequence_m
 * the pointer rna is not freed
 */
void free_RNA_sequence_m(RNA_sequence_m *rna);

/*
 * read a multi-fasta file
 *
 * inputFile:        the file to read
 * nb_seq:           fill with the number of sequences 
 *
 * RETURN: a list of RNA_sequence_m
 */
RNA_sequence_m **get_multiple_bracket(char * inputFile, int *nb_seq, int *EQ);

/* Transform a RNA_sequence_m into a postorder_tree_m */
postorder_tree_m *RNAsequence2postorder_m(RNA_sequence_m *rna);

void clustal_w_output(RNA_sequence_m *align, char * outputFile);

void display_alignment_m(RNA_sequence_m *align, char * outputFile);

void multi_fasta(RNA_sequence_m *align, char * outputFile);

int alignment_density(int total_length, int nb_seq, RNA_sequence_m * align);

void create_alignment_RNA_sequences_m(RNA_sequence_m *rnaA, RNA_sequence_m *rnaB, int **supertree, 
				      RNA_sequence_m *alignment, int startA, int endA, int startB,int endB);

int covariations(RNA_sequence_m *list); 

#endif /*RNA_MULTIPLE_STRUCTURE_H_*/
