/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or */
/* (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/********************************/
/* RNA_multiple_structure.c     */
/********************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "../headers/edit_distance.h"
#include "../headers/stack.h"
#include "../headers/score.h"
#include "../headers/RNA_structure.h"
#include "../headers/RNA_multiple_structure.h"


int covariations(RNA_sequence_m *list){
 
  int i, j; 
  int nb_windows=0; 
  int length=0;
  int id_column;
  int gap_column; 

  for(j=1; j<list->s[0].size; j++){
    /* on teste une position */
    gap_column=(list->s[0].p[j].nucleotide==4);
    id_column=1;
    for(i=1; i<list->nb_seq; i++){
      id_column = id_column && (list->s[i].p[j].nucleotide == list->s[0].p[j].nucleotide);
      gap_column=gap_column||  (list->s[i].p[j].nucleotide == 4);
    }
    if ((!id_column)&&(!gap_column)&&(length==0)) { length=6; nb_windows++;} 
    length--;
    if (length<0) length=0;
  }
  return nb_windows*1000/(list->s[0].size-1); 
}

/* display a RNA_sequence_m */
void display_RNA_sequence_m(RNA_sequence_m *list){

  int i, j;

  for(i=0; i<list->nb_seq; i++){

    printf("%s\n", list->s[i].name);
   
    for(j=1; j<list->s[i].size; j++){
      printf("%c", int_to_RNAchar(list->s[i].p[j].nucleotide));
    }

    printf("\n");

    for(j=1; j<list->s[i].size; j++){
      printf("%c", int_2_char(list->s[i].p[j].pairing) );
    }
    printf("\n");
  }

  printf("pairing:\n");
  for(j=1; j<list->s[0].size; j++){
    printf("%c", int_2_char(list->super_pairing[j]) );
  }
  printf("\npartner:\n");
  for(j=1; j<list->s[0].size; j++){
    printf("%i ", list->super_partner[j] );
  }
  printf("\n");

  for(i=1; i< list->s[0].size-1; i++)
    printf("%c", list->eq[i]);
  printf("\n\n");
}


/********************************************/
/***  MEMORY ALLOCATION FUNCTIONALITIES   ***/
/********************************************/

RNA_sequence_m *alloc_RNA_sequence_m(int size, int nb_seq){

  RNA_sequence_m *seq;
  
  seq = (RNA_sequence_m *)malloc( sizeof(RNA_sequence_m) );
  seq->nb_seq = nb_seq;
  seq->s = (RNA_sequence *)malloc( nb_seq * sizeof(RNA_sequence) );
  seq->super_pairing = (int *)malloc( sizeof(int) * size );
  seq->super_partner = (int *)malloc( sizeof(int) * size );
  seq->eq = (char *)malloc( sizeof(char) * (size-1) );

  return seq;
}

void free_RNA_sequence_m(RNA_sequence_m *rna){
  int i;

  free(rna->eq);

  for(i=0; i<rna->nb_seq; i++){
    free(rna->s[i].p);
    free(rna->s[i].name);
  }
  free(rna->s);
  free(rna->super_pairing);
  free(rna->super_partner);
}

void free_postorder_tree_m(postorder_tree_m *tree){

  int i;

  for(i=0; i<tree->size; i++){
    free(tree->n[i].labelX);
    free(tree->n[i].labelY);
    free(tree->n[i].paired);
  }

  free(tree->n);
  /* free(tree); */
}

/* copy the RNA_sequence_m temp in the RNA_sequence_m test */
void copy_sequence(RNA_sequence_m *test, RNA_sequence_m *temp){
  int k, l;

  test->s = (RNA_sequence *)malloc( sizeof(RNA_sequence) * temp->nb_seq );
  test->nb_seq = temp->nb_seq;
  test->eq = (char *)malloc( sizeof(char) * (temp->s[0].size-1) );
  test->super_pairing = (int *)malloc( sizeof(int) * (temp->s[0].size) );
  test->super_partner = (int *)malloc( sizeof(int) * (temp->s[0].size) );
  for(k=0; k < temp->s[0].size-1; k++)
    test->eq[k] = temp->eq[k];
  for(k=0; k < temp->s[0].size; k++){
    test->super_pairing[k] = temp->super_pairing[k];
    test->super_partner[k] = temp->super_partner[k];
  }
  for(k=0; k<temp->nb_seq; k++){
    test->s[k].name = (char *)malloc( sizeof(char) * (strlen(temp->s[k].name)+1) );
    strcpy(test->s[k].name, temp->s[k].name);
    test->s[k].p = (position *)malloc( sizeof(position) * temp->s[k].size);
    test->s[k].size = temp->s[k].size;
    for(l=0; l<temp->s[k].size; l++){
      test->s[k].p[l].nucleotide = temp->s[k].p[l].nucleotide;
      test->s[k].p[l].pairing = temp->s[k].p[l].pairing;
      test->s[k].p[l].partner = temp->s[k].p[l].partner;
    }
  }
}

RNA_sequence_m **get_multiple_bracket(char *inputFile, int *nb_seq, int *EQ ){

  RNA_sequence_m **RNA_list = NULL;
  FILE *fp;
  int num_seq, size, i, j;
  int pos, eq_factors=0;
  int buffer_size = 0;  
  int *q;
  char *buffer = NULL;
  char c, temp[100];
  char *p1, *p2;


  /* file opening */
  fp = fopen(inputFile, "r");
  if( !fp ){
    printf("error when loading file: %s\n", inputFile);
    exit(2);
  }

  /* copy the file into buffer */
  i = 0;
  *nb_seq = 0;
  while( !feof(fp) ){
    fscanf(fp, "%c", &c);
    i++;
    buffer = realloc(buffer, i*sizeof(char));
    buffer[i-1] = c;
    if( c == '>' )
      (*nb_seq)++;
  }
  buffer = realloc(buffer, (i+1)*sizeof(char));
  buffer_size = i+1;
  buffer[buffer_size-1] = '\0';
  fclose(fp);
  
  RNA_list = (RNA_sequence_m **)malloc( sizeof(RNA_sequence_m *) * *nb_seq );
  num_seq = 0;
  i = 0; 
  while( num_seq < *nb_seq ){

    if( buffer[i] != '>' ){
      printf("[EE] the file %s has a wrong format !\n", inputFile);
      exit(3);
    }
    /* get the sequences names */
    i++;
    while( buffer[i] == ' ' ) i++;
    j = i;
    while( buffer[i] != '\n' ){
      temp[i-j] = buffer[i];
      i++;
    }
    temp[i-j] = '\0';

    /* compute sequence length */
    pos = i;
    size = 0;
    while( is_RNA_char(buffer[i]) == 1 || buffer[i] == '\n'){
      if( buffer[i] != '\n')
	size++;
      i++;
    }
    p1 = (char *)malloc( sizeof(char) * (size) );
    p2 = (char *)malloc( sizeof(char) * (size) );
    q  = (int *)malloc( sizeof(int) * (size) );
    RNA_list[num_seq] = alloc_RNA_sequence_m(size+1, 1);

    /* get the sequence and the structure for the current RNA */
    i = pos;
    j = 0;
    while( j<size ){
      if( buffer[i] != '\n' ){
	p1[j] = buffer[i];
	j++;
      }
      i++;
    }
    j = 0;
    while( j<size ){
      if( buffer[i] != '\n' ){
	p2[j] = buffer[i];
	j++;
      }
      i++;
    }

    i++;
    /* eq constraint */
    if( *EQ ){
      /* if no eq factors in the file, ignore eq constraints */
      if( !( buffer[i] != '>' && buffer[i] >= 32 && buffer[i] <= 125 ) || i>buffer_size-size ){
	eq_factors = 0;
	for(j=0; j<size; j++)
	  RNA_list[num_seq]->eq[j] = ' ';
      }
      /* else get eq factors */
      else{
	eq_factors = 1;
	j=0;
	while( j < size ){
	  if( buffer[i] != '\n' ){
	    RNA_list[num_seq]->eq[j] = buffer[i];
	    j++;
	  }
	  i++;
	}
      }
    }
    else{
      for(j=0; j<size; j++)
	RNA_list[num_seq]->eq[j] = ' ';
    }
    while( i<buffer_size && !(buffer[i] == '>') ) i++;
    retrace_partner(0, p2, q, size, inputFile);
    
    RNA_list[num_seq]->s[0].p = malloc( sizeof(position)*(size+1) );
    RNA_list[num_seq]->s[0].size = size+1;
    RNA_list[num_seq]->s[0].name = (char *)malloc( sizeof(char) * (strlen(temp)+1) );
    strcpy(RNA_list[num_seq]->s[0].name, temp);
    for(j=0;j<size;j++){
      RNA_list[num_seq]->s[0].p[j+1].nucleotide = RNAcharacter( p1[j], inputFile );
      RNA_list[num_seq]->s[0].p[j+1].pairing = char_2_int( p2[j] );
      RNA_list[num_seq]->s[0].p[j+1].partner = q[j]+1;
      RNA_list[num_seq]->super_pairing[j+1] = char_2_int( p2[j] );
      RNA_list[num_seq]->super_partner[j+1] = q[j]+1;
    }
    
    num_seq++;
    free(p1);
    free(p2);
    free(q);
  }
 
  if( *EQ && eq_factors == 0){
      printf("[!!] no eq factors found in %s\n\t -> eq constraints ignored\n\n", inputFile);
      *EQ = 0;
  }
  free(buffer);

  return RNA_list;
}


int sequence_density(RNA_sequence_m ** list_rna, int nb_seq){

  int total_length=0;
  int total_pairs=0;
  int i,j;
  for (i=0; i<nb_seq; i++){
    total_length+=list_rna[i]->s[0].size - 1;
    for (j=1; j<list_rna[i]->s[0].size; j++)
      if (list_rna[i]->s[0].p[j].pairing==1) total_pairs++;
  }
  return (total_pairs*200/total_length);

}

void RNAsequence_postorder_m_aux(postorder_tree_m *tree, RNA_sequence_m *rna, int x, 
				 int y, int *num, int *sizeFather){
  int size = 1;
    
  if( x>y )return ;
  
  /*(x,y) is a base-pair */
  if( rna->super_pairing[x] != 0 && rna->super_partner[x] == y ){
    RNAsequence_postorder_m_aux( tree, rna, x+1, y-1, num, &size);
    *sizeFather = *sizeFather + size;
    *num = *num + 1;
    tree->n[*num].size = size;
    tree->n[*num].positionX = x;
    tree->n[*num].positionY = y;
  }else{
    if( rna->super_partner[x] == x){/* x is a free base */
      *num = *num + 1;
      size = 1;
      *sizeFather = *sizeFather + size;
      tree->n[*num].size = size;
      tree->n[*num].positionX = x;
      tree->n[*num].positionY = x;
      RNAsequence_postorder_m_aux( tree, rna, x+1, y, num, sizeFather );			
    }else{  /* x is paired with another base than y */
      RNAsequence_postorder_m_aux( tree, rna, x, rna->super_partner[x], num, sizeFather );
      RNAsequence_postorder_m_aux( tree, rna, rna->super_partner[x]+1, y, num, sizeFather);
    }
  }
}

postorder_tree_m *RNAsequence2postorder_m(RNA_sequence_m *rna){

  int num=0;
  int size=0;
  int i, j;
  postorder_tree_m *tree;
  postorder_tree_m *tmp_tree;

  if(rna->nb_seq < 1)
    return NULL;

  /* allocation of temporary tree */
  tmp_tree = malloc( sizeof(postorder_tree_m) );
  tmp_tree->n = malloc( sizeof(node_m) * (rna->s[0].size) );
  for(i=0; i<rna->s[0].size; i++)
    tmp_tree->n[i].size = 0;

  /* construction of the tree with the first sequence of the list */
  RNAsequence_postorder_m_aux(tmp_tree, rna, 1, rna->s[0].size - 1, &num, &size);

  /* allocation of the final tree */
  tree = (postorder_tree_m *)malloc( sizeof(postorder_tree_m) );
  tree->n=malloc( sizeof(node_m) * (num+1) );
  for(i=0; i<num+1; i++){
    tree->n[i].labelX = (char *)malloc( sizeof(char) * rna->nb_seq );
    tree->n[i].labelY = (char *)malloc( sizeof(char) * rna->nb_seq );
    tree->n[i].paired = (int *)malloc( sizeof(int) * rna->nb_seq );
  }

  /* copy of the temporary tree into the final one */
  tree->nb_seq = rna->nb_seq;
  tree->size = num+1;
  tree->n[0].positionX=0; /*new*/
  tree->n[0].positionY=0; /*new*/
  tree->n[0].size=0; /*new*/
  for(i=1; i<num+1; i++){ 
    tree->n[i].size = tmp_tree->n[i].size;
    tree->n[i].positionX = tmp_tree->n[i].positionX;
    tree->n[i].positionY = tmp_tree->n[i].positionY;
  }

  free(tmp_tree->n);
  free(tmp_tree);
  
  for(i=1; i<tree->size; i++){
    for(j=0; j<rna->nb_seq; j++){
      tree->n[i].labelX[j] = rna->s[j].p[tree->n[i].positionX].nucleotide;

      if(tree->n[i].size > 1){
	tree->n[i].labelY[j] = rna->s[j].p[tree->n[i].positionY].nucleotide;
	if(rna->s[j].p[tree->n[i].positionX].pairing == -1 && 
	   rna->s[j].p[tree->n[i].positionX].partner == tree->n[i].positionY )
	  tree->n[i].paired[j] = 1;
	else
	  tree->n[i].paired[j] = 0;
      }
      else{
	tree->n[i].labelY[j] = 0;
	tree->n[i].paired[j] = 0;
      }
    }
  }

  return tree;
}



void clustal_w_output(RNA_sequence_m *align, char *outputFile){

  int i, j, k, l, m;
  int alignment_length = align->s[0].size;
  FILE *fp;
  char c, c1;

  if ( strlen(outputFile) > 0 )
    fp = fopen(outputFile, "w");
  else
    fp = stdout;

  fprintf(fp,"CLUSTAL W multiple sequence alignment\n\n");
  
  /* displaying nucleotides */
  for(j=0; j < (alignment_length/maxChar) + 1; j++){

    for(k=0; k<align->nb_seq; k++){
      /* names */
      l = 0;
      m = 0;
      while( m<10 && align->s[k].name[l] != '\n' && align->s[k].name[l] != '\r' && align->s[k].name[l] != '\0' ){
	if( align->s[k].name[l] != ' '){
	  fprintf(fp,"%c", align->s[k].name[l]);
	  m++;
	}
	l++;
      }
      for(m=l; m<10; m++)
	fprintf(fp," ");
      fprintf(fp,"     ");

      /* nucleotides */
      for(i=(j*maxChar)+1; i<=((j*maxChar)+maxChar) && i<alignment_length;i++)
	fprintf(fp, "%c", int_to_RNAchar((int)align->s[k].p[i].nucleotide) );

      if( j == alignment_length/maxChar )
	fprintf(fp, " %i\n", alignment_length -1);
      else
	fprintf(fp, " %i\n", (j+1)*60);
    }

    /* stars */ 
    for(m=0; m<15; m++)
	fprintf(fp," ");
    for(i=(j*maxChar)+1; i<=((j*maxChar)+maxChar) && i<alignment_length;i++){
      c = align->s[0].p[i].nucleotide;
      c1 = c;
      for(k=1; k<align->nb_seq; k++)
	if( align->s[k].p[i].nucleotide != c )
	  c1 = align->s[k].p[i].nucleotide;
      if( c == c1)
	fprintf(fp, "*");
      else
	fprintf(fp, " ");
    }
    fprintf(fp,"\n\n");      
  }

  fprintf(fp,"\n");
}

int alignment_density(int max_length, int nb_seq, RNA_sequence_m * align){
postorder_tree_m *tree = NULL;
 int i,j; 
 int size_align;
 tree = RNAsequence2postorder_m(align);
 size_align=(tree->size-1)*(tree->nb_seq);
 for (i=1;i<tree->size; i++){
   if (tree->n[i].size>1)
     for (j=0; j<tree->nb_seq; j++){
       if (tree->n[i].paired[j]!=1)
	 size_align++;
     }
 }
 free_postorder_tree_m(tree);
 return (size_align*100/max_length);
}


void display_alignment_m(RNA_sequence_m *align, char * outputFile){
  int i, j, k, l, m;
  int alignment_length = align->s[0].size;
  FILE *fp;
  char c, c1;

  if ( strlen(outputFile) > 0 )
    fp = fopen(outputFile, "w");
  else
    fp = stdout;

  fprintf(fp,"\n\n");
  
  /* displaying nucleotides */
  for(j=0; j < (alignment_length/maxChar) + 1; j++){

    for(k=0; k<align->nb_seq; k++){
      /* names */
      for(l=0; l<10 && align->s[k].name[l] != '\n' && align->s[k].name[l] != '\r' && align->s[k].name[l] != '\0'; l++)
	fprintf(fp,"%c", align->s[k].name[l]);
      for(m=l; m<10; m++)
	fprintf(fp," ");
      fprintf(fp,"     ");

      /* nucleotides */
      for(i=(j*maxChar)+1; i<=((j*maxChar)+maxChar) && i<alignment_length;i++)
	fprintf(fp, "%c", int_to_RNAchar((int)align->s[k].p[i].nucleotide) );
      fprintf(fp, "\n");
    }

    /* stars */ 
    for(m=0; m<15; m++)
	fprintf(fp," ");
    for(i=(j*maxChar)+1; i<=((j*maxChar)+maxChar) && i<alignment_length;i++){
      c = align->s[0].p[i].nucleotide;
      c1 = c;
      for(k=1; k<align->nb_seq; k++)
	if( align->s[k].p[i].nucleotide != c )
	  c1 = align->s[k].p[i].nucleotide;
      if( c == c1)
	fprintf(fp, "*");
      else
	fprintf(fp, " ");
    }
    fprintf(fp,"\n\n");      
  }

  fprintf(fp,"\n");
 
  /* displaying brackets */
  for(j=0; j < (alignment_length/maxChar) + 1; j++){
    
    for(k=0; k<align->nb_seq; k++){
      /* names */
      for(l=0; l<10 && align->s[k].name[l] != '\n' && align->s[k].name[l] != '\r' && align->s[k].name[l] != '\0'; l++)
	fprintf(fp,"%c", align->s[k].name[l]);
      for(m=l; m<10; m++)
	fprintf(fp," ");
      fprintf(fp,"     ");
      
      /* brackets */
      for(i=(j*maxChar)+1; i<=((j*maxChar)+maxChar) && i<alignment_length;i++)
	fprintf(fp, "%c",int_2_char(align->s[k].p[i].pairing));
      fprintf(fp,"\n");
    }

    /* stars */ 
    for(m=0; m<15; m++)
	fprintf(fp," ");
    for(i=(j*maxChar)+1; i<=((j*maxChar)+maxChar) && i<alignment_length;i++){
      c = int_2_char(align->s[0].p[i].pairing);
      c1 = c;
      for(k=1; k<align->nb_seq; k++)
	if( int_2_char(align->s[k].p[i].pairing) != c )
	  c1 = int_2_char(align->s[k].p[i].pairing);
      if( c == c1)
	fprintf(fp, "*");
      else
	fprintf(fp, " ");
    }
    fprintf(fp,"\n\n");      
  }

  if( strlen(outputFile) > 0 )
    fclose(fp);
}


/*
 * Print the alignment in fasta format
 */
void multi_fasta(RNA_sequence_m *align, char * outputFile){

  int i, j;
  FILE *fp;

  if ( strlen(outputFile) > 0 )
    fp = fopen(outputFile, "w");
  else
    fp = stdout;

  for(i=0; i<align->nb_seq; i++){
    /* names */
    fprintf(fp, ">");
    for(j=0;j<strlen(align->s[i].name); j++)
      if(align->s[i].name[j] !='\n')
	fprintf(fp, "%c", align->s[i].name[j]);
    fprintf(fp, "\n");

    /* nucleotides */
    for(j=1; j<align->s[i].size; j++)
      fprintf(fp, "%c", int_to_RNAchar((int)align->s[i].p[j].nucleotide) );
    fprintf(fp, "\n");

    /* brackets */
    for(j=1; j<align->s[i].size; j++)
      fprintf(fp, "%c", int_2_char(align->s[i].p[j].pairing) );
    fprintf(fp, "\n");
  }
  
  if( strlen(outputFile) > 0)
    fclose(fp);
}


void fill_alignment(RNA_sequence_m *alignment, RNA_sequence_m *rnaA, RNA_sequence_m *rnaB, int pos, int x, int y, int type){

  int j;

  /* match */
  if(type == 0){
    for(j=0; j<rnaA->nb_seq; j++){
	alignment->s[j].p[pos].nucleotide = rnaA->s[j].p[x].nucleotide;
	alignment->s[j].p[pos].pairing = rnaA->s[j].p[x].pairing;
	alignment->s[j].p[pos].partner = rnaA->s[j].p[x].partner;
    }
    for(j=0; j<rnaB->nb_seq; j++){
      alignment->s[j + rnaA->nb_seq].p[pos].nucleotide = rnaB->s[j].p[y].nucleotide;
      alignment->s[j + rnaA->nb_seq].p[pos].pairing = rnaB->s[j].p[y].pairing;
      alignment->s[j + rnaA->nb_seq].p[pos].partner = rnaB->s[j].p[y].partner;
    }
  }
  /* deletion in A */
  else if( type == -1){
    for(j=0; j<rnaA->nb_seq; j++){
      alignment->s[j].p[pos].nucleotide = 4;
      alignment->s[j].p[pos].pairing = 2;
      alignment->s[j].p[pos].partner = -1;
    }
    
    for(j=0; j<rnaB->nb_seq; j++){
      alignment->s[j + rnaA->nb_seq].p[pos].nucleotide = rnaB->s[j].p[y].nucleotide;
      alignment->s[j + rnaA->nb_seq].p[pos].pairing = rnaB->s[j].p[y].pairing;
      alignment->s[j + rnaA->nb_seq].p[pos].partner = rnaB->s[j].p[y].partner;
    }
  }
  /* deletion in B */
  else if( type == 1){ 
    for(j=0; j<rnaA->nb_seq; j++){
      alignment->s[j].p[pos].nucleotide = rnaA->s[j].p[x].nucleotide;
      alignment->s[j].p[pos].pairing = rnaA->s[j].p[x].pairing;
      alignment->s[j].p[pos].partner = rnaA->s[j].p[x].partner;
    }
    
    for(j=0; j<rnaB->nb_seq; j++){
      alignment->s[j + rnaA->nb_seq].p[pos].nucleotide = 4;
      alignment->s[j + rnaA->nb_seq].p[pos].pairing = 2;
      alignment->s[j + rnaA->nb_seq].p[pos].partner = -1;
    }
  }
}


/* input: rnaA, rnaB, supertree
 * output: alignment
 */
void create_alignment_RNA_sequences_m(RNA_sequence_m *rnaA, RNA_sequence_m *rnaB, int **supertree, 
				      RNA_sequence_m *alignment, int startA, int endA, int startB, int endB){

  int i, j, x, y, alignment_length = 0;
  int *q;
  char *p;

  alignment_length = endA + endB - startA - startB + 2; 
  for (i=startA; i<=endA; i++)
    for (j=startB; j<=endB; j++)
      if(supertree[i][j]==1) alignment_length--;

  /* memory allocation for the alignment */
  alignment->s = (RNA_sequence *)malloc( sizeof(RNA_sequence) * (rnaA->nb_seq + rnaB->nb_seq) ); 
  alignment->nb_seq = rnaA->nb_seq + rnaB->nb_seq;
  alignment->eq = (char *)malloc( sizeof(char) * alignment_length );
  alignment->super_pairing = (int *)malloc( sizeof(int) * (alignment_length+1) );
  alignment->super_partner = (int *)malloc( sizeof(int) * (alignment_length+1) );
  for(i=0; i<rnaA->nb_seq + rnaB->nb_seq; i++){
    alignment->s[i].p = (position *)malloc( sizeof(position) * (alignment_length+1) );
    alignment->s[i].size = alignment_length + 1;
  }

  x=endA; 
  y=endB;
  for(i=alignment_length; i>0;i--){
    if(supertree[x][y]==1){ 
      /* base-match or mismatch */
      alignment->eq[i-1] = rnaA->eq[x-1] == 0 ? rnaB->eq[y-1] : rnaA->eq[x-1];
      alignment->super_pairing[i] = rnaA->super_pairing[x] != 0 ? rnaA->super_pairing[x] : rnaB->super_pairing[y];
      fill_alignment(alignment, rnaA, rnaB, i, x, y, 0);
      x--;y--;
    }else if (supertree[x][y]==-1 ){ /* base deletion */
      alignment->eq[i-1] = rnaA->eq[x-1];
      alignment->super_pairing[i] = rnaA->super_pairing[x];
      fill_alignment(alignment, rnaA, rnaB, i, x, y, 1);
      x--;
    }
    else if (supertree[x][y]==-2 ){ /* base insertion */
	alignment->eq[i-1] = rnaB->eq[y-1];
	alignment->super_pairing[i] = rnaB->super_pairing[y];	
	fill_alignment(alignment, rnaA, rnaB, i, x, y, -1);
	y--;
    }
    else{
      printf("\n Not a valid alignment !!! ");
      printf("x,y -> supertree[x][y]: %d, %d -> %d\n", x, y,supertree[x][y]);
      exit(1);
    }
  }
  /* copy sequences names */
  for(j=0; j<rnaA->nb_seq; j++){
    alignment->s[j].name = (char *)malloc( sizeof(char) * (strlen(rnaA->s[j].name) + 1) );
    strcpy(alignment->s[j].name, rnaA->s[j].name);
  }
  for(j=0; j<rnaB->nb_seq; j++){
    alignment->s[j + rnaA->nb_seq].name = (char *)malloc( sizeof(char) * (strlen(rnaB->s[j].name) + 1) );
    strcpy(alignment->s[j + rnaA->nb_seq].name, rnaB->s[j].name);
  }

  /*   partner renumerating */
  for(i=0; i<alignment->nb_seq; i++){
    q = (int *)malloc( sizeof(int) * alignment->s[i].size );
    p = (char *)malloc( sizeof(char) * alignment->s[i].size );
    for(j=1; j<alignment->s[i].size; j++)
      p[j] = int_2_char(alignment->s[i].p[j].pairing);

    retrace_partner(1, p, q, alignment->s[i].size, NULL);

    for(j=1; j<alignment->s[i].size; j++)
      alignment->s[i].p[j].partner = q[j];
 
    for(j=1; j<alignment->s[0].size; j++)
      alignment->super_partner[j] = q[j];
   
    free(q);
    free(p);
  }

  /*q = (int *)malloc( sizeof(int) * alignment->s[0].size );
  p = (char *)malloc( sizeof(char) * alignment->s[0].size );
  for(j=1; j<alignment->s[0].size; j++)
    p[j] = int_2_char(alignment->super_pairing[j]);

  printf("ret2\n");
  retrace_partner(1, p, q, alignment->s[0].size, NULL);
  */
  

 

  /* free memory */
  /*free(q);
  free(p);  
  */
}
