/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or */
/* (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */


/**************************/
/*  global alignment      */
/**************************/ 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "../headers/score.h"
#include "../headers/constraint.h"
#include "../headers/stack.h"
#include "../headers/RNA_structure.h"
#include "../headers/arc_alignment.h"
#include "../headers/graph.h"
#include "../headers/matrix.h"
#include "../headers/arc_alignment_graph.h"

 
int optimal_arc_score(RNA_sequence *rnaA, RNA_sequence *rnaB, int primary, int similar, int eq, char *eqA, char *eqB){

  postorder_tree * treeA;
  postorder_tree * treeB;
  int score; 
  
  treeA=(postorder_tree * ) malloc(sizeof(postorder_tree));
  treeB=(postorder_tree * ) malloc(sizeof(postorder_tree));
  
  RNA_sequence2postorder(rnaA,treeA);
  RNA_sequence2postorder(rnaB,treeB);

 /*   prise en compte des contraintes */
  constraint=new_matrix_char(treeA,treeB);	
  no_constraint(constraint,treeA->size,treeB->size);

  if( eq ){
    full_constraint(constraint, treeA->size, treeB->size);
    eq_constraint(constraint, (postorder_tree_m *)treeA, (postorder_tree_m *)treeB, eqA, eqB, rnaA->size, rnaB->size, 0);
  }
    
  if (similar==1)
    diagonal_constraint(constraint,treeA->size,treeB->size, rnaA->size/4);  

  if(primary==1)
    primary_sequence_constraint(constraint,treeA,treeB,rnaA,rnaB);

       
  build_alignment_graph(treeA, treeB);
  /*display_alignment_graph(alignment_graph, treeA->size-1, treeB->size-1);*/
 
  computeTables(treeA, treeB,0,0,treeA->size-1, treeB->size-1);
  score=get_tables(0,treeA->size-1,treeB->size-1,ABCD); 
  return score;
}


/* Construction of the optimal mapping: 
   the mapping is a set of pairs of nodes of treeA x treeB. 
   It is stored in a stack. */
int trace_back_alignment_mapping(postorder_tree * treeA, postorder_tree * treeB, int **  supertree){
  int score; 
  int i, j, k, l,positionA, positionB; 
  seven_stack *  p;
  int old_i=0; 
  int old_j=0;
  int t;

  p= malloc(sizeof(seven_stack));
  init_seven_stack(p,treeA->size+treeB->size);
  computeTables(treeA, treeB, 0,0,treeA->size-1, treeB->size-1);
  t=best_table(treeA->size-1, treeB->size-1,ABCD);
  score=table[t][treeA->size-1][treeB->size-1];
 
  build_optimal_path(treeA,treeB,0,0,treeA->size-1,treeB->size-1,t,0,0,p,supertree);
   while(!is_empty_seven_stack(p)){
    seven_pop(p,&i,&j, &positionA, &positionB,&k,&l,&t);
     if ((i!=old_i) ||(j!=old_j)){
       old_i=i; old_j=j;
       computeTables(treeA, treeB, i,j,k,l);
     }   
     build_optimal_path(treeA, treeB, i,j, k, l,t, positionA, positionB,p,supertree);
  }
  return score;
}

int optimal_arc_alignment(RNA_sequence * rnaA, RNA_sequence * rnaB,  RNA_sequence* alignA, RNA_sequence* alignB, int primary,
			  int similar, int eq, char *eqA, char *eqB){

  int i,j,score;
  int ** supertree;
  postorder_tree * treeA;
  postorder_tree * treeB;

  treeA=(postorder_tree * ) malloc(sizeof(postorder_tree));
  treeB=(postorder_tree * ) malloc(sizeof(postorder_tree));
  
  RNA_sequence2postorder(rnaA,treeA);
  RNA_sequence2postorder(rnaB,treeB);

  supertree = (int **)malloc( sizeof(int*)*(rnaA->size+1) );
    for (i=0; i<=rnaA->size; i++)
      supertree[i]= (int *)malloc( sizeof(int)*(rnaB->size+1) );
    for (i=0; i<=rnaA->size; i++)
       for (j=0; j<=rnaB->size; j++)
	 supertree[i][j]=0;

    
  /*   prise en compte des contraintes */
  constraint=new_matrix_char(treeA,treeB);	
  no_constraint(constraint,treeA->size,treeB->size);

  if( eq ){
    full_constraint(constraint, treeA->size, treeB->size);
    eq_constraint(constraint, (postorder_tree_m *)treeA, (postorder_tree_m *)treeB, eqA, eqB, rnaA->size, rnaB->size, 0);
  }
    
  if (similar==1)
    diagonal_constraint(constraint,treeA->size,treeB->size, rnaA->size/4+1);  

  if(primary==1)
    primary_sequence_constraint(constraint,treeA,treeB,rnaA,rnaB);
  
  build_alignment_graph(treeA, treeB);
  /*display_alignment_graph(alignment_graph, treeA->size-1, treeB->size-1);*/ 
  score= trace_back_alignment_mapping(treeA, treeB, supertree);
 
  
  create_alignment_RNA_sequences(rnaA,rnaB, supertree, alignA,  alignB, 1, rnaA->size-1, 1, rnaB->size-1);
    
  return score;
}



