/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or */
/* (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "../headers/score.h"
#include "../headers/constraint.h"
#include "../headers/stack.h"
#include "../headers/RNA_structure.h"
#include "../headers/arc_alignment_graph.h"
#include "../headers/graph.h"
#include "../headers/matrix.h"


/* regarde si le noeud x marque le debut d'une nouvelle tige */
int new_stem(postorder_tree * tree, int x){
  if (x>1)
    return (tree->n[x].size!=tree->n[x-1].size+1);
  else
    return 1;
}

/* number of closed subforests ending at position i, 
 * including i */
void tree_closed(postorder_tree * tree, short int * tab_closed){
  int i; 
  int j; 
  int * father;
  father=(int*)malloc(sizeof(int)*tree->size); 
  for (i=1;i<tree->size; i++){
    father[i]=-1;
  }
  for (i=1;i<tree->size; i++){
    if (tree->n[i].size>1) 
      for (j=i-1; j>i-tree->n[i].size; j=j-tree->n[j].size)
        father[j]=i;
  }
  for (i=1; i<tree->size;i++){  
    j=i-tree->n[i].size; 
    if ((j>0) && ((father[i]==-1) ||(father[i]==father[j])))
      tab_closed[i]=tab_closed[j]+1;
    else 
      tab_closed[i]=1;
  }

  free(father);
}

/* right_sibling[i]=x : x est le frere droit de i, s'il existe, -1 sinon */
void right_sibling(postorder_tree * tree, int* tab){
 int i ;  
 for (i=1;i<tree->size; i++)
   tab[i]=-1;
 for (i=1;i<tree->size; i++)
   tab[i-tree->n[i].size]=i;
}

void is_closed_forest(char ** tab, postorder_tree * tree){
  int x=0;
  int fils[tree->size];
  int y, z,t;
  int i,j;
  for (i=1; i<tree->size; i++)
    for (j=1; j<tree->size; j++)
      tab[i][j]=0;

  for (i=1; i<tree->size; i++){
    if (tree->n[i].size>1){ /* noeud interne */
      x=0;
      for (j=i-1;  j>i-tree->n[i].size; j=j-tree->n[j].size){ /* recherche des fils */
	fils[x]=j; x++;
      }
      for (j=0; j<x;j++){
	t=fils[j];
	for (y=0;y<=j; y++){
	  z=fils[y];
	  tab[t-tree->n[t].size+1][z]=1;
	  tab[z][t-tree->n[t].size+1]=1;
	}
      }
    }/* fin noeud interne */
  } 
  x=0;
 
  for (j=tree->size-1;  j>0; j=j-tree->n[j].size){
    fils[x]=j; x++;
   }
  for (j=0; j<x;j++){
    t=fils[j];
    for (y=0;y<=j; y++){
      z=fils[y];
      tab[t-tree->n[t].size+1][z]=1;
      tab[z][t-tree->n[t].size+1]=1;
    }
  } 
}

/* rightmost_sibling[i]=x : x est le frere le plus a droite, s'il existe, i sinon*/
void rightmost_sibling(postorder_tree * tree, int* tab){
  int * tab2;
  int i; 
  tab2=(int *) malloc(sizeof(int)*(tree->size+1));
  right_sibling(tree, tab2);
  for (i=tree->size-1; i>0; i--){
    if (tab2[i]!=-1) 
      tab[i]=tab[tab2[i]];
    else 
      tab[i]=i;
  }
}


/* pas de cout d'ouverture */
int * tree_full_deletion(postorder_tree * tree){
  int w=0;
  int i,x;
  int * full_deletion; 
  full_deletion=(int*)malloc(sizeof(int)*tree->size);
  for (i=1; i<tree->size; i++){
    w=0;
    for (x=i; x>i-tree->n[i].size; x--)
      if (tree->n[x].size==1)
	w+=base_deletion(tree->n[x].labelX); 
      else
	w+=arc_removing(tree->n[x].labelX, tree->n[x].labelY);
    full_deletion[i]=w;
  }
  return full_deletion; 
}

/* retourne max{table[i][x][y]+z}, avec i donn par table_list (ABCD,ACD, etc.) */
int get_tables(int z, int x, int y, char* table_list){
  int w=INT_MIN;
  int i;

  if (constraint[x][y]!=ALLOWED)
    return INT_MIN; 
  else {
  for (i=0; i<4; i++)
    if ((table_list[i]=='1') && (table[i][x][y]>w))
      w=table[i][x][y];
  
  if (w>INT_MIN){
    return z+w; 
  }
  else 
    return INT_MIN;
  }
}


int best_table(int x, int y, char* table_list){
  int w=INT_MIN;
  int t=0;
  int i;
  if ((x>=0) && (y>=0) && (constraint[x][y]==ALLOWED)) 
  for (i=0; i<4; i++)
    if ((table_list[i]=='1') && (table[i][x][y]>w)){
      w=table[i][x][y];
      t=i;
    }
  return t;
}

void computeTables_Cell(postorder_tree * treeA, postorder_tree * treeB,int x,int y, int i, int j){
  int  d;
  edge_gap ed;
  int w0=INT_MIN; 
  int w1=INT_MIN;
  int w2=INT_MIN;
  int w3=INT_MIN; /* scores de table[i][x][y],  respectivement */ 
  int w; 
  int is_new_stem=0;

  /*******************/
  /* implicit edges  */
  /*******************/
    
  /* full deletion in A */
  if (x>i){
    w1=get_tables(full_deletionA[x], x-treeA->n[x].size,y,B);
    w=get_tables(full_deletionA[x]+gap_open,  x-treeA->n[x].size,y,ACD);
    if (w>w1) w1=w;
  } 

  /* full deletion in B */
  if (y>j){
    w=get_tables(full_deletionB[y],x,y-treeB->n[y].size,B);
    if (w>w1) w1=w;
    w=get_tables(full_deletionB[y]+gap_open,x,y-treeB->n[y].size,ACD);
    if (w>w1) w1=w;
  }
  
  /* base-mismatch */
  if ((x>i)&& (y>j) && (treeA->n[x].size==1) && (treeB->n[y].size==1)) { 
    w0=get_tables(base_mismatch(treeA->n[x].labelX,treeB->n[y].labelX), x-1,y-1,ABCD);
  }
  
  /*******************/
  /* explicit edges  */
  /*******************/
 
  for (d=0; d< alignment_graph[x][y].current_edge +1; d++){
    get_edge_gap(&alignment_graph[x][y], &ed, d);
    if (ed.edit_operation<5) /* operation on treeA */
      is_new_stem=new_stem(treeA,x);
    else /* operation sur treeB */
      is_new_stem=new_stem(treeB,y);
    if ((ed.targetX>=i) && (ed.targetY>=j)){
      w=get_tables(ed.weight,ed.targetX,ed.targetY, ABCD);
      if ((ed.edit_operation== 0) || ((ed.origin==3) && !is_new_stem)){ /* arc-mismatch or "stem" state*/
	if (w>w3) w3=w;
      }
	else { /* "maybe steem" state*/
	  if (w>w2) w2=w;
	}
    }
  }

  table[0][x][y]=w0;
  table[1][x][y]=w1;
  table[2][x][y]=w2;
  table[3][x][y]=w3;
}

/* computes the score of the optimal path in the rectangle[i,j]-[k,l] */
void computeTables(postorder_tree * treeA, postorder_tree * treeB,int i,int j, int k, int l){
  int x,y;

  /* initialisation */
  table[0][i][j]=0;
  table[1][i][j]=gap_open; 
  table[2][i][j]=0;
  table[3][i][j]=INT_MIN;

  /* first row */
  for (x=i+1; x<=k; x++)
    if  (constraint[x][j]==ALLOWED){
      table[0][x][j]=INT_MIN;
      table[1][x][j]=get_tables(full_deletionA[x], x-treeA->n[x].size,j,B);   
      table[2][x][j]=INT_MIN; 
      table[3][x][j]=INT_MIN;
    }   

  /* first column */
  for (y=j+1; y<=l; y++)
    if (constraint[i][y]==ALLOWED){
      table[0][i][y]=INT_MIN;
      table[1][i][y]=get_tables(full_deletionB[y],i,y-treeB->n[y].size,B); 
      table[2][i][y]=INT_MIN;
      table[3][i][y]=INT_MIN;
    }
  
  /* other cells */
  for (x=i+1; x<=k; x++) 
    if (is_closedA[i+1][x])
      for (y=j+1; y<=l; y++)
	if (is_closedB[j+1][y])
	  if (constraint[x][y]==ALLOWED)
	    computeTables_Cell(treeA, treeB,x,y,i,j);
}

 
void compute_new_value(postorder_tree * tree, int u, int opening, int extension, int x, int y, char eo, edge_gap *ed2, edge_gap * ed3){
  int w, z;  
 
  /* u is a new stem */
  if (new_stem(tree, u)){ /* ouverture de gap obligatoire : maybe stem obligatoire*/
    w=get_tables(opening,x,y,ABCD); 
    if (w>(*ed2).weight){
      (*ed2).weight=w;
      (*ed2).edit_operation=eo;
      (*ed2).origin=best_table(x,y,ABCD);
    }
  } 
  else {/* extension possible : deux etats possibles*/  
    /* maybe stem */
    w=get_tables(opening,x,y,AB);
    z=get_tables(extension,x,y,C);
    if (z>w) {
      if (z>(*ed2).weight){
	(*ed2).weight=z;
	(*ed2).edit_operation=eo;
	(*ed2).origin=2;
      }
    }
    else if (w>(*ed2).weight){
      (*ed2).weight=w;
      (*ed2).edit_operation=eo;
      (*ed2).origin=best_table(x,y,AB);
    }
    /* stem */
    w=get_tables(extension,x,y,D);
    if (w>(*ed3).weight){
      (*ed3).weight=w;
      (*ed3).edit_operation=eo;
      (*ed3).origin=3;
    }
  }
}

/* build the alignment graph for two trees A and B */
/* Each edge has type "edge" */
void build_alignment_graph(postorder_tree * treeA, postorder_tree * treeB){
  int i,j,k,l,m,n,x,y;
  short int * tree_closedA;
  short int * tree_closedB; 
  int * rightmostA;
  int * rightmostB;
  int * ancestorA;
  int * ancestorB;
  int z;
  edge_gap ed2, ed3; 
  int w;
  int degree=0;
  int nee=0; /* utilite ? */
  
  tree_closedA=(short int*)malloc(sizeof(short int)*(treeA->size+1));
  tree_closedB=(short int*)malloc(sizeof(short int)*(treeB->size+1)); 
  tree_closed(treeA,tree_closedA);
  tree_closed(treeB,tree_closedB);
  
  ancestorA=(int*)malloc(sizeof(int)*(treeA->size+1));
  ancestorB=(int*)malloc(sizeof(int)*(treeB->size+1)); 
  ancestor(treeA,ancestorA);
  ancestor(treeB,ancestorB);
  
  rightmostA=(int*)malloc(sizeof(int)*(treeA->size+1)); 
  rightmostB=(int*)malloc(sizeof(int)*(treeB->size+1));
  rightmost_sibling(treeA,rightmostA); 
  rightmost_sibling(treeB,rightmostB);
    
  is_closedA= (char**) malloc(sizeof(char*)*(treeA->size+1));
  for (i=0; i<=treeA->size; i++)
    is_closedA[i]=(char*) malloc(sizeof(char)*(treeA->size+1));
  is_closed_forest(is_closedA, treeA);
 
  is_closedB= (char**) malloc(sizeof(char*)*(treeB->size+1));
  for (i=0; i<=treeB->size; i++)
    is_closedB[i]=(char*) malloc(sizeof(char)*(treeB->size+1));
  is_closed_forest(is_closedB, treeB);
    
  /***********************************/
  /* insertions et deletions totales */
  /***********************************/
    full_deletionA=tree_full_deletion(treeA); 
    full_deletionB=tree_full_deletion(treeB);

   /**********************************/
  /*    skeleton of the graph       */
  /**********************************/ 
  
  alignment_graph= (vertex_gap**) malloc(sizeof(vertex_gap*)*(treeA->size+1));
  for (i=0; i<=treeA->size; i++)
    alignment_graph[i]=(vertex_gap*) malloc(sizeof(vertex_gap)*(treeB->size+1));

  for (i=1; i<treeA->size; i++){
    for (j=1; j<treeB->size; j++){
      if (constraint[i][j]==ALLOWED){
        if ((treeA->n[i].size>1) &&  (treeB->n[j].size>1)) degree=2*(tree_closedB[j]+tree_closedA[i]);
	else if (treeA->n[i].size>1) degree=2* tree_closedB[j]; 
	else if (treeB->n[j].size>1) degree=2* tree_closedA[i];
	else degree=0; 
	init_edge_stack_gap(&alignment_graph[i][j],degree);
      }
      else init_edge_stack_gap(&alignment_graph[i][j],0);
    }
  }
    
  
      /*if ((treeA->n[i].size>1) &&  (treeB->n[j].size>1)) degree=(2+ tree_closedB[j]+tree_closedA[i])*2;
	else if ((treeA->n[i].size==1) && (treeB->n[j].size>1)) degree=(1+tree_closedA[i])*2;
	else if ((treeA->n[i].size>1) && (treeB->n[j].size==1)) degree=(1+tree_closedB[j])*2; 
	else degree=0; 
	init_edge_stack_gap(&alignment_graph[i][j],degree);
      }
    }
    }*/

  /*******************************************/
  /* allocation des quatre tables            */
  /*******************************************/
  table=(int***)malloc(sizeof(int**)*4);
  for (i=0; i<4; i++){
    table[i]=(int**) malloc(sizeof(int*)*(treeA->size+1));
    for (j=0; j<=treeA->size; j++)
      table[i][j]=(int*) malloc(sizeof(int)*(treeB->size+1));
  }

  /*****************************/ 
  /*  weight of the edges      */
  /*****************************/
  
  for (i=treeA->size-1; i>=1; i--){
    if (treeA->n[i].size==1) {
      for (j=treeB->size-1; j>=1; j--){
	if  ( (treeB->n[j].size==1) && (constraint[i-1][j-1]==ALLOWED)){
	  k=ancestorA[i]; 
	  l=ancestorB[j];
	  m=rightmostA[k];
	  n=rightmostB[l];
	  	  
	  /*************************************************/
	  /*        initialisation des quatre tables       */
	  /*************************************************/
	  
	  table[0][i-1][j-1]=0;
	  table[1][i-1][j-1]=gap_open;
	  table[2][i-1][j-1]=0;
	  table[3][i-1][j-1]=INT_MIN;
	  
	  /* deletions  */
	  for (x=i; x<=m; x++)
	    if (constraint[x][j-1]==ALLOWED){
	      table[0][x][j-1]=INT_MIN;
	      table[1][x][j-1]=get_tables(full_deletionA[x], x-treeA->n[x].size,j-1,B);
	      table[2][x][j-1]=INT_MIN;
	      table[3][x][j-1]=INT_MIN;
	    }
	  
	  /* insertions */
	  for (y=j; y<=n; y++)
	    if (constraint[i-1][y]==ALLOWED){
	      table[0][i-1][y]=INT_MIN;		
	      table[1][i-1][y]=get_tables(full_deletionB[y],i-1,y-treeB->n[y].size,B); 
	      table[2][i-1][y]=INT_MIN;
	      table[3][i-1][y]=INT_MIN;
	    }
	  
	  
	  /**********************/
	  /*       zone I       */
	  /**********************/
	  for (x=i; x<=k; x++){
	    if (is_closedA[i][x]){
	      for (y=j; y<=l; y++){  
		if (is_closedB[j][y])
		  if (constraint[x][y]==ALLOWED){
		  computeTables_Cell(treeA,treeB,x,y,i-1,j-1);
		  ed2.table=2;
		  ed3.table=3;
		  /**************************************/
		  /* arc allant de (x,y) vers (i-1,j-1) */
		  /**************************************/
		  
		  ed2.targetX= i-1; 
		  ed2.targetY= j-1;
		  ed3.targetX= i-1; 
		  ed3.targetY= j-1;
		  ed2.weight=INT_MIN; 
		  ed3.weight=INT_MIN; 
		  
		  /* arc_mismatch entre  x et y : "stem" state */
		  if (is_subtree(treeA,i,x) && is_subtree(treeB,j,y) ) { 		
		    w=get_tables(arc_mismatch(treeA->n[x].labelX, treeA->n[x].labelY,treeB->n[y].labelX,treeB->n[y].labelY),x-1,y-1,ABD);
		    z=get_tables(arc_mismatch(treeA->n[x].labelX, treeA->n[x].labelY,treeB->n[y].labelX,treeB->n[y].labelY) - gap_open,x-1,y-1,C);
		    if (z>w) {
		      if (z> ed3.weight){
			 ed3.weight=z;
			 ed3.edit_operation=0;
			 ed3.origin=2; 
		      }
		    }
		    else if (w>ed3.weight){
		      ed3.weight=w;
		      ed3.edit_operation=0;
		      ed3.origin=best_table(x-1,y-1,ABD);
		    }
		  }
		  
		  /* arc-removing de x */      
		  if (is_subtree(treeA,i,x)) 
		    compute_new_value(treeA, x, arc_removing(treeA->n[x].labelX, treeA->n[x].labelY)+gap_open,arc_removing(treeA->n[x].labelX, treeA->n[x].labelY),x-1,y,1, &ed2,&ed3);		
		  
		  /* arc-altering de x en y avec match en 3'*/
		  if  (is_subtree(treeA,i,x) && (treeB->n[y].size==1)) 
		    compute_new_value(treeA,x,arc_altering2(treeA->n[x].labelX, treeA->n[x].labelY,treeB->n[y].labelX )+gap_open,arc_altering2(treeA->n[x].labelX, treeA->n[x].labelY,treeB->n[y].labelX ),x-1,y-1,3, &ed2,&ed3);
		  
		  /* arc-removing de y */
		  if (is_subtree(treeB,j,y)) 
		    compute_new_value(treeB,y,arc_removing(treeB->n[y].labelX, treeB->n[y].labelY)+gap_open,arc_removing(treeB->n[y].labelX, treeB->n[y].labelY),x,y-1,5, &ed2,&ed3);		
		  
		  /* arc-altering de y en x avec match en 3'*/
		  if ( is_subtree(treeB,j,y) && (treeA->n[x].size==1) )
		    compute_new_value(treeB,y,arc_altering2(treeB->n[y].labelX,treeB->n[y].labelY, treeA->n[x].labelX)+gap_open,arc_altering2(treeB->n[y].labelX,treeB->n[y].labelY, treeA->n[x].labelX), x-1,y-1,7, &ed2, &ed3);
		 		  
		  if (table[2][x][y]<ed2.weight){
		    update_edge_gap(&alignment_graph[x][y], ed2,&nee);
		    table[2][x][y]=ed2.weight;
		  }
		  
		  if (table[3][x][y]<ed3.weight){
		    update_edge_gap(&alignment_graph[x][y], ed3,&nee);
		    table[3][x][y]=ed3.weight;
		  }
	   
	  	     
		  /**************************************/
		  /* arc allant de (x,y) vers (i-1,j-2) */
		  /**************************************/
		  if ((j>1) &&  (constraint[i-1][j-2]==ALLOWED) && is_subtree(treeA,i,x) && is_closedB[j-1][y]  && (treeB->n[j-1].size==1) ) {
		    
		    ed2.targetX=i-1;
		    ed2.targetY=j-2;
		    ed2.weight=INT_MIN;
		    ed3.targetX=i-1;
		    ed3.targetY=j-2;
		    ed3.weight=INT_MIN;
		    
		    /* arc-altering de x en j-1, avec match en 5'*/
		    compute_new_value(treeA,x,arc_altering1(treeA->n[x].labelX, treeA->n[x].labelY,treeB->n[j-1].labelX )+gap_open,arc_altering1(treeA->n[x].labelX, treeA->n[x].labelY,treeB->n[j-1].labelX ),x-1,y,4,&ed2,&ed3);
		    
		    /* arc-breaking de x  en (j-1,y)*/		
		    if  (treeB->n[y].size==1)
		      compute_new_value(treeA,x,arc_breaking(treeA->n[x].labelX, treeA->n[x].labelY, treeB->n[j-1].labelX, treeB->n[y].labelX)+gap_open,arc_breaking(treeA->n[x].labelX, treeA->n[x].labelY, treeB->n[j-1].labelX, treeB->n[y].labelX),x-1,y-1,2,&ed2,&ed3);
		    
		    if  (ed2.weight>INT_MIN)
		      update_edge_gap(&alignment_graph[x][y], ed2,&nee);
		    if   (ed3.weight>INT_MIN)
		      update_edge_gap(&alignment_graph[x][y], ed3,&nee);
		  }
		  
		  /**************************************/	         
		  /* arc allant de (x,y) vers (i-2,j-1) */
		  /**************************************/	
		  
		  if ((i>1) && (constraint[i-2][j-1]==ALLOWED) && is_subtree(treeB,j,y) && is_closedA[i-1][x]  && (treeA->n[i-1].size==1)){
		    
		    ed2.targetX=i-2; 
		    ed2.targetY=j-1;	
		    ed2.weight=INT_MIN;
		    ed3.targetX=i-2; 
		    ed3.targetY=j-1;	
		    ed3.weight=INT_MIN;
		    
		     /* arc-altering de y en i-1 */
		    compute_new_value(treeB,y,arc_altering1(treeB->n[y].labelX,treeB->n[y].labelY,treeA->n[i-1].labelX )+gap_open, arc_altering1(treeB->n[y].labelX,treeB->n[y].labelY,treeA->n[i-1].labelX ), x,y-1,8,&ed2,&ed3);
		    
		    /* arc-breaking de y  en (i-1,x)*/
		    if  (treeA->n[x].size==1) 
		      compute_new_value(treeB,y,arc_breaking(treeB->n[y].labelX, treeB->n[y].labelY, treeA->n[i-1].labelX, treeA->n[x].labelX)+gap_open,arc_breaking(treeB->n[y].labelX, treeB->n[y].labelY, treeA->n[i-1].labelX, treeA->n[x].labelX),x-1,y-1,6,&ed2,&ed3); 
		    
		    if (ed2.weight>INT_MIN)
		      update_edge_gap(&alignment_graph[x][y], ed2,&nee);
		    if (ed3.weight>INT_MIN)
		      update_edge_gap(&alignment_graph[x][y], ed3,&nee);
		  }
		}
	      }
	    }
	  }/* end zone I */
	  
	  /*********************/
	  /*     zone II       */
	  /*********************/
	  
	  for (x=k+1; x<=m; x++){
	    if (is_closedA[i][x]){
	      for (y=j; y<=l; y++){
		if (is_closedB[j][y])
		  if (constraint[x][y]==ALLOWED){
		  computeTables_Cell(treeA, treeB,x,y,i-1,j-1);
		  
		  /***************************************/
		  /* arc allant de (x,y) vers (i-1, j-1) */
		  /***************************************/
		  
		  ed2.targetX= i-1; 
		  ed2.targetY= j-1;	      
		  ed2.weight=table[2][x][y];
		  ed3.targetX= i-1; 
		  ed3.targetY= j-1;	      
		  ed3.weight=table[3][x][y];
		  
		  /* arc-removing de y */
		  if (is_subtree(treeB,j,y))
		    compute_new_value(treeB,y,arc_removing(treeB->n[y].labelX, treeB->n[y].labelY)+gap_open,arc_removing(treeB->n[y].labelX, treeB->n[y].labelY),x,y-1,5,&ed2,&ed3);
		  
		  /* arc-altering de y en x */
		  if  ((is_subtree(treeB,j,y)) && (treeA->n[x].size==1))
		    compute_new_value(treeB,y,arc_altering2(treeB->n[y].labelX,treeB->n[y].labelY, treeA->n[x].labelX)+gap_open,arc_altering2(treeB->n[y].labelX,treeB->n[y].labelY, treeA->n[x].labelX),x-1,y-1,7,&ed2,&ed3);
		  		  
		  if (table[2][x][y]<ed2.weight){
		    update_edge_gap(&alignment_graph[x][y], ed2,&nee);
		    table[2][x][y]=ed2.weight;
		  }
		  
		  if (table[3][x][y]<ed3.weight){
		    update_edge_gap(&alignment_graph[x][y], ed3,&nee);
		    table[3][x][y]=ed3.weight;
		  }
		  
		  /***********************************/
		  /* arc menant de (x,y) a (i-2,j-1) */
		  /***********************************/
		  
		  if ((i>1) &&  (constraint[i-2][j-1]==ALLOWED) && is_subtree(treeB,j,y) && is_closedA[i-1][x]  && (treeA->n[i-1].size==1)){
		    
		    ed2.targetX= i-2; 
		    ed2.targetY= j-1;
		    ed2.weight=INT_MIN;
		    ed3.targetX= i-2; 
		    ed3.targetY= j-1;
		    ed3.weight=INT_MIN;
		    
		    /* arc-altering de y en i-1 */
		    compute_new_value(treeB,y, arc_altering1(treeB->n[y].labelX,treeB->n[y].labelY, treeA->n[i-1].labelX)+gap_open,arc_altering1(treeB->n[y].labelX,treeB->n[y].labelY, treeA->n[i-1].labelX),x,y-1,8,&ed2,&ed3);
		    /* arc-breaking de y  en (i-1,x)*/		    
		    if (treeA->n[x].size==1)
		      compute_new_value(treeB,y,arc_breaking(treeB->n[y].labelX, treeB->n[y].labelY, treeA->n[i-1].labelX, treeA->n[x].labelX)+gap_open,arc_breaking(treeB->n[y].labelX, treeB->n[y].labelY, treeA->n[i-1].labelX, treeA->n[x].labelX),x-1,y-1,6,&ed2,&ed3);
		    
		    if (ed2.weight>INT_MIN)
		      update_edge_gap(&alignment_graph[x][y], ed2,&nee);
		    if (ed3.weight>INT_MIN)
		      update_edge_gap(&alignment_graph[x][y], ed3,&nee);
			
		  }/* fin arcs menant de (x,y) a (i-2,j-1) */
		} /* if closedB */
	      } /* for y */
	    } /* if closedA */
	  }/* fin zone II - for x*/
	  
	  /************************/
	  /*      zone III        */
	  /************************/
	  for (x=i; x<=k; x++) {
	    if (is_closedA[i][x]){
	      for (y=l+1; y<=n; y++){
		if (is_closedB[j][y]) 
		  if (constraint[x][y]==ALLOWED){
		  computeTables_Cell(treeA, treeB,x,y,i-1,j-1);
		  
		  /***********************************/
		  /* arc menant de (x,y)  (i-1,j-1) */
		  /***********************************/
		  
		  ed2.targetX= i-1; 
		  ed2.targetY= j-1;	      
		  ed2.weight=table[2][x][y];
		  ed3.targetX= i-1; 
		  ed3.targetY= j-1;	      
		  ed3.weight=table[3][x][y];
		  
		  /* arc-removing de x */      
		  if (is_subtree(treeA,i,x)) 
		    compute_new_value(treeA,x,arc_removing(treeA->n[x].labelX, treeA->n[x].labelY)+gap_open,arc_removing(treeA->n[x].labelX, treeA->n[x].labelY),x-1,y,1, &ed2,&ed3);
		  
		  /* arc-altering de x en y, match en 3' */
		  if (is_subtree(treeA,i,x) &&  (treeB->n[y].size==1) )
		    compute_new_value(treeA,x,arc_altering2(treeA->n[x].labelX, treeA->n[x].labelY,treeB->n[y].labelX )+gap_open,arc_altering2(treeA->n[x].labelX, treeA->n[x].labelY,treeB->n[y].labelX ),x-1,y-1,3,&ed2,&ed3);
		  		  
		  if (table[2][x][y]<ed2.weight){
		    update_edge_gap(&alignment_graph[x][y], ed2,&nee);
		    table[2][x][y]=ed2.weight;
		  }
		  
		  if (table[3][x][y]<ed3.weight){
		    update_edge_gap(&alignment_graph[x][y], ed3,&nee);
		    table[3][x][y]=ed3.weight;
		  }
		  
		  /* arc menant de (x,y) a (i-1,j-2) */
		  
		  if ((j>1)  && (constraint[i-1][j-2]==ALLOWED) && is_subtree(treeA,i,x) && is_closedB[j-1][y] && (treeB->n[j-1].size==1) ){
		    ed2.targetX= i-1; 
		    ed2.targetY= j-2;
		    ed2.weight=INT_MIN;
		    ed3.targetX= i-1; 
		    ed3.targetY= j-2;
		    ed3.weight=INT_MIN;
			    
		    /* arc-altering de x en j-1, match en 5' */
		    compute_new_value(treeA,x,arc_altering1(treeA->n[x].labelX, treeA->n[x].labelY,treeB->n[j-1].labelX )+gap_open,arc_altering1(treeA->n[x].labelX, treeA->n[x].labelY,treeB->n[j-1].labelX ),x-1,y,4,&ed2,&ed3);
		    
		    /* arc-breaking de x  en (j-1,y)*/
		    if (treeB->n[y].size==1)   
		      compute_new_value(treeA,x,arc_breaking(treeA->n[x].labelX, treeA->n[x].labelY, treeB->n[j-1].labelX, treeB->n[y].labelX)+gap_open, arc_breaking(treeA->n[x].labelX, treeA->n[x].labelY, treeB->n[j-1].labelX, treeB->n[y].labelX),x-1,y-1,2,&ed2,&ed3);
		    
		    if (ed2.weight>INT_MIN)
		      update_edge_gap(&alignment_graph[x][y], ed2,&nee);
		    if (ed3.weight>INT_MIN)
		      update_edge_gap(&alignment_graph[x][y], ed3,&nee);
		  } /* j>1 */
		}
	      }
	    }
	  }/* fin zone III */
	}
      }  
    } 
  }
  
  free(rightmostA);
  free(rightmostB);
  free(tree_closedA);
  free(tree_closedB);
}

/* optimal path in the alignment delineated by the
   rectangle  i,j,k,l  with (k,l) in  table t*/
void build_optimal_path(postorder_tree * treeA,postorder_tree * treeB, int i,int j,int k,int l, int t,  int positionA, int positionB, seven_stack * p, int ** supertree){
  int x,y;
  int d; 
  int finished=0;
  edge_gap ed;   
   
  x=k; 
  y=l;
  ed.edit_operation=-1;
  while (x+y>i+j){

    finished=0;

    /*******************/
    /* implicit edges  */
    /*******************/
    
    if (t==0){ /* base-mismatch */
      finished=1;    
      supertree[treeA->n[x].positionX][treeB->n[y].positionX]=1;
      x--; y--; 
      t=best_table(x,y,ABCD);
    }
    else if (t==1) { /* full deletion */
      finished=1; 
      if ((x>i) && (table[1][x][y]==get_tables(full_deletionA[x], x-treeA->n[x].size, y, B))){ 
	for (d=x; d>x-treeA->n[x].size; d--){
	  supertree[treeA->n[d].positionX][max2(positionB,treeB->n[y].positionY)]=-1;
	  supertree[treeA->n[d].positionY][max2(positionB,treeB->n[y].positionY)]=-1;
	}
	x-=treeA->n[x].size; 
      }
      else if ((x>i) && (table[1][x][y]==get_tables(full_deletionA[x]+gap_open, x-treeA->n[x].size, y, ACD))){
	for (d=x; d>x-treeA->n[x].size; d--){	
	  supertree[treeA->n[d].positionX][max2(positionB,treeB->n[y].positionY)]=-1;
	  supertree[treeA->n[d].positionY][max2(positionB,treeB->n[y].positionY)]=-1;
	}	
	x-=treeA->n[x].size; 
	t=best_table(x,y,ACD); 
      }
      else if ((y>j) && (table[1][x][y]==get_tables(full_deletionB[y],x, y-treeB->n[y].size, B))){
	for (d=y; d>y-treeB->n[y].size; d--){
	  supertree[max2(positionA,treeA->n[x].positionY)][treeB->n[d].positionX]=-2;
	  supertree[max2(positionA,treeA->n[x].positionY)][treeB->n[d].positionY]=-2;
	}
	y-=treeB->n[y].size;	
      }
      else  if ((y>j) && (table[1][x][y]==get_tables(full_deletionB[y]+gap_open,x, y-treeB->n[y].size, ACD))){
	for (d=y; d>y-treeB->n[y].size; d--){	
	  supertree[max2(positionA,treeA->n[x].positionY)][treeB->n[d].positionX]=-2;
	  supertree[max2(positionA,treeA->n[x].positionY)][treeB->n[d].positionY]=-2;
	}
	y-=treeB->n[y].size;
	t=best_table(x,y,ACD);	
      }
    }
	
    /*******************/
    /* explicit edges  */
    /*******************/

    for (d=0; (!finished) && (d<=alignment_graph[x][y].current_edge); d++){
      get_edge_gap(&alignment_graph[x][y], &ed, d);
      if ((ed.targetX>=i)&&(ed.targetY>=j) && (table[t][x][y]==get_tables(ed.weight,ed.targetX,ed.targetY,ABCD))){
	finished=1; 
	switch (ed.edit_operation) {
	case 0 :  /* arc-(mis)match */
	  supertree[treeA->n[x].positionX][treeB->n[y].positionX]=1; 
	  supertree[treeA->n[x].positionY][treeB->n[y].positionY]=1; 
	  seven_push(p,ed.targetX,ed.targetY,treeA->n[x].positionX, treeB->n[y].positionX,x-1,y-1,ed.origin); 
	  break;
	case 5 :  /* arc_insertion */ 
	  supertree[max2(positionA,treeA->n[ed.targetX].positionY)][treeB->n[y].positionX]=-2;
	  supertree[treeA->n[x].positionY][treeB->n[y].positionY]=-2;
	  seven_push(p, ed.targetX,ed.targetY,  max2(positionA,treeA->n[ed.targetX].positionY),  treeB->n[y].positionX,x, y-1,ed.origin);
	  break; 
	case 1:  /* arc_deletion */   
	  supertree[treeA->n[x].positionX][max2(positionB,treeB->n[ed.targetY].positionY)]=-1;
	  supertree[treeA->n[x].positionY][treeB->n[y].positionY]=-1;
	  seven_push(p,ed.targetX,ed.targetY, treeA->n[x].positionX, max2(positionB,treeB->n[ed.targetY].positionY),x-1,y,ed.origin);
	  break;
	case 6:/* arc_breaking sur treeB */
	  supertree[treeA->n[ed.targetX+1].positionX][treeB->n[y].positionX]=1;
	  supertree[treeA->n[x].positionX][treeB->n[y].positionY]=1;
	  seven_push(p,ed.targetX+1,ed.targetY, treeA->n[ed.targetX+1].positionX,treeB->n[y].positionX,x-1,y-1,ed.origin); 
	  break;
	case 2 : /* arc_breaking sur treeA */
	   supertree[treeA->n[x].positionX][treeB->n[ed.targetY+1].positionX]=1;
	  supertree[treeA->n[x].positionY][treeB->n[y].positionX]=1;
	  seven_push(p,ed.targetX,ed.targetY+1,treeA->n[x].positionX,treeB->n[ed.targetY+1].positionX,x-1,y-1,ed.origin);
	  break;
	case 3 :  /* arc_altering sur treeA avec match sur la base en 3' */
	  supertree[treeA->n[x].positionY][treeB->n[y].positionX]=1;
	  supertree[treeA->n[x].positionX][max2(positionB,treeB->n[ed.targetY].positionY)]=-1;
	  seven_push(p,ed.targetX,ed.targetY,treeA->n[x].positionX,max2(positionB,treeB->n[ed.targetY].positionX),x-1,y-1,ed.origin);
	  break;
	case 7 : /* arc_altering sur treeB avec match sur la base en 3' */
	  supertree[treeA->n[x].positionX][treeB->n[y].positionY]=1;
	  supertree[max2(positionA,treeA->n[ed.targetX].positionY)][treeB->n[y].positionX]=-2;
	  seven_push(p,ed.targetX,ed.targetY, max2(positionA,treeA->n[ed.targetX].positionY),treeB->n[y].positionX,x-1,y-1,ed.origin); 
	  break;
	case 4 :  /* arc_altering sur treeA avec match sur la base en 5' */
	  supertree[treeA->n[x].positionX][treeB->n[ed.targetY+1].positionX]=1; 
	  supertree[treeA->n[x].positionY][treeB->n[y].positionY]=-1;
	  seven_push(p,ed.targetX,ed.targetY+1,treeA->n[x].positionX,max2(positionB,treeB->n[ed.targetY+1].positionX),x-1,y,ed.origin); 
	  break;
	case 8 : /* arc_altering sur treeB avec match sur la base en 5' */
	  supertree[treeA->n[ed.targetX+1].positionX][treeB->n[y].positionX]=1;
	  supertree[treeA->n[x].positionY][treeB->n[y].positionY]=-2; 
	  seven_push(p,ed.targetX+1,ed.targetY,max2(positionA,treeA->n[ed.targetX+1].positionX),treeB->n[y].positionX,x,y-1,ed.origin);
	  break;
	default :
	  printf("*** program error *** \n");fflush(stdout);
	  break;
	} /* end of switch */
	x=ed.targetX; 
	y=ed.targetY;
	t=best_table(x,y,ABCD);
      }/* end of if */
    }/* end of for */
    if (finished==0) { printf("Program error -- No rule: %d, %d,%d,%d\n",i,j,k,l);  exit (1);}
  }
}
