/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or */
/* (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include<stdio.h>
#include<stdlib.h>
#include<malloc.h>
#include<string.h>
#include "../headers/clustering.h"
#include "../headers/RNA_structure.h"
#include "../headers/multiple_arc_alignment.h"

void display_tab(int **tab, int nb_seq){

  int i, j;

  printf("      ");
  for(i=0; i<nb_seq; i++)
    printf("%5i ", i);
  printf("\n");
  printf("     ");
  for(i=0; i<nb_seq; i++)
    printf("------");
  printf("\n");
  
  for(i=0; i<nb_seq; i++){
    printf("%3i|  ",i);
    for(j=0; j<i; j++){
      printf("%5i ", tab[i][j]);
    }
    for(j=0; j<nb_seq-i; j++)
      printf("      ");
    printf("|\n");
  }
  printf("     ");
  for(i=0; i<nb_seq; i++)
    printf("------");
  printf("\n");
 
}

void display_cluster(cluster **tab, int nb_seq){

  int i, j;

  printf("      ");
  for(i=0; i<nb_seq; i++)
    printf("%5i ", i);
  printf("\n");
  printf("     ");
  for(i=0; i<nb_seq; i++)
    printf("------");
  printf("\n");
  
  for(i=0; i<nb_seq; i++){
    printf("%3i|  ",i);
    for(j=0; j<nb_seq; j++){
      printf("%2i:%2i ", tab[i][j].left, tab[i][j].right);
    }
    printf("|\n");
  }
  printf("     ");
  for(i=0; i<nb_seq; i++)
    printf("------");
  printf("\n");
 
}

cluster **cluster_init(int nbre_clusters){
  cluster **liste = NULL;
  int i,j;
 
  /* matrix of cluster allocation */
  liste = (cluster **)malloc(sizeof(cluster *) * nbre_clusters );
  for(i=0;i<nbre_clusters;i++){
    liste[i] = (cluster *)malloc(sizeof(cluster) * nbre_clusters );
  }

  /* 
   * left = right = -1 => leaves of the tree
   * At the beginning, each cluster is a leaf 
   */
  for(i=0; i<nbre_clusters; i++){
    for(j=0; j<nbre_clusters; j++){
      liste[i][j].left = -1;
      liste[i][j].right = -1;
      liste[i][j].seq = (int *)malloc(sizeof(int));
      liste[i][j].seq[0] = j;
      liste[i][j].nbre_seq = 1;
    }
  }

  return liste;
}

/*
 * look for the maximal score in the similarity matrix
 */
void max_score(int **similarite, int *i, int *j, int nbre_clusters, int *mat){
  
  int k,l;
  int max, test;

  max = similarite[1][0];
  test = 0;
  k=1;
  l=0;
  /* look for the first allowed value */
  while(test == 0){
    if(mat[k] == 1){
      if(mat[l] == 1){
	max = similarite[k][l];
	*i = k;
	*j = l;
	test = 1;
      }
    }
    /* next column */
    l++;
    /* if we were at the last column */
    if(l == k){
      k++;/* then next row */
      l=0;/* and first column */
      /* if we were at the last row */
      if(k == nbre_clusters)
	test = 1;/* then we stop */
    }
  }
  
  /* and we look for the max */
  for(k=0; k<nbre_clusters; k++)
    if(mat[k] == 1)
      for(l=0; l<k; l++)
	if(mat[l] == 1)
	  if( similarite[k][l] > max ){
	    max = similarite[k][l];
	    *i = k;
	    *j = l;
	  }
}

int mean(int a, int b){
   return (a+b)/2;
}

void update_sim(int **similarity, int i, int j, int nb_clusters, int *mat){

  int k;

  /* update of the jth row */
  for(k=0; k<j; k++)
    if(mat[k] == 1){
      similarity[j][k] = mean( similarity[j][k], similarity[i][k] );
    }
  
  /* update of the jth column */
  for(k=j+1; k<nb_clusters; k++)
    if(mat[k] == 1){
      similarity[k][j] = mean( similarity[k][j], similarity[i][j] );
    }
}

void perform_clustering(cluster **liste_cluster, int **similarite, int nbre_clusters){

  int nbre;
  int i=0,j=0, k, l;
  int level;
  int *mat = NULL;

  /*
   * During clustering, we need to remove one row and one column from the 
   * matrix at each level. To avoid allocation at each step, we use a boolean 
   * vector that indicates wich rows and columns are forbidden.
   */
  mat = (int *)malloc(sizeof(int) * (nbre_clusters));

  for(k=0; k<nbre_clusters; k++)
    mat[k] = 1;/* we allow every row and every column */

  nbre = nbre_clusters;
  level = 0;
  while(nbre > 1){
    /* look for the nearest structures */
    max_score(similarite, &i, &j, nbre_clusters, mat);
    level++;

    /* make a cluster with the two structures found */
    liste_cluster[level][j].left = j;
    liste_cluster[level][j].right = i;
    /* update the sequences of the new cluster */
    liste_cluster[level][j].nbre_seq = liste_cluster[level-1][j].nbre_seq +
      liste_cluster[level-1][i].nbre_seq;

    liste_cluster[level][j].seq = (int *)realloc(liste_cluster[level][j].seq, 
						 sizeof(int) * (liste_cluster[level][j].nbre_seq) );
    for(k=0; k<liste_cluster[level-1][j].nbre_seq; k++)
      liste_cluster[level][j].seq[k] = liste_cluster[level-1][j].seq[k];
    for(k=liste_cluster[level-1][j].nbre_seq; k<liste_cluster[level][j].nbre_seq; k++)
      liste_cluster[level][j].seq[k] = liste_cluster[level-1][i].seq[k-liste_cluster[level-1][j].nbre_seq];

    for(k=0; k<nbre_clusters; k++){
      if(k != j){
	if(liste_cluster[level-1][k].left == -1 && liste_cluster[level-1][k].right == -1){
	  liste_cluster[level][k].left = liste_cluster[level-1][k].left;
	  liste_cluster[level][k].right = liste_cluster[level-1][k].right;
	  liste_cluster[level][k].seq = (int *)realloc( liste_cluster[level][k].seq,
							liste_cluster[level-1][k].nbre_seq * sizeof(int) );
	  for(l=0; l<liste_cluster[level-1][k].nbre_seq; l++)
	    liste_cluster[level][k].seq[l] = liste_cluster[level-1][k].seq[l];
	  liste_cluster[level][k].nbre_seq = liste_cluster[level-1][k].nbre_seq;
	}
	else{
	  liste_cluster[level][k].left = k;
	  liste_cluster[level][k].right = -1;

	  liste_cluster[level][k].seq = (int *)realloc(liste_cluster[level][k].seq,
						       sizeof(int)*liste_cluster[level-1][k].nbre_seq);
	  for(l=0; l<liste_cluster[level-1][k].nbre_seq; l++)
	    liste_cluster[level][k].seq[l] = liste_cluster[level - 1][k].seq[l];
	  liste_cluster[level][k].nbre_seq = liste_cluster[level-1][k].nbre_seq;
	}
      }
    }
    /* we delete the ith row and the ith column */
    mat[i] = 0;

    /* update of the similarity matrix */
    update_sim(similarite, i, j, nbre_clusters, mat);

    nbre--;
  }
  free(mat);
}

RNA_sequence_m *clustering(int **similarity, int nb_seq, RNA_sequence_m **list, int *sc, int EQ, int SIMILAR){

  int i, j;
  cluster **cluster_list = NULL;
  int left, right;
  RNA_sequence_m *temp;
  RNA_sequence_m *test;
  int score=0;

  cluster_list = cluster_init(nb_seq);
  
  /* start clustering */
  perform_clustering(cluster_list, similarity, nb_seq);

  temp = (RNA_sequence_m *)malloc( sizeof(RNA_sequence_m) * nb_seq );
  test = (RNA_sequence_m *)malloc( sizeof(RNA_sequence_m) );

  for(i=0; i<nb_seq; i++){
    temp[i].s = NULL;
  }

  for(i=1; i<nb_seq; i++){
    for(j=0; j<nb_seq; j++){
      
      left = cluster_list[i][j].left;
      right = cluster_list[i][j].right;
      /* merge two clusters */
      if( left != -1 && right != -1){
	
	/* merge two leaves */
	if(cluster_list[i-1][left].left == -1 && cluster_list[i-1][right].left == -1){
	  score = optimal_multiple_arc_alignment_mapping( list[left], list[right], &(temp[left]), EQ, SIMILAR );
	}
	/* merge leaf and a node */
	else if( cluster_list[i-1][left].left != -1 && cluster_list[i-1][right].left == -1){

	  /* copy of temp[left] */
	  copy_sequence(test, &(temp[left]));
	  free_RNA_sequence_m(&temp[left]);

	  score = optimal_multiple_arc_alignment_mapping( test, list[right], &(temp[left]), EQ, SIMILAR );
	  
	  free_RNA_sequence_m(test);
	}
	else if( cluster_list[i-1][left].left == -1 && cluster_list[i-1][right].left != -1){
	  score = optimal_multiple_arc_alignment_mapping( &(temp[right]), list[left], &(temp[left]), EQ, SIMILAR );
	}
	/* merge two nodes */
	else if( cluster_list[i-1][left].left != -1 && cluster_list[i-1][right].left != -1){

	  /* copy of temp[left] */
	  copy_sequence(test, &(temp[left]));
	  free_RNA_sequence_m(&temp[left]);

	  score = optimal_multiple_arc_alignment_mapping( test, &(temp[right]), &(temp[left]), EQ, SIMILAR );

	  free_RNA_sequence_m(test);
	}
      }
    }
  }
  
  i = 0;
  while( temp[i++].nb_seq < nb_seq );

  *sc = score;

  copy_sequence(test, &(temp[i-1])); 

  /* free memory */
  for(i=0;i<nb_seq;i++)
    for(j=0;j<nb_seq;j++)
      if(cluster_list[i][j].nbre_seq >= 1)
	free(cluster_list[i][j].seq);

  for(i=0;i<nb_seq;i++)
    free(cluster_list[i]);

  free(cluster_list);

  for(i=0;i<nb_seq;i++){
    if(temp[i].s != NULL)
      free_RNA_sequence_m(&temp[i]);
  }

  free(temp);

  /* return the alignment */
  return test;
}
