/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or */
/* (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../headers/constraint.h"
#include "../headers/matrix.h"



/*initialization of the matrix*/
void no_constraint(char ** cons,int m, int n){
	int i,j;
	for(i=0;i<m;i++)
	  for(j=0;j<n;j++)
	    cons[i][j] = ALLOWED;
}


void full_constraint(char **cons, int m, int n){
  int i, j;
  for(i=0; i<m; i++)
    for(j=0; j<n; j++)
      cons[i][j] = FORBIDDEN;
}


void display_constraint(char ** cons, int m, int n){
	int i,j;
	for(i=0;i<m;i++)
	  for(j=0;j<n;j++)
	    if (cons[i][j] == ALLOWED) printf ("no constraint: %d, %d\n",i,j);
}


void diagonal_constraint(char ** cons, int m, int n, int d){
  int i,j;
  for(i=0;i<m;i++)
    for(j=0;j<n;j++)
      if ((i-j<d) && (j-i<d)){ 
	if(cons[i][j] != FORBIDDEN)
	  cons[i][j] = ALLOWED; 
      }
      else 
	cons[i][j]=FORBIDDEN;

}



/*ajoute une contrainte*/
void add_constraint(char ** cons,postorder_tree * treeA,postorder_tree * treeB,int x,int y){
	int i,j;
	int h,l;
	h=x-treeA->n[x].size;
	l=y-treeB->n[y].size;

	if((treeA->n[x].size==1 && treeB->n[y].size!=1) || (treeA->n[x].size!=1 && treeB->n[y].size==1))return;

	/*	printf ("contrainte : %d, %d\n", x, y);*/

	for(i=0;i<h; i++)
	  for (j=l+2; j<treeB->size; j++)
	      cons[i][j] = (char)1;
	for (i=h+2; i<x; i++){
	  for (j=0; j<l; j++)
	    cons[i][j] = (char)1;
	  for (j=y+2; j<treeB->size; j++)
	    cons[i][j] = (char)1;
	}
	for (i=x+2; i<treeA->size; i++)
	  for (j=0; j<y; j++)
	     cons[i][j] = (char)1;
}

/*trouve l'indice du noeud dans l'arbre correspondant a l'indice de la base dans la squence*/
int sequence_position_to_postorder_position(postorder_tree * tree,int index){
	int i;
	for(i=1;i<tree->size;i++)
		if(tree->n[i].positionX==index)
			return i;
		else if(tree->n[i].positionY==index)
			return i;
	return -1;
}

int sequence_position_to_postorder_position_m(postorder_tree_m * tree,int index){
  int i;
  for(i=1;i<tree->size;i++)
    if(tree->n[i].positionX==index)
      return i;
    else if(tree->n[i].positionY==index)
      return i;
  return -1;
}

/*extract the sequence from an RNA_sequence object */
void extract_plain_sequence(RNA_sequence *arn,sequence_list_t seq){
	int i;

	seq->length=(arn->size)-1;
	seq->nucl=malloc(sizeof(char)*(seq->length)+1);
	for(i=1;i<arn->size;i++){
		seq->nucl[i-1]=arn->p[i].nucleotide;
	}
	seq->nucl[arn->size-1] = '\0';
} 



/*ajoute les contraintes venant des blocs des squences alignes*/
void put_constraint(char ** cons,block_t *block,postorder_tree * treeA,postorder_tree * treeB){
  int i,j,positionA, positionB;
  int size=(block[0]->length)+1;
  char ** verif_cons=NULL;
  
  verif_cons = new_matrix_char(treeA,treeB);
  
  no_constraint(verif_cons,treeA->size,treeB->size);
  for(i=1;i<size;i++){
       for(j=0;j<block[i]->length;j++){
      positionA= sequence_position_to_postorder_position(treeA,(block[i]->begin[0]+j)+1);
      positionB= sequence_position_to_postorder_position(treeB,(block[i]->begin[1]+j)+1);
      if(positionA!=-1 && positionB!=-1 && verif_cons[positionA][positionB]== (char)0){ 
	verif_cons[positionA][positionB]=(char )1;
	add_constraint(cons,treeA,treeB,positionA,positionB);	
      }
    }
  }

  free_matrix_char(verif_cons, treeA);
  /*free(verif_cons); */

}

/* prend deux squences, les aligne, recupre les blocs aligns et les ajoute dans la matrice des contraintes */
void primary_sequence_constraint(char ** cons, postorder_tree * treeA, postorder_tree * treeB, RNA_sequence *rnaA,RNA_sequence *rnaB){
  

  sequence_list_t seq_rnaA = NULL;
  sequence_list_t seq_rnaB = NULL;
  block_list_t list_seq_arn_A_B = NULL;
  
  seq_rnaA = (sequence_list_t)malloc(sizeof(sequence_t));
  seq_rnaB = (sequence_list_t)malloc(sizeof(sequence_t));
  seq_rnaA->name = NULL;
  seq_rnaA->nucl = NULL;
  seq_rnaB->name = NULL;
  seq_rnaB->nucl = NULL;

  extract_plain_sequence(rnaA,seq_rnaA);
  extract_plain_sequence(rnaB,seq_rnaB);
 
  list_seq_arn_A_B=find_blocks(seq_rnaA, seq_rnaB);
 
  /*display_block_t(list_seq_arn_A_B,seq_rnaA_rnaB);*/
 
  put_constraint(cons,list_seq_arn_A_B,treeA,treeB);	
  
  /* ajout Antoine ?*/
  seq_rnaA->nucl = mon_free(seq_rnaA->nucl);
  seq_rnaA->name = mon_free(seq_rnaA->name);
  seq_rnaA = mon_free(seq_rnaA);

  seq_rnaB->nucl = mon_free(seq_rnaB->nucl);
  seq_rnaB->name = mon_free(seq_rnaB->name);
  seq_rnaB = mon_free(seq_rnaB);
  /* fin ajout */

  free_block_list (list_seq_arn_A_B);
}


void extract_eq_blocks_m(char *eqA, square_t **tab, int *size, char *eqB, 
			 int eqA_size, int eqB_size, postorder_tree_m *treeA){

  int i=0, j, temp_size, k, tree_pos;
  square_t *temp_tab=NULL;
  char *tab_eq_temp, *already;

  already = (char *)calloc( treeA->size+1, sizeof(char) );

  tab_eq_temp = (char *)calloc( 93, sizeof(char) );
  for(j=0; j<eqB_size-1; j++)
    if(eqB[j] != ' ')
      tab_eq_temp[(int)(eqB[j]) - 33] = 1;

  /* get the common constraints */
  temp_size = 1;
  while( i < eqA_size-1 ){
    tree_pos = sequence_position_to_postorder_position_m(treeA, i+1);
    if( eqA[i] != ' ' && tab_eq_temp[(int)(eqA[i]) - 33] && already[tree_pos] == 0){
      temp_tab = (square_t *)realloc(temp_tab, sizeof(square_t) * (temp_size) );
      temp_tab[temp_size-1].num = eqA[i];
      temp_tab[temp_size-1].begin = i+1;
      already[tree_pos] = 1;
      j = eqA[i];
      while( i<eqA_size-1 && j == eqA[i] ){
	i++;
	already[sequence_position_to_postorder_position_m(treeA, i)] = 1;
      }
      temp_tab[temp_size-1].end = i;
      temp_size++;
      i--;
    }
    i++;
  }
 
/*   printf("\ntabA %i\n", temp_size); */
/*   for(i=0; i<temp_size-1; i++) */
/*     printf("%c: %i\t%i\n", temp_tab[i].num, temp_tab[i].begin, temp_tab[i].end); */
 
  if(temp_size>1){
    *tab = (square_t *)malloc( sizeof(square_t) );
    (*tab)[0].num = temp_tab[0].num;
    (*tab)[0].begin = temp_tab[0].begin;
    (*tab)[0].end = temp_tab[0].end;
  
    /* merge constraints with the same number */
    *size = 1;
    k = 1;
    for(i=1; i<temp_size-1; i++){
      if( temp_tab[i-1].num == temp_tab[i].num )
	(*tab)[k-1].end = temp_tab[i].end;
      else{
	(*size)++;
	*tab = (square_t *)realloc(*tab, sizeof(square_t) * (i+1) );
	(*tab)[k].num = temp_tab[i].num;
	(*tab)[k].begin = temp_tab[i].begin;
	(*tab)[k].end = temp_tab[i].end;
	k++;
      }
    }
  }
  else{
    *tab = (square_t *)malloc( sizeof(square_t) );
    *size = 0;
  }
  
  free(already);
  free(temp_tab);
  free(tab_eq_temp);
}


void extract_eq_blocks(char *eqA, square_t **tab, int *size, char *eqB, int eqA_size, int eqB_size, postorder_tree *treeA){

  int i=0, j, temp_size, k, tree_pos;
  square_t *temp_tab=NULL;
  char *tab_eq_temp, *already;

  already = (char *)calloc( treeA->size+1, sizeof(char) );

  tab_eq_temp = (char *)calloc( 93, sizeof(char) );
  for(j=0; j<eqB_size-1; j++)
    if(eqB[j] != ' ')
      tab_eq_temp[(int)(eqB[j]) - 33] = 1;

  /* get the common constraints */
  temp_size = 1;
  while( i < eqA_size-1 ){
    tree_pos = sequence_position_to_postorder_position(treeA, i+1);
    if( eqA[i] != ' ' && tab_eq_temp[(int)(eqA[i]) - 33] && already[tree_pos] == 0){
      temp_tab = (square_t *)realloc(temp_tab, sizeof(square_t) * (temp_size) );
      temp_tab[temp_size-1].num = eqA[i];
      temp_tab[temp_size-1].begin = i+1;
      j = eqA[i];
      already[tree_pos] = 1;
      while( i<eqA_size-1 && j == eqA[i]){
	i++;
	already[sequence_position_to_postorder_position(treeA, i)] = 1;
      }
      temp_tab[temp_size-1].end = i;
      temp_size++;
      i--;
    }
    i++;
  }

  *tab = (square_t *)malloc( sizeof(square_t) );
  (*tab)[0].num = temp_tab[0].num;
  (*tab)[0].begin = temp_tab[0].begin;
  (*tab)[0].end = temp_tab[0].end;

  if(temp_size > 1){
    /* merge constraints with the same number */
    *size = 1;
    k = 1;
    for(i=1; i<temp_size-1; i++){
      if( temp_tab[i-1].num == temp_tab[i].num )
	(*tab)[k-1].end = temp_tab[i].end;
      else{
	(*size)++;
	*tab = (square_t *)realloc(*tab, sizeof(square_t) * (i+1) );
	(*tab)[k].num = temp_tab[i].num;
	(*tab)[k].begin = temp_tab[i].begin;
	(*tab)[k].end = temp_tab[i].end;
	k++;
      }
    }
  }
  else{
    *tab = (square_t *)malloc( sizeof(square_t) );
    *size = 0;
  }

  free(already);
  free(temp_tab);
  free(tab_eq_temp);
}


void swap_values(int *a, int *b){

  int c;

  c = *a;
  *a = *b;
  *b = c;
}

void eq_constraint(char **cons, postorder_tree_m *treeA, postorder_tree_m *treeB,
		   char *eqA, char *eqB, int eqA_size, int eqB_size, int multiple){

  int i=0, j=0, k;
  int sizeA=1, sizeB=1;
  square_t *tabA=NULL, *tabB=NULL;
  int firstA, lastA, firstB, lastB;
  
  if(multiple){
    extract_eq_blocks_m(eqA, &tabA, &sizeA, eqB, eqA_size, eqB_size, treeA);
    extract_eq_blocks_m(eqB, &tabB, &sizeB, eqA, eqB_size, eqA_size, treeB);
  }
  else{
    extract_eq_blocks(eqA, &tabA, &sizeA, eqB, eqA_size, eqB_size, (postorder_tree *)treeA);
    extract_eq_blocks(eqB, &tabB, &sizeB, eqA, eqB_size, eqA_size, (postorder_tree *)treeB);
  }

  /* sizeA--; */
  /*   sizeB--; */

  /* first constraints */
  if(sizeA>1){
    if(multiple){
      lastA  = sequence_position_to_postorder_position_m(treeA, tabA[1].begin);
      lastB  = sequence_position_to_postorder_position_m(treeB, tabB[1].begin);
    }
    else{
      lastA  = sequence_position_to_postorder_position((postorder_tree *)treeA, tabA[1].begin);
      lastB  = sequence_position_to_postorder_position((postorder_tree *)treeB, tabB[1].begin);
    }
  }
  else{
    lastA = treeA->size-1;
    lastB = treeB->size-1;
  }

  for(j=0; j<=lastA; j++)
    for(k=0; k<=lastB; k++)
      cons[j][k] = ALLOWED;

  for(i=1; i<sizeA-1; i++){
    if(multiple){
      firstA = sequence_position_to_postorder_position_m(treeA, tabA[i-1].end);
      lastA  = sequence_position_to_postorder_position_m(treeA, tabA[i+1].begin);
      firstB = sequence_position_to_postorder_position_m(treeB, tabB[i-1].end);
      lastB  = sequence_position_to_postorder_position_m(treeB, tabB[i+1].begin);
    }
    else{
      firstA = sequence_position_to_postorder_position((postorder_tree *)treeA, tabA[i-1].end);
      lastA  = sequence_position_to_postorder_position((postorder_tree *)treeA, tabA[i+1].begin);
      firstB = sequence_position_to_postorder_position((postorder_tree *)treeB, tabB[i-1].end);
      lastB  = sequence_position_to_postorder_position((postorder_tree *)treeB, tabB[i+1].begin);
    }
    if(firstA > lastA)
      swap_values(&firstA, &lastA);
    if(firstB > lastB)
      swap_values(&firstB, &lastB);

    for(j=firstA; j<=lastA; j++)
      for(k=firstB; k<=lastB; k++)
	cons[j][k] = ALLOWED;
  }


  /* last constraints */
  if(sizeA>1){
    if(multiple){
      firstA = sequence_position_to_postorder_position_m(treeA, tabA[sizeA-2].end);
      lastA  = sequence_position_to_postorder_position_m(treeA, eqA_size-1);
      firstB = sequence_position_to_postorder_position_m(treeB, tabB[sizeB-2].end);
      lastB  = sequence_position_to_postorder_position_m(treeB, eqB_size-1);
    }
    else{
      firstA = sequence_position_to_postorder_position((postorder_tree *)treeA, tabA[sizeA-2].end);
      lastA  = sequence_position_to_postorder_position((postorder_tree *)treeA, eqA_size-1);
      firstB = sequence_position_to_postorder_position((postorder_tree *)treeB, tabB[sizeB-2].end);
      lastB  = sequence_position_to_postorder_position((postorder_tree *)treeB, eqB_size-1);
    }
    if(firstA > lastA)
      swap_values(&firstA, &lastA);
    if(firstB > lastB)
      swap_values(&firstB, &lastB);
    for(j=firstA; j<=lastA; j++)
      for(k=firstB; k<=lastB; k++)
	cons[j][k] = ALLOWED;
  }

  /* free memory */
  free(tabA);
  free(tabB);
}


/*affiche l'alignement d'une liste de blocs*/
void display_block_t(block_t *block,sequence_list_t* seq){
  int i,j;
  int size=(block[0]->length);
  printf("\n");
  for(i=1;i<=size;i++){
    printf("%d  ",block[i]->begin[0]+1);
    for(j=0;j<(block[i]->length);j++)
      printf("%c",seq[0]->nucl[j+block[i]->begin[0]]);
    printf("  %d\n%d  ",block[i]->begin[0]+block[i]->length+1,block[i]->begin[1]+1);
    for(j=0;j<(block[i]->length);j++)
      printf("%c",seq[1]->nucl[j+block[i]->begin[1]]);
    printf("  %d\n\n",block[i]->begin[1]+block[i]->length+1);
  }
}

