/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or */
/* (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */


void create_alignment_RNA_sequences_m(RNA_sequence_m *rnaA, RNA_sequence_m *rnaB, stack *alignment_stack,
				      RNA_sequence_m *alignment, int startA, int endA, int startB,int endB){

  int i, j, x, y, alignment_length = 0;
  char *matchA, *matchB; /* 1 for matching positions, O otherwise */
  int last_bracket=0;  /* 0 if the last '(' was in A, 1 if it was in B */
  int *q;
  char *p;

  alignment_length = endA + endB - startA - startB + 2; 
  
  matchA = (char *)malloc( sizeof(char) * (rnaA->s[0].size + 1) );
  matchB = (char *)malloc( sizeof(char) * (rnaB->s[0].size + 1) );
  for(i=0; i<rnaA->s[0].size; i++)
    matchA[i] = 0;
  for(i=0; i<rnaB->s[0].size; i++)
    matchB[i] = 0;

  matchA[rnaA->s[0].size] = 2;
  matchB[rnaB->s[0].size] = 2;

  while(!is_empty_stack(alignment_stack)){
    pop(alignment_stack, &x, &y);
    matchA[x] = 1;
    matchB[y] = 1;
    alignment_length--;
  }

  /* memory allocation for the alignment */
  alignment->s = (RNA_sequence *)malloc( sizeof(RNA_sequence) * (rnaA->nb_seq + rnaB->nb_seq) ); 
  alignment->nb_seq = rnaA->nb_seq + rnaB->nb_seq;
  alignment->eq = (char *)malloc( sizeof(char) * alignment_length );
  alignment->super_pairing = (int *)malloc( sizeof(int) * (alignment_length+1) );
  alignment->super_partner = (int *)malloc( sizeof(int) * (alignment_length+1) );
  for(i=0; i<rnaA->nb_seq + rnaB->nb_seq; i++){
    alignment->s[i].p = (position *)malloc( sizeof(position) * (alignment_length+1) );
    alignment->s[i].size = alignment_length + 1;
  }

  x=startA; 
  y=startB;
  for(i=1; i<=alignment_length; i++){

    if(matchA[x] == 1 && matchB[y] == 1){ 
      /* match or mismatch */
      alignment->eq[i-1] = rnaA->eq[x-1] == 0 ? rnaB->eq[y-1] : rnaA->eq[x-1];
      alignment->super_pairing[i] = rnaA->super_pairing[x] != 0 ? rnaA->super_pairing[x] : rnaB->super_pairing[y];
      fill_alignment(alignment, rnaA, rnaB, i, x, y, 0);
      if (rnaA->super_pairing[x]==-1) last_bracket=0;
      if (rnaB->super_pairing[y]==-1) last_bracket=1;
      x++;y++;
    }else if(matchA[x]==0 && matchB[y]==0 && rnaA->super_pairing[x] == 1 && rnaB->super_pairing[y] == 1 ){
      /* base deletion */
      if(last_bracket == 0){
	alignment->eq[i-1] = rnaA->eq[x-1];
	alignment->super_pairing[i] = rnaA->super_pairing[x];
	fill_alignment(alignment, rnaA, rnaB, i, x, y, 1);
	x++;
      }
      else{
	/* base insertion */
	alignment->eq[i-1] = rnaB->eq[y-1];
	alignment->super_pairing[i] = rnaB->super_pairing[y];	
	fill_alignment(alignment, rnaA, rnaB, i, x, y, -1);
	y++;
      }
    }else if(matchA[x]==0 && matchB[y]==0 && ( (rnaA->super_pairing[x] == 1 && rnaB->super_pairing[y] == -1) ||
					      (rnaA->super_pairing[x] == -1 && rnaB->super_pairing[y] == 1)) ){
      
      if(rnaA->super_pairing[x] == 1){
	/* base deletion */
	alignment->eq[i-1] = rnaA->eq[x-1];
	alignment->super_pairing[i] = rnaA->super_pairing[x];
	fill_alignment(alignment, rnaA, rnaB, i, x, y, 1);
	x++;
      }else{
	/* base insertion */
	alignment->eq[i-1] = rnaB->eq[y-1];
	alignment->super_pairing[i] = rnaB->super_pairing[y];
	fill_alignment(alignment, rnaA, rnaB, i, x, y, -1);
	y++;
      }
    }else if( matchA[x]==0 ){ 
      /* base deletion */
      alignment->eq[i-1] = rnaA->eq[x-1];
      alignment->super_pairing[i] = rnaA->super_pairing[x];
      fill_alignment(alignment, rnaA, rnaB, i, x, y, 1);
      if (rnaA->super_pairing[x]==-1) last_bracket=0;
      x++;
    }else if( matchB[y] == 0 ){ 
      /* base insertion */
      alignment->eq[i-1] = rnaB->eq[y-1];
      alignment->super_pairing[i] = rnaB->super_pairing[y];
      fill_alignment(alignment, rnaA, rnaB, i, x, y, -1);
      if (rnaB->super_pairing[y]==-1) last_bracket=1;
      y++;
    }else{
      printf("\n Not a valid alignment !!! ");
      exit(1);
    }
  }

  /* copy sequences names */
  for(j=0; j<rnaA->nb_seq; j++){
    alignment->s[j].name = (char *)malloc( sizeof(char) * (strlen(rnaA->s[j].name) + 1) );
    strcpy(alignment->s[j].name, rnaA->s[j].name);
  }
  for(j=0; j<rnaB->nb_seq; j++){
    alignment->s[j + rnaA->nb_seq].name = (char *)malloc( sizeof(char) * (strlen(rnaB->s[j].name) + 1) );
    strcpy(alignment->s[j + rnaA->nb_seq].name, rnaB->s[j].name);
  }

  /*   partner renumerating */
  for(i=0; i<alignment->nb_seq; i++){
    q = (int *)malloc( sizeof(int) * alignment->s[i].size );
    p = (char *)malloc( sizeof(char) * alignment->s[i].size );
    for(j=1; j<alignment->s[i].size; j++)
      p[j] = int_2_char(alignment->s[i].p[j].pairing);

    retrace_partner(1, p, q, alignment->s[i].size, NULL);

    for(j=1; j<alignment->s[i].size; j++)
      alignment->s[i].p[j].partner = q[j];
    
    free(q);
    free(p);
  }

  q = (int *)malloc( sizeof(int) * alignment->s[0].size );
  p = (char *)malloc( sizeof(char) * alignment->s[0].size );
  for(j=1; j<alignment->s[0].size; j++)
    p[j] = int_2_char(alignment->super_pairing[j]);

  retrace_partner(1, p, q, alignment->s[0].size, NULL);

  for(j=1; j<alignment->s[0].size; j++)
    alignment->super_partner[j] = q[j];

  /* free memory */
  free(q);
  free(p);  
  free(matchA);
  free(matchB);
}
