/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or */
/* (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "../headers/score.h"
#include "../headers/constraint.h"
#include "../headers/stack.h"
#include "../headers/RNA_structure.h"
#include "../headers/matrix.h"

static int ** table = NULL;
static char ** constraint=NULL; 


static int get_table(int x, int y){
  if (constraint[x][y]==ALLOWED)
    return table[x][y];
  else
    return -1000000;
}


/* compute the matrix TD : Tree Distance 
   sub : bolean matrix that indicates whether the optimal score is achieved with a substitution operation 
   constraint : constraint matrix */
void edit_distance(postorder_tree * treeA,postorder_tree * treeB,int ** TD,int ** table,int ** sub,char ** constraint){
  int i,j, k, l, x, y;
  int calA=0; 
  int calB=0; 
  int calC=0;
  
  int w; 
  int * ancestorA;
  int * ancestorB;
  
  ancestorA=(int*)malloc(sizeof(int)*treeA->size);
  ancestorB=(int*)malloc(sizeof(int)*treeB->size); 
  ancestor(treeA,ancestorA);
  ancestor(treeB,ancestorB);
  
  for(i=treeA->size-1;i>=1; i--){
    if (treeA->n[i].size==1) {
      for (j=treeB->size-1; j>=1; j--){
	if  (treeB->n[j].size==1) {
	  
	  if(constraint[i][j]==ALLOWED){
	    
	    table[i-1][j-1]=0;
	    k=ancestorA[i]; 
	    l=ancestorB[j];
	    
	    for (x=i; x<=k; x++)
	      if (treeA->n[x].size==1)
		table[x][j-1]=table[x-1][j-1]+base_deletion(treeA->n[x].labelX);
	      else
		table[x][j-1]=table[x-1][j-1]+arc_removing(treeA->n[x].labelX, treeA->n[x].labelY);
	    for (y=j; y<=l; y++)
	      if (treeB->n[y].size==1)
		table[i-1][y]=table[i-1][y-1]+base_deletion(treeB->n[y].labelX);
	      else
		table[i-1][y]=table[i-1][y-1]+arc_removing(treeB->n[y].labelX, treeB->n[y].labelY); /* a remplacer par full_deletion */
	    
	    for (x=i; x<=k; x++){
	      for (y=j; y<=l; y++)
		if  (constraint[x][y]==ALLOWED){
		  if (treeA->n[x].size==1) 
		    calA=get_table(x-1,y)+base_deletion(treeA->n[x].labelX); /*deletion */
		  else 
		    calA=get_table(x-1,y)+arc_removing(treeA->n[x].labelX, treeA->n[x].labelY); /*deletion */
		  if (treeB->n[y].size==1)
		    calB=get_table(x,y-1)+base_deletion(treeB->n[y].labelX); /*insertion*/
		  else 
		    calB=get_table(x,y-1)+arc_removing(treeB->n[y].labelX, treeB->n[y].labelY); /*insertion*/
		  
		  table[x][y]=MAX(calA,calB);
		  
		  if ((treeA->n[x].size==1)&&(treeB->n[y].size==1)){ 	     
		    TD[x][y]=base_mismatch(treeA->n[x].labelX,treeB->n[y].labelX);
		    calC=get_table(x-1,y-1)+TD[x][y]; /*substitution*/
		  }
		  else if  ((treeA->n[x].size>1)&&(treeB->n[y].size>1) && !(is_subtree(treeA,i,x) && is_subtree(treeB,j,y))){ 
		  if (sub[x][y]) 
		    calC=get_table(x-treeA->n[x].size,y-treeB->n[y].size)+TD[x][y];
		  else
		    calC=INT_MIN;
		}
		else calC=INT_MIN;
		
		table[x][y]=MAX(get_table(x,y),calC);
		
		if (is_subtree(treeA,i,x) && is_subtree(treeB,j,y)){
		  w=get_table(x-1,y-1)+arc_mismatch(treeA->n[x].labelX, treeA->n[x].labelY,treeB->n[y].labelX, treeB->n[y].labelY);
		  
		  if (w>get_table(x,y)){
		    table[x][y]=w;
		    TD[x][y]=w;
		    sub[x][y]=1;
		  }
		  else{
		    TD[x][y]=INT_MIN;  
		  }
		}
	      }/* end for y*/
	    }/* end for x*/
	  }
	}
      }
    }
  }
}

/* compute the score of the optimal alignment using the TD matrix */
int score_TD(postorder_tree * treeA,postorder_tree * treeB,int ** TD, int ** sub,char ** constraint){
  int x,y;
  int ** g=NULL;
  int score; 

  g = new_matrix(treeA,treeB);
  g[0][0]=0;
  for(x=1;x<treeA->size;x++)g[x][0]=g[x-1][0]+indel(treeA,x); 
  for(y=1;y<treeB->size;y++)g[0][y]=g[0][y-1]+indel(treeB,y); 
  for(x=1;x<treeA->size;x++){
    for(y=1;y<treeB->size;y++){
      if (constraint[x][y]==ALLOWED){
      if  (sub[x][y]==1)  /* substitution */
      	g[x][y]=max3(g[x-1][y]+indel(treeA,x),g[x][y-1]+indel(treeB,y),g[x-treeA->n[x].size][y-treeB->n[y].size]+TD[x][y]);
      else 	
	 g[x][y]=max2(g[x-1][y]+indel(treeA,x),g[x][y-1]+indel(treeB,y));
      }
    }
  }
  score = g[treeA->size-1][treeB->size-1];
  free_matrix(g, treeA);
  return score; 
}

/* optimal path in the edit graph g delineated by the
   rectangle  i,j,k,l  
   peut-on avancer le calcul de edit_operation a edit_distance.c ?*/
void optimal_path(postorder_tree * treeA,postorder_tree * treeB,RNA_sequence* rnaA, RNA_sequence* rnaB, int ** TD,int ** g,char ** edit_operation,int ** sub,int i,int j,int k,int l, stack * p, int ** supertree, char** constraint, int mini,int minj){
  int x,y;
  

  printf("i,j,k,l=%d,%d,%d,%d\n", i,j,k,l);

  g[i][j]=0;
  for(x=i+1;x<=k;x++){g[x][j]=g[x-1][j]+indel(treeA,x);edit_operation[x][j]=1;}
  for(y=j+1;y<=l;y++){g[i][y]=g[i][y-1]+indel(treeB,y);edit_operation[i][y]=2;}
  for(x=i+1;x<=k;x++){
    for(y=j+1;y<=l;y++){
      if (sub[x][y]==1){  /* substitution */
	g[x][y]=max3(g[x-1][y]+indel(treeA,x),g[x][y-1]+indel(treeB,y),g[x-treeA->n[x].size][y-treeB->n[y].size]+TD[x][y]);
	if(g[x][y]==g[x-1][y]+indel(treeA,x))edit_operation[x][y]=1;
	if(g[x][y]==g[x][y-1]+indel(treeB,y))edit_operation[x][y]=2;
	if(g[x][y]==g[x-treeA->n[x].size][y-treeB->n[y].size]+TD[x][y]){ edit_operation[x][y]=3;}
      }else{ /* no substitution */
	g[x][y]=max2(g[x-1][y]+indel(treeA,x),g[x][y-1]+indel(treeB,y)); 
	if(g[x][y]==g[x-1][y]+indel(treeA,x))edit_operation[x][y]=1;
	if(g[x][y]==g[x][y-1]+indel(treeB,y))edit_operation[x][y]=2;
      }
    }
  }


  x=k;y=l;
 
  while(x+y>i+j){
    
    switch (edit_operation[x][y]) {
    case 0 : /* fin */
      x=-1;y=-1; 
      break;
    case 1 : /* deletion */
      
      if (y==0){
        if  (treeA->n[x].size>1){
	  if (treeA->n[x].labelX!='X'){
	    supertree[treeA->n[x].positionY][minj]=-1;
	    supertree[treeA->n[x].positionX][minj]=-1;
	  }
	}
	else
	  supertree[treeA->n[x].positionX][minj]=-1;
      }      
      if (y==j){
        if  (treeA->n[x].size>1){
	  if (treeA->n[x].labelX!='X'){
	    supertree[treeA->n[x].positionX][minj]=-1;
	    supertree[treeA->n[x].positionY][minj]=-1;
	  }
	}
	else
	   supertree[treeA->n[x].positionX][minj]=-1;
      }   

      if ((treeA->n[x].size==1) && (treeB->n[y].size==1)){
	supertree[treeA->n[x].positionX][MAX(minj,treeB->n[y].positionX)]=-1;
      }
      if ((treeA->n[x].size==1) &&  (treeB->n[y].size>1)){
	supertree[treeA->n[x].positionX][MAX(minj,treeB->n[y].positionY)]=-1;
      }
      if ((treeA->n[x].size>1) &&  (treeB->n[y].size==1)){
	 if (treeA->n[x].labelX!='X') supertree[treeA->n[x].positionY][MAX(minj,treeB->n[y].positionX)]=-1;
	for (k=minj; k<treeB->n[y].positionX; k++)
	   if (treeA->n[x].labelX!='X') supertree[treeA->n[x].positionX][k]=-1;
      }
      if ((treeA->n[x].size>1) &&  (treeB->n[y].size>1)){
	if (treeA->n[x].labelX!='X'){
	supertree[treeA->n[x].positionY][treeB->n[y].positionY]=-1;
	for (k=minj; k<treeB->n[y].positionY; k++)
	  supertree[treeA->n[x].positionX][k]=-1;
	}
      }
      x=x-1;
      break;
    case 2 : /* insertion */
      printf("insertion : %d, %d\n", treeA->n[x].size, treeB->n[y].size);
      
      if (x==0){
        if  (treeB->n[y].size>1){
	  if (treeB->n[y].labelX!='X'){
	    supertree[mini][treeB->n[y].positionY]=-2;
	    supertree[mini][treeB->n[y].positionX]=-2;
	  }
	}
	else
	  supertree[mini][treeB->n[y].positionX]=-2;
      }      
      if (x==i){
        if  (treeB->n[y].size>1){
	  if (treeB->n[y].labelX!='X'){
	    supertree[mini][treeB->n[y].positionY]=-2;
	    supertree[mini][treeB->n[y].positionX]=-2;
	  }
	}
	else
	  supertree[mini][treeB->n[y].positionX]=-2;
      }      
      if ((treeA->n[x].size==1) && (treeB->n[y].size==1)){ 
	printf("insertion1\n");
	supertree[MAX(mini, treeA->n[x].positionX)][treeB->n[y].positionX]=-2;
      }
      if ((treeA->n[x].size==1) &&  (treeB->n[y].size>1)){
	if (treeB->n[y].labelX!='X'){
	  supertree[treeA->n[x].positionX][treeB->n[y].positionY]=-2;
	  for (k=mini; k<treeA->n[x].positionX; k++)
	    supertree[k][treeB->n[y].positionX]=-2;
	}
      }
      if ((treeA->n[x].size>1) &&  (treeB->n[y].size==1)){
	supertree[MAX(mini,treeA->n[x].positionY)][treeB->n[y].positionX]=-2;
      }
      if ((treeA->n[x].size>1) &&  (treeB->n[y].size>1)){
	if (treeB->n[y].labelX!='X'){
	  supertree[MAX(mini,treeA->n[x].positionY)][treeB->n[y].positionY]=-2;
	  for (k=mini; k<treeA->n[x].positionY; k++)
	    supertree[k][treeB->n[y].positionX]=-2;
	}
      }
      y=y-1;
      break;
    case 3 : /*(mis)match */ /*a modifier pour le full edit */
        if (treeA->n[x].labelX!='X') {
	supertree[treeA->n[x].positionX][treeB->n[y].positionX]=1;
	if (treeA->n[x].size>1){
	  supertree[treeA->n[x].positionY][treeB->n[y].positionY]=1;
	  push(p,x,y);
	}
      }
      else {
	 push(p,x,y);
      }
      x=x-treeA->n[x].size;
      y=y-treeB->n[y].size;
      break;      
    }
  }
  
}

/* Construction of the optimal mapping: 
   the mapping is a set of pairs of nodes of treeA x treeB. 
   It is stored in a stack. */
int trace_back_mapping(postorder_tree * treeA,postorder_tree * treeB,RNA_sequence* rnaA, RNA_sequence* rnaB, int ** TD,int ** sub, int** supertree, char ** constraint){
  int x,y;
  int score; 
  char ** edit_operation=new_matrix_char(treeA,treeB);
  int ** g=new_matrix(treeA,treeB);
  stack *  p;
  p= malloc(sizeof(stack));
  /* p->elements = malloc(sizeof(pair)*treeA->size+treeB->size);*/
  
  
  init_stack(p,treeA->size+treeB->size);
  optimal_path(treeA,treeB,rnaA, rnaB,TD,g,edit_operation,sub,0,0,treeA->size-1,treeB->size-1,p,supertree,constraint,0,0);
  while(!is_empty_stack(p)){
    pop(p,&x,&y);
    optimal_path(treeA,treeB,rnaA, rnaB, TD,g,edit_operation,sub,x-treeA->n[x].size,y-treeB->n[y].size,x-1,y-1,p,supertree,constraint, treeA->n[x].positionX, treeB->n[y].positionX);
  }
    score=g[treeA->size-1][treeB->size-1]; 
    free_matrix_char(edit_operation, treeA);
    free_matrix(g, treeA);
    p->elements = mon_free(p->elements);
    p = mon_free(p);
    return score; 
  }

int optimal_edit(RNA_sequence* rnaA, RNA_sequence* rnaB, int primary, int similar,  char* nameCtA, char * nameCtB, char * nameFileOut, RNA_sequence* alignA, RNA_sequence* alignB, int full_edit, int eq, char *eqA, char *eqB, int do_mapping){
  int ** TD = NULL;
  int ** sub = NULL;
  postorder_tree * treeA;
  postorder_tree * treeB;
  int sc=0;
int ** supertree; 
  int i,j;	
  
  treeA=( postorder_tree * ) malloc(sizeof(postorder_tree));
  treeB=( postorder_tree * ) malloc(sizeof(postorder_tree));
  
  if (full_edit==1) {
    RNAsequence2triplet_postorder(rnaA,treeA);
    RNAsequence2triplet_postorder(rnaB,treeB);
  }
  else{ 
    RNA_sequence2postorder(rnaA,treeA);
    RNA_sequence2postorder(rnaB,treeB);
  }
  
constraint=new_matrix_char(treeA,treeB);	
no_constraint(constraint,treeA->size,treeB->size);	
if (similar==1)
    diagonal_constraint(constraint,treeA->size,treeB->size, treeA->size/5);		
  if(primary==1)
    primary_sequence_constraint(constraint,treeA,treeB,rnaA,rnaB);

  TD=new_matrix(treeA,treeB);
  table=new_matrix(treeA,treeB);
  sub=new_matrix(treeA,treeB);
  for (i=1; i<treeA->size; i++)
    for (j=1; j<treeB->size; j++)
      sub[i][j]=(treeA->n[i].size==1) && (treeB->n[j].size==1);
  edit_distance(treeA,treeB,TD,table,sub,constraint);
  
  if (do_mapping==0){
    sc=score_TD(treeA, treeB, TD, sub, constraint);
  }
  else {
    supertree = (int **)malloc( sizeof(int*)*(rnaA->size+1) );
    for (i=0; i<=rnaA->size; i++)
      supertree[i]= (int *)malloc( sizeof(int)*(rnaB->size+1) );
    for (i=0; i<=rnaA->size; i++)
      for (j=0; j<=rnaB->size; j++)
	supertree[i][j]=0;
    sc=trace_back_mapping(treeA,treeB,rnaA,rnaB,TD,sub,supertree, constraint); 
    for (i=0; i<rnaA->size; i++){
      for (j=0; j<rnaB->size; j++)
	printf("%3d ",supertree[i][j]); 
      printf("\n");  
    } 
    printf("fin de l'affichage du supertree\n");
      create_alignment_RNA_sequences(rnaA,rnaB, supertree, alignA, alignB, 1, rnaA->size-1, 1, rnaB->size-1);
      printf("fin de creation de l'alignement\n");
   
    
  }
  free_matrix(TD, treeA);
  free_matrix(table, treeA);
  free_matrix(sub, treeA);
  free_matrix_char(constraint, treeA);
  treeA->n = mon_free(treeA->n);
  treeB->n = mon_free(treeB->n);
  treeA = mon_free(treeA);
  treeB = mon_free(treeB);	
  return sc;
}


/* computes the score of the optimal alignment of two trees */
/* without building the optimal mapping */
int optimal_edit_score(RNA_sequence* rnaA, RNA_sequence* rnaB, int primary, int similar, int full_edit, int eq, char *eqA, char *eqB){
  RNA_sequence* alignA =NULL; 
  RNA_sequence * alignB=NULL;
  char * nameCtA=NULL; 
  char * nameCtB=NULL;
  char * nameFileOut=NULL;
  return optimal_edit(rnaA, rnaB, primary, similar, nameCtA, nameCtB, nameFileOut, alignA, alignB, full_edit, eq, eqA, eqB, 0); 
}


int optimal_edit_mapping(RNA_sequence* rnaA, RNA_sequence* rnaB, int primary, int similar, char* nameCtA, char * nameCtB, char * nameFileOut, RNA_sequence* alignA, RNA_sequence* alignB, int full_edit, int eq, char *eqA, char *eqB){
  return optimal_edit(rnaA, rnaB, primary, similar, nameCtA, nameCtB, nameFileOut, alignA, alignB, full_edit, eq, eqA, eqB, 1); 
}
