/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or */
/* (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "../headers/score.h"
#include "../headers/constraint.h"
#include "../headers/stack.h"
#include "../headers/RNA_structure.h"
#include "../headers/graph.h"
#include "../headers/matrix.h"
#include "../headers/arc_alignment_graph.h"

/***********************************************************/
/* local alignment                                         */
/***********************************************************/

/* table between  [i][j] and [k][l]
 * max_x: X-coordinate of the maximal value
 * max_y: Y-coordinate of the maximal value 
 * max_cell: maximal value
 */

void computeTables_local(postorder_tree * treeA, postorder_tree * treeB,int i,int j, int k, int l, int * max_x, int * max_y, int * max_cell, int * max_table){
  int x,y,z;
  *max_cell=0;
  
  /* initialisation */
  table[0][i][j]=0;
  table[1][i][j]=0; 
  table[2][i][j]=0;
  table[3][i][j]=0;
  
  /* first row */
  for (x=i+1; x<=k; x++)
    if  (constraint[x][j]==ALLOWED){
      table[0][x][j]=0;
      table[1][x][j]=0;   
      table[2][x][j]=0;
      table[3][x][j]=0;
    }   
  
  /* first column */
  for (y=j+1; y<=l; y++)
    if (constraint[i][y]==ALLOWED){
      table[0][i][y]=0; 
      table[1][i][y]=0; 
      table[2][i][y]=0;
      table[3][i][y]=0;
    }
  
  /* other cells */
  for (x=i+1; x<=k; x++) 
    if (is_closedA[i+1][x])
      for (y=j+1; y<=l; y++)
	if (is_closedB[j+1][y]){
	  if (constraint[x][y]==ALLOWED){
	    computeTables_Cell(treeA, treeB,x, y, i, j);
	    for (z=0; z<4; z++){
	      if (table[z][x][y]<0) 
		table[z][x][y]=0;
	      else if (table[z][x][y]>*max_cell){
		*max_table=z;
		*max_x=x;
		*max_y=y;
		*max_cell=table[z][x][y];
	      } 
	    }
	  }
	}
}


int minimal_size(int s){
  return s/10;
}

/* search for the best pair of closed subforests */
void best_pair_of_closed_subforests(postorder_tree * treeA, postorder_tree * treeB, int * i, int *j, int *k, int *l, int *source_table){
  int x, y;
  int max_cell, max_x, max_y,max_table; 
  int score=0; 
  int nodeA=0, nodeB=0;
  int finished; 
  edge_gap ed; 
  int d;
  int min_size=5;
  
  /* external closed subforest */
  computeTables_local(treeA, treeB, 0,0, treeA->size-1, treeB->size-1,  &max_x, &max_y, &max_cell, &max_table);
  if (max_cell>score) {
    score=max_cell; *k=max_x; *l=max_y; nodeA=0; nodeB=0; *source_table=max_table;  min_size=minimal_size(score);
  }
  
  /* internal closed subforests  */
  for (x=treeA->size-1; x>=1; x--)
    if (treeA->n[x].size>min_size)
      for (y=treeB->size-1; y>=1; y--)
	if (treeB->n[y].size>min_size) {
	  /* look at children */
	  computeTables_local(treeA, treeB, x-treeA->n[x].size, y-treeB->n[y].size, x-1, y-1, &max_x, &max_y, &max_cell,&max_table);
	  if (max_cell>score) {
	    score=max_cell; *k=max_x; *l=max_y; nodeA= x-treeA->n[x].size; nodeB= y-treeB->n[y].size; *source_table=max_table;
	    min_size=minimal_size(score);
	  }
	}
    
  /* search for the beginning of the path coming from  (k,l) in the table corresponding to  nodeA, nodeB*/
  computeTables_local(treeA, treeB, nodeA, nodeB, *k, *l, &max_x, &max_y, &max_cell,&max_table);
  x=*k; 
  y=*l;
  while (table[*source_table][x][y]>0){ 
    finished=0;
    /*******************/
    /* implicit edges  */
    /*******************/
    if (*source_table==0){  /* base-match */   
      x--; y--; 
      *source_table=best_table(x,y,ABCD);
      finished=1;
    }
    else  if (*source_table==1) { /* full deletion */
      finished=1;
      if (table[1][x][y]==get_tables(full_deletionA[x], x-treeA->n[x].size, y, B)){ 
	x-=treeA->n[x].size;
      }
      else if (table[1][x][y]==get_tables(full_deletionA[x]+gap_open, x-treeA->n[x].size, y, ACD)){
	x-=treeA->n[x].size; 
	*source_table=best_table(x,y,ACD); 
      }
      else if (table[1][x][y]==get_tables(full_deletionB[y],x, y-treeB->n[y].size, B)){
	y-=treeB->n[y].size;	
      }
      else  if (table[1][x][y]==get_tables(full_deletionB[y]+gap_open,x, y-treeB->n[y].size, ACD)){
	y-=treeB->n[y].size;
	*source_table=best_table(x,y,ACD);	
      }
    }
    
    /*******************/
    /* explicit edges  */
    /*******************/
    for (d=0;  (!finished) && (d<=alignment_graph[x][y].current_edge); d++){
      get_edge_gap(&alignment_graph[x][y], &ed, d);
      if (table[*source_table][x][y]==get_tables(ed.weight,ed.targetX,ed.targetY,ABCD)){
	finished=1; 
	x=ed.targetX; 
	y=ed.targetY;
	*source_table=best_table(x,y,ABCD);
      }
    }
  }
  *i=x; 
  *j=y;
  *source_table=best_table(*k,*l,ABCD);
}

      
int trace_back_local_alignment(postorder_tree * treeA, postorder_tree * treeB, int ** supertree, int * startA, int * endA, int * startB, int * endB){
  int score; 
  int i, j, k, l,x, source_table, positionA, positionB; 

  seven_stack *  p;
  int old_i=0; 
  int old_j=0;;
  
  p=malloc(sizeof(seven_stack));

  best_pair_of_closed_subforests(treeA,treeB, &i,&j,&k,&l,&source_table);
  if (treeA->n[k].size==1)
    *endA=treeA->n[k].positionX; 
  else 
    *endA=treeA->n[k].positionY;
 if (treeB->n[l].size==1)
    *endB=treeB->n[l].positionX; 
  else 
    *endB=treeB->n[l].positionY;

 *startA=treeA->size*2;
 *startB=treeB->size*2;
 for (x=i+1; x<=k; x++)
   if (treeA->n[x].positionX <*startA) *startA=treeA->n[x].positionX;

 for (x=j+1; x<=l; x++)
   if (treeB->n[x].positionX <*startB) *startB=treeB->n[x].positionX;

   init_seven_stack(p,treeA->size+treeB->size);
  computeTables(treeA, treeB,i,j,k,l); /* utile ? on peut peut-etre reutiliser la table de best_pair etc. */
  /* source_table=best_table(k,l,ABCD);  utile ? */
  score=table[source_table][k][l];
  build_optimal_path(treeA, treeB, i,j,k,l, source_table, i, j,p,supertree);
 
  while(!is_empty_seven_stack(p)){
    seven_pop(p,&i,&j, &positionA, &positionB,&k,&l,&source_table);
    if ((i!=old_i) ||(j!=old_j)){
      old_i=i; old_j=j;
      computeTables(treeA, treeB, i,j,k,l);
    }   
    build_optimal_path(treeA, treeB, i,j, k, l,source_table, positionA, positionB,p,supertree);
  }
  return score;
}


int optimal_local(RNA_sequence * rnaA, RNA_sequence * rnaB,  RNA_sequence* alignA, RNA_sequence* alignB, int primary, int eq,
			    char *eqA, char *eqB, int do_mapping){

  int i,j,score;
  int ** supertree;
  postorder_tree * treeA;
  postorder_tree * treeB;
  int startA, startB, endA, endB; 
  int source_table;

  treeA=(postorder_tree * ) malloc(sizeof(postorder_tree));
  treeB=(postorder_tree * ) malloc(sizeof(postorder_tree));
  
  RNA_sequence2postorder(rnaA,treeA);
  RNA_sequence2postorder(rnaB,treeB);
	
  
 /*   prise en compte des contraintes */
  constraint=new_matrix_char(treeA,treeB);	
  no_constraint(constraint,treeA->size,treeB->size);
  
  if(primary==1)
    primary_sequence_constraint(constraint,treeA,treeB,rnaA,rnaB);

  build_alignment_graph(treeA, treeB);
 

 if (do_mapping){
   supertree = (int **)malloc( sizeof(int*)*(rnaA->size+1) );
   for (i=0; i<=rnaA->size; i++)
     supertree[i]= (int *)malloc( sizeof(int)*(rnaB->size+1) );
   for (i=0; i<=rnaA->size; i++)
     for (j=0; j<=rnaB->size; j++)
       supertree[i][j]=0;
   score= trace_back_local_alignment(treeA, treeB,supertree, &startA, &endA, &startB, &endB);
   create_alignment_RNA_sequences(rnaA,rnaB, supertree, alignA,  alignB, startA, endA, startB, endB);
 }   
 else {
   best_pair_of_closed_subforests(treeA, treeB, &startA, &startB, &endA, &endB, &source_table);
   computeTables(treeA,treeB,startA,startB,endA,endB);
   /* source_table=best_table(endA,endB,ABCD);*/
  score=table[source_table][endA][endB];
  }

  return score;
}
 



int optimal_local_alignment(RNA_sequence * rnaA, RNA_sequence * rnaB,  RNA_sequence* alignA, RNA_sequence* alignB, int primary, int eq,
			    char *eqA, char *eqB){

  int i,j,score;
  int ** supertree;
  postorder_tree * treeA;
  postorder_tree * treeB;
  int startA, startB, endA, endB; 

  treeA=(postorder_tree * ) malloc(sizeof(postorder_tree));
  treeB=(postorder_tree * ) malloc(sizeof(postorder_tree));
  
  RNA_sequence2postorder(rnaA,treeA);
  RNA_sequence2postorder(rnaB,treeB);
	

 supertree = (int **)malloc( sizeof(int*)*(rnaA->size+1) );
   for (i=0; i<=rnaA->size; i++)
     supertree[i]= (int *)malloc( sizeof(int)*(rnaB->size+1) );
   for (i=0; i<=rnaA->size; i++)
     for (j=0; j<=rnaB->size; j++)
       supertree[i][j]=0;
  
 /*   taking constraints into account */
  constraint=new_matrix_char(treeA,treeB);	
  no_constraint(constraint,treeA->size,treeB->size);
  
  if(primary==1)
    primary_sequence_constraint(constraint,treeA,treeB,rnaA,rnaB);

  build_alignment_graph(treeA, treeB);
  
 /* display_alignment_graph(alignment_graph, treeA->size-1, treeB->size-1); */
  score= trace_back_local_alignment(treeA, treeB, supertree, &startA, &endA, &startB, &endB);
  create_alignment_RNA_sequences(rnaA,rnaB, supertree, alignA,  alignB, startA, endA, startB, endB);
    
  return score;
}



int optimal_local_score(RNA_sequence * rnaA, RNA_sequence * rnaB, int primary, int eq, char *eqA, char *eqB){

  int score;
  postorder_tree * treeA;
  postorder_tree * treeB;
  int startA, startB, endA, endB; 
  int source_table;

  treeA=(postorder_tree * ) malloc(sizeof(postorder_tree));
  treeB=(postorder_tree * ) malloc(sizeof(postorder_tree));
  
  RNA_sequence2postorder(rnaA,treeA);
  RNA_sequence2postorder(rnaB,treeB);
	
 /* taking into account  constraints */
  constraint=new_matrix_char(treeA,treeB);	
  no_constraint(constraint,treeA->size,treeB->size);
  if(primary==1)
    primary_sequence_constraint(constraint,treeA,treeB,rnaA,rnaB);

  build_alignment_graph(treeA, treeB);
  best_pair_of_closed_subforests(treeA, treeB, &startA, &startB, &endA, &endB, &source_table);
  computeTables(treeA,treeB,startA,startB,endA,endB);

  /* source_table=best_table(endA,endB,ABCD);*/
  score=table[source_table][endA][endB];
  return score;
}
