/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or */
/* (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <getopt.h>
#include "../headers/RNA_structure.h"
#include "../headers/RNA_multiple_structure.h"
#include "../headers/score.h"
#include "../headers/edit_distance.h"
#include "../headers/arc_alignment.h"
#include "../headers/multiple_arc_alignment.h"
#include "../headers/clustering.h"
#include "../headers/local_arc_alignment.h"
#include "../headers/tree_alignment.h"
#include "../headers/arc_alignment_without_gap.h"

extern int getopt(int,char * const *,const char *);
extern char *optarg;
extern int optind;
extern int opterr;
extern int optopt;


void pairwise_alignment(RNA_sequence_m **list, int EQ, int SIMILAR, int FASTA, 
			int PRIMARY, int MAPPING, int ARC, int LOCAL, int ALIGN, 
			int FULL_EDIT, int FILEOUT, int gap_open, char *nameFileOut);

/*compare a set of RNA structures*/ 
int main(int argc, char *argv[]){
  
  int i;
  double elapsedTime;
  int ALIGN=1;
  int ARC=1;
  int PRIMARY=0;
  int LOCAL=0;
  int MAPPING=1;
  int FILEOUT=0;
  int FILESCORE=0;
  int FULL_EDIT=0;
  int SIMILAR=0;
  int USE_MATRIX_FILE=0;
  int FASTA=0 ;
  int EQ=0 ;
  int CLUSTAL=0;
  int NOGAP=0;
  int c=0;
  clock_t stopTime;
  clock_t startTime = clock();
  RNA_sequence_m **list;
  char *nameFileIn=argv[1];
  char *nameFileOut=NULL; /*=argv[3];*/ 
  char *nameFileScore=NULL; 
  char *nameFileMatrix=NULL;
  int opt_index;
  int nb_seq;

  static const char *optstring = "S:o:M:lefghcnms";
  struct option options[] = {
    {"fasta",          no_argument,       &FASTA,      1},
    {"eq",             no_argument,       &EQ,         1},
    {"clustalw",       no_argument,       &CLUSTAL,    1},
    {0,                0,                 0,           0}
  };
  
  if (argc<2) {
    printf ("Usage : \n gardenia inputFile [-S scoreFile] [-o outputFile] [-M matrixFile] [-lefghcnms] [--fasta] [--eq] \n");
    printf ("\t -S      : score file\n");
    printf ("\t --fasta : output in fasta format\n");
    printf ("\t -o      : output file\n");
    printf ("\t --eq    : eq file\n");
    printf ("\t -s      : similar structures\n");
    printf ("\t -M      : use scoring matrices\n");
    printf ("\nAvailable only for pairwise alignments:\n");
    printf ("\t -l      : local comparison\n");
    printf ("\t -e      : tree edit distance\n");
    printf ("\t -f      : tree edit distance with full edit operations \n");
    printf ("\t -g      : tree alignment \n");
    printf ("\t -h      : tree alignment with full edit operations\n");
    printf ("\t -c      : constraints from the primary structure \n");
    printf ("\t -n      : no mapping\n");
    printf ("\t -m      : multiple alignment (no affine gap cost)\n");

    exit(EXIT_SUCCESS);
  }
  
  while ((c = getopt_long(argc,argv, optstring, options, &opt_index)) != -1) {
    switch (c) {
    case 'S' : /*score file */
      nameFileScore = (char*) malloc (strlen(optarg)+1 * sizeof(char));
      strcpy (nameFileScore, optarg);
      FILESCORE=1;
      break;
    case 'o' : /*result file */
      nameFileOut = (char*) malloc (strlen(optarg)+1 * sizeof(char));
      strcpy (nameFileOut, optarg);
      FILEOUT=1;
      break;
    case 's' : /* similar structures */
      SIMILAR=1;
      break;
    case 'M' : /* matrix score */
      nameFileMatrix = (char *)malloc(strlen(optarg)+1 * sizeof(char));
      strcpy(nameFileMatrix, optarg);
      USE_MATRIX_FILE = 1;
      break;
    case 'l' : /* local comparison */
      LOCAL=1;
      break;
    case 'e' : /* tree edit distance */
      ARC=0;
      ALIGN=0;
      break;
    case 'f' : /* tree edit distance, full edit */
      ARC=0;
      ALIGN=0;
      FULL_EDIT=1;
      break;
    case 'g' : /* tree alignment */
      ARC=0;
      ALIGN=1;
      break;
    case 'h' : /* tree alignment, full edit */
      ARC=0;
      ALIGN=1;
      FULL_EDIT=1;
      break;
    case 'c' : /* constraints frome the primary structure */
      PRIMARY=1;
      break;
    case 'm' : /* multiple alignment: no gap */
      NOGAP=1;
      break;
    case 'n' : /* mapping */
      MAPPING=0;
      break;
    }
  }
  printf("Start\n");
  printf("File in  : %s\n",nameFileIn); 
   
  if(PRIMARY) printf("Sequence constraint : true\n");
  else printf("Sequence constraint : false\n");
  
  if (EQ) printf("Eq constraint : true \n");
  else  printf("Eq constraint : false \n");

  if (FASTA) printf("outputFile type : multifasta (clustalw-like)\n");
  
  if (MAPPING) printf("Construction of the mapping : true\n");
  else printf("Construction of the mapping : false\n");
  
  if (!ARC)
    if (FULL_EDIT) printf("RNA model : enriched tree\n");
    else printf("RNA model : simple tree\n");
  else
    printf("RNA model : arc annotated sequence\n");

  if (ALIGN) printf("Comparison model : alignment \n");
  else printf("Comparison model : edit distance \n");
  
  if (SIMILAR) printf("k-strip algorithm : yes \n");
  else  printf("k-strip algorithm : no \n"); 

  printf("\n");
 
  list = get_multiple_bracket( nameFileIn, &nb_seq, &EQ );
  
  if (FILESCORE==0){
    if (USE_MATRIX_FILE==1){
      if ((ARC) && (nb_seq==2) && (NOGAP==0)) {
	nameFileScore = (char*) malloc (16 * sizeof(char));
	strcpy (nameFileScore, "ribosumgaps.txt"); 
      }
      else{
	nameFileScore = (char*) malloc (12 * sizeof(char));
	strcpy (nameFileScore, "ribosum.txt"); 
      }
    }
    else{
      if ((ARC) && (nb_seq==2) && (NOGAP==0)){
	nameFileScore = (char*) malloc (9 * sizeof(char));
	strcpy (nameFileScore, "gaps.txt");
      }
      else {
	nameFileScore = (char*) malloc (10 * sizeof(char));
	strcpy (nameFileScore, "score.txt"); 
      }
    }
  }
  

  if(FILEOUT == 0){ 
    nameFileOut = (char*) malloc (1 * sizeof(char));
    nameFileOut[0] = '\0';
  }
  
  get_score(nameFileScore, nameFileMatrix, &base_mat, &base_mis, &base_del, &arc_mat, &arc_mismat, 
	    &arc_mis, &arc_rem, &arc_alt_mat, &arc_alt_mis, &arc_bre_mat, &arc_bre_mismat,
	    &arc_bre_mis, &gap_open,USE_MATRIX_FILE,ARC,FULL_EDIT);

  if (NOGAP==1) gap_open=0;

  print_score( nameFileMatrix, base_mat, base_mis, base_del, arc_mat, arc_mismat, 
	    arc_mis, arc_rem, arc_alt_mat, arc_alt_mis, arc_bre_mat, arc_bre_mismat,
	    arc_bre_mis, gap_open,USE_MATRIX_FILE,ARC,FULL_EDIT);

 
  printf("Number of sequences : %d\n", nb_seq); 

  if ( nb_seq > 2 ){ /* multiple alignment */
    printf("Computing the multiple alignment...\n");
    multiple_alignment(list, nb_seq, EQ, SIMILAR, FASTA, CLUSTAL, nameFileOut );
  }
  else if (nb_seq == 2){/* pairwise alignment */
    printf("Computing the pairwise alignment...\n");
    pairwise_alignment(list, EQ, SIMILAR, FASTA, PRIMARY, MAPPING, ARC, LOCAL,
		       ALIGN, FULL_EDIT, FILEOUT, gap_open, nameFileOut);
  }
  else{
    printf("Input file must contain at least two sequences. \n");
  }  

  /* free memory */
  free(nameFileOut);
  free(nameFileScore);
  if(USE_MATRIX_FILE)
    free_matrices();
  for(i=0; i<nb_seq; i++){   
    free_RNA_sequence_m(list[i]);
    free(list[i]);
  }
  free(list);

  stopTime = clock();
  elapsedTime = (stopTime - startTime) / (CLOCKS_PER_SEC / (double) 1000.0);
  printf("Elapsed time: %1.0f ms\n",elapsedTime);
    
  return 0;
}



void pairwise_alignment(RNA_sequence_m **list, int EQ, int SIMILAR, int FASTA, int PRIMARY, 
			int MAPPING, int ARC, int LOCAL, int ALIGN, int FULL_EDIT, int FILEOUT,
			int gap_open, char *nameFileOut){

  RNA_sequence *rnaA, *rnaB, *alignA, *alignB;
  int score = 0;

  alignA=(RNA_sequence *) malloc(sizeof(RNA_sequence));
  alignB=(RNA_sequence *) malloc(sizeof(RNA_sequence));
  
  rnaA = &(list[0]->s[0]);
  rnaB = &(list[1]->s[0]);
    
  /*  display_RNA_sequence_in_bracket_notation(rnaA);
      display_RNA_sequence_in_bracket_notation(rnaB);  */
  
  if(ARC){
    if( gap_open == 0){
      if(LOCAL)
	printf("[!!] local alignment not supported with these options...\n");

      if(MAPPING)
	score = optimal_arc_alignment_without_gap_mapping(rnaA, rnaB, alignA, alignB, SIMILAR, PRIMARY);
      else
	score = optimal_arc_alignment_without_gap_score(rnaA, rnaB, SIMILAR, PRIMARY);
    }
    else{
      if (MAPPING){
	if (LOCAL)
	  score = optimal_local_alignment(rnaA, rnaB, alignA, alignB,PRIMARY, EQ, list[0]->eq, list[1]->eq);
	else
	  score = optimal_arc_alignment(rnaA, rnaB, alignA, alignB,PRIMARY, SIMILAR, EQ, list[0]->eq, list[1]->eq);
      }
      else{
	if(LOCAL)
	  score=optimal_local_score(rnaA, rnaB,PRIMARY, EQ, list[0]->eq, list[1]->eq);
	else
	  score=optimal_arc_score(rnaA, rnaB,PRIMARY,SIMILAR, EQ, list[0]->eq, list[1]->eq);
      }
    }
  }
  else{
    if(LOCAL)
      printf("[!!] local alignment not supported with these options ...\n");

    if(!ALIGN){ /* edition */
      if(MAPPING)
	score=optimal_edit_mapping(rnaA, rnaB, PRIMARY, SIMILAR, rnaA->name, rnaB->name, nameFileOut, 
			      alignA, alignB, FULL_EDIT, EQ, list[0]->eq, list[1]->eq);
      else
	score=optimal_edit_score(rnaA, rnaB, PRIMARY, SIMILAR, FULL_EDIT, EQ, list[0]->eq, list[1]->eq);
    }
    else{ /* alignment */
      if( MAPPING )
	score = optimal_tree_alignment_mapping(rnaA, rnaB, alignA, alignB, FULL_EDIT,SIMILAR,PRIMARY);
      else
      	score = optimal_tree_alignment_score(rnaA, rnaB, FULL_EDIT,SIMILAR,PRIMARY);
    }
  }

  printf("\nScore : %d\n", score);  
  
  if(MAPPING){
    if(LOCAL)  
      printf("Pairwise local alignment ");
    else 
      printf("Pairwise global alignment ");
    if(FILEOUT)
      printf("available in %s\n", nameFileOut);
    else 
      printf("\n\n");
    if( !FASTA )
      display_alignment(rnaA->name, rnaB->name, alignA, alignB, nameFileOut);
    else
      pairwise_fasta_output(rnaA->name, rnaB->name, alignA, alignB, nameFileOut);
  }

  free(alignA);
  free(alignB);
}
