/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or */
/* (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "../headers/multiple_arc_alignment.h"
#include "../headers/score_m.h"
#include "../headers/stack.h"
#include "../headers/RNA_structure.h"
#include "../headers/RNA_multiple_structure.h"
#include "../headers/graph.h"
#include "../headers/matrix.h"
#include "../headers/constraint.h"
#include "../headers/clustering.h"

static vertex** alignment_graph;
static int ** table;
static int * full_deletionA, * full_deletionB;
static char ** is_closedA;
static char ** is_closedB;
static char **constraint;


/* number of closed subforests ending at position i, 
 * including i
 */
static void tree_closed_m(postorder_tree_m *tree, short int *tab_closed){
  int i; 
  int j; 
  /*  tab_closed[tree->size]=0;*/
  for (i=1; i<tree->size;i++){
    j = i - tree->n[i].size; 
    if ( (j>0) && (tree->n[j].positionY == tree->n[i].positionX-1) )
      tab_closed[i] = tab_closed[j]+1;
    else 
      tab_closed[i] = 1;
  }
}

/* ancestor[i]=x : x is the heighest node of treesuch that
 * i is a leftmost leaf 
 */
static void ancestor_m(postorder_tree_m *tree, int *tab_ancestor){
  int i ; 
  for (i=1;i<tree->size; i++){
    tab_ancestor[i] = i;
  }
  for (i=1;i<tree->size; i++){
    tab_ancestor[i-tree->n[i].size+1] = i;
  }
}

/* right_sibling[i]=x : x is the right sibling of  i, if it exists, -1 otherwise 
 */
static void right_sibling_m(postorder_tree_m *tree, int *tab){
  int i ;  
  for (i=1;i<tree->size; i++){
    tab[i] = -1;
  }
  for (i=1;i<tree->size; i++){
    tab[i-tree->n[i].size] = i;
  }
}


/* if no father : -1 */
static void father_m(postorder_tree_m *tree, int *tab ){
  int i ,j; 
  for (i=1;i<tree->size; i++){
    tab[i] = -1;
  }
  for (i=1;i<tree->size; i++){
    if (tree->n[i].size>1) 
      for (j=i-1; j>i-tree->n[i].size; j=j-tree->n[j].size)
	tab[j] = i;
  }
}

static void is_closed_forest_m(char ** tab, postorder_tree_m *tree){
  int x = 0;
  int *child; 
  int y, z, t;
  int i, j;
  
  child = (int *)malloc( sizeof(int) * tree->size );

  for (i=1; i<tree->size; i++)
    for (j=1; j<tree->size; j++)
      tab[i][j] = 0;

  for (i=1; i<tree->size; i++){
    if (tree->n[i].size>1){ /* internal node of tree */
      x = 0;
      for (j=i-1;  j>i-tree->n[i].size; j=j-tree->n[j].size){ /* looking for children */
	child[x] = j;
	x++;
      }
      for (j=0; j<x;j++){
	t = child[j];
	for (y=0; y<=j; y++){
	  z = child[y];
	  tab[t-tree->n[t].size+1][z] = 1;
	  tab[z][t-tree->n[t].size+1] = 1;
	}
      }
    }/* end internal node */
  }
  x=0;
 
  for (j=tree->size-1;  j>0; j=j-tree->n[j].size){
    child[x] = j;
    x++;
  }
  for (j=0; j<x;j++){
    t = child[j];
    for (y=0;y<=j; y++){
      z = child[y];
      tab[t-tree->n[t].size+1][z] = 1;
      tab[z][t-tree->n[t].size+1] = 1;
    }
  }
  free(child);
}


/* rightmost_sibling[i]=x : x is the rightmost sibling of i, if it exists, 
 * i otherwise */
static void rightmost_sibling_m(postorder_tree_m *tree, int *tab){
  int * tab2;
  int i; 
  tab2=(int *) malloc( sizeof(int) * tree->size );
  right_sibling_m(tree, tab2);
  for (i=tree->size-1; i>0; i--){
    if (tab2[i]!=-1) 
      tab[i] = tab[tab2[i]];
    else 
      tab[i] = i;
  }
  free(tab2);
}

static int is_subtree_m(postorder_tree_m *tree, int i, int j){
  return ( (tree->n[j].size>1) && (j-tree->n[j].size+1==i) );
}

/* static int is_relevant_pair_m(postorder_tree_m * treeA, postorder_tree_m *treeB, int i, int x, int j, int y){ */
 
/*   /\* return 1;*\/  */
/*   if (is_subtree_m(treeA,i,x)) return 1;  */
/*   if (is_subtree_m(treeB,j,y)) return 1;  */
/*   if ( (x < treeA->size-1) && (is_subtree_m(treeA,i,x+1)) ) return 1; */
/*   if ( (y < treeB->size-1) && (is_subtree_m(treeB,j,y+1)) ) return 1; */
/*   if ( y == treeB->size-1) return 1;  */
/*   if ( x == treeA->size-1) return 1;  */
/*   return 0; */
/* } */

static int tree_full_deletion_m(postorder_tree_m *tree, int i, int nb_seqB){
  int w = 0;
  int x;
  int *paired = NULL;
  char *gap = NULL;

  paired = (int *)malloc(sizeof(int));
  paired[0] = 0;
  gap = (char *)malloc(sizeof(char)*nb_seqB);
  
  for (x=0; x<nb_seqB; x++)
    gap[x] = 4;

  for (x=i; x>i-tree->n[i].size; x--)
    if ( tree->n[x].size == 1 )
      w += base_deletion_m( tree->n[x].labelX, tree->nb_seq, nb_seqB);
    else
      w += score_sp(tree->n[x].labelX, tree->n[x].labelY, gap, gap, tree->nb_seq, 1, tree->n[x].paired, paired);

  free(gap);
  free(paired);
  return w; 
}


static int get_table(int x, int y){
  if( constraint[x][y] ==  ALLOWED )
    return table[x][y];
  else
    return -1000000;
}

static void computeTable_Cell_m(postorder_tree_m *treeA, postorder_tree_m *treeB, int x, int y, int i, int j){
  int  d, max_d,w;
  edge ed;
  int w_max = -1000000;

  if (constraint[x][y]==ALLOWED){
  
  /*** we look at all edges coming from (x,y) ***/
  
  /* implicit edges  */
  
  /* full deletion */
  if ( x>i )
    w_max = get_table(x-treeA->n[x].size, y) + full_deletionA[x];
  /* full insertion */
  if ( y>j ){
    w = get_table(x, y-treeB->n[y].size) + full_deletionB[y];
    if ( w>w_max ) { w_max = w; }
  }
  /* substitutions between free bases */
  if ( (x>i) && (y>j) && (treeA->n[x].size == 1) && (treeB->n[y].size == 1) ){  
    w = get_table(x-1, y-1) + base_mismatch_m(treeA->n[x].labelX, treeB->n[y].labelX, treeA->nb_seq, treeB->nb_seq);
    if ( w>w_max ) { w_max = w; }
  }

  /* explicit edges  */
  max_d = alignment_graph[x][y].current_edge + 1;
  for(d=0; d<max_d; d++){
    get_edge(&alignment_graph[x][y], &ed, d);

    if( (ed.targetX >= i) && (ed.targetY >= j) ){
      w = ed.weight + get_table(ed.targetX, ed.targetY);
      if( w>w_max ) { w_max = w; }
    }
  }

  table[x][y] = w_max;
  }
}

static void computeTable_m(postorder_tree_m *treeA, postorder_tree_m *treeB, int i, int j, int k, int l){
  int x,y;
  table[i][j] = 0;
  
  for(x=i+1; x<=k; x++)
    table[x][j] = full_deletionA[x] + get_table(x-treeA->n[x].size, j); /*a revoir */
  for(y=j+1; y<=l; y++)
    table[i][y] = full_deletionB[y] + get_table(i, y-treeB->n[y].size); /*a revoir */

  for(x=i+1; x<=k; x++)
    if( is_closedA[i+1][x] )
      for(y=j+1; y<=l; y++)
	if( is_closedB[j+1][y] )
	  computeTable_Cell_m(treeA, treeB, x, y, i, j);	
}

static void build_alignment_graph_m(postorder_tree_m *treeA, postorder_tree_m *treeB){
  int i,j,k,l,m,n,x,y;
  short int *tree_closedA;
  short int *tree_closedB; 
  int *rightmostA;
  int *rightmostB;
  int *ancestorA;
  int *ancestorB;
  int *fatherA;
  int *fatherB;
  int *pairedA, *pairedB;
  char *gapA, *gapB;
  edge ed; 
  int w;

  int degree = 0;

  ed.edit_operation = 0;
  tree_closedA = (short int *)malloc( sizeof(short int) * treeA->size );
  tree_closedB = (short int *)malloc( sizeof(short int) * treeB->size ); 
  tree_closed_m(treeA, tree_closedA);
  tree_closed_m(treeB, tree_closedB);
  
  ancestorA = (int *)malloc( sizeof(int) * treeA->size );
  ancestorB = (int *)malloc( sizeof(int) * treeB->size ); 
  ancestor_m(treeA, ancestorA);
  ancestor_m(treeB, ancestorB);
 
  rightmostA = (int *)malloc( sizeof(int) * treeA->size ); 
  rightmostB = (int *)malloc( sizeof(int) * treeB->size );
  rightmost_sibling_m(treeA, rightmostA);
  rightmost_sibling_m(treeB, rightmostB);
 
  fatherA = (int *)malloc( sizeof(int) * treeA->size );
  fatherB = (int *)malloc( sizeof(int) * treeB->size ); 
  father_m(treeA, fatherA);
  father_m(treeB, fatherB);

  is_closedA = (char **)malloc( sizeof(char *) * treeA->size );
  for(i=1; i<treeA->size; i++){
    is_closedA[i] = (char*)malloc( sizeof(char) * treeA->size );
  }
  is_closedB = (char **)malloc( sizeof(char *) * treeB->size );
  for(i=1; i<treeB->size; i++){
    is_closedB[i] = (char *)malloc( sizeof(char) * treeB->size );
  }
  is_closed_forest_m(is_closedA, treeA);
  is_closed_forest_m(is_closedB, treeB);
     
  
  pairedA = (int *)malloc( treeA->nb_seq * sizeof(int) );
  gapA = (char *)malloc( sizeof(char) * treeA->nb_seq );
  for(i=0; i<treeA->nb_seq; i++){
    pairedA[i] = 0;
    gapA[i] = 4;
  }
  pairedB = (int *)malloc( treeB->nb_seq * sizeof(int) );
  gapB = (char *)malloc( sizeof(char) * treeB->nb_seq );
  for(i=0; i<treeB->nb_seq; i++){
    pairedB[i] = 0;
    gapB[i] = 4;
  }

  /**********************************/
  /*    skeleton of the graph       */
  /**********************************/ 

  for (i=1; i<treeA->size; i++){
    for (j=1; j<treeB->size; j++){
      if (constraint[i][j]==ALLOWED){
        if ((treeA->n[i].size>1) &&  (treeB->n[j].size>1)) degree=2*(tree_closedB[j]+tree_closedA[i]);
	else if (treeA->n[i].size>1) degree=2* tree_closedB[j]; 
	else if (treeB->n[j].size>1) degree=2* tree_closedA[i];
	else degree=0; 
	init_edge_stack(&alignment_graph[i][j],degree);
      } 
      else init_edge_stack(&alignment_graph[i][j],0);
    }
  }
  
  /*******************************/ 
  /*     weight of the edges     */
  /*******************************/
  
  for(i=treeA->size-1; i>=1; i--){
    if( treeA->n[i].size == 1 ) {
      for(j=treeB->size-1; j>=1; j--){
	if( treeB->n[j].size == 1 && constraint[i-1][j-1] == ALLOWED ) { 

	  k = ancestorA[i];
	  l = ancestorB[j];
	  m = rightmostA[k];
	  n = rightmostB[l];

	  /* table[x][y] : cost of an optimal path from (x,y) to (i-1,j-1).  
	   * table and alignment_graph are computed simultaneously */
	  
	  table[i-1][j-1] = 0;
	  for(x=i; x<=m; x++)
	    table[x][j-1] = full_deletionA[x]+ get_table(x-treeA->n[x].size, j-1); /*a revoir */
 	  for(y=j; y<=n; y++)
	    table[i-1][y] = full_deletionB[y]+ get_table(i-1, y-treeB->n[y].size); /*a revoir */  
	  	  
	  /**********************/
	  /*       zone I       */
	  /**********************/
	  for(x=i; x<=k; x++){
	    if( is_closedA[i][x] ){
	      for(y=j; y<=l; y++){  
		if( is_closedB[j][y] && constraint[x][y] == ALLOWED ){
		  
		  computeTable_Cell_m(treeA, treeB, x, y, i-1, j-1);
		  
		  /*** edge going from vertex (x,y) to vertex (i-1,j-1) ***/

		  ed.targetX = i-1; 
		  ed.targetY = j-1;	
		  ed.weight = table[x][y]; 
		  
		  /* arc_mismatch between arcs x in A and y in B */
		  if( (is_subtree_m(treeA, i, x)) && (is_subtree_m(treeB, j, y)) ){
		    w = score_sp( treeA->n[x].labelX, treeA->n[x].labelY, treeB->n[y].labelX, treeB->n[y].labelY,
				  treeA->nb_seq, treeB->nb_seq, treeA->n[x].paired, treeB->n[y].paired) 
		      + get_table(x-1, y-1);

		    update(w, 0, &ed);
		  }
		  
		  /* arc-removing of x in A */      
		  if( is_subtree_m(treeA, i, x) ) {
 
		    w = score_sp(treeA->n[x].labelX, treeA->n[x].labelY, gapB, gapB,
				 treeA->nb_seq, treeB->nb_seq, treeA->n[x].paired, pairedB) + get_table(x-1, y);
		    update(w, 1, &ed);
		  }
		  
		  /* arc-altering of x in y */
		  if( (is_subtree_m(treeA, i, x)) && (treeB->n[y].size == 1) ){

		    w = score_sp(treeA->n[x].labelX, treeA->n[x].labelY, gapB, treeB->n[y].labelX, 
				 treeA->nb_seq, treeB->nb_seq, treeA->n[x].paired,pairedB) 
		      + get_table(x-1, y-1);

		    update(w, 3, &ed);
		  }
		  
		  /* arc-removing of y in B */
		  if( is_subtree_m(treeB, j, y) ) {

		    w = score_sp( gapA, gapA, treeB->n[y].labelX, treeB->n[y].labelY,
				 treeA->nb_seq, treeB->nb_seq, pairedA, treeB->n[y].paired) + get_table(x, y-1);
		    update(w, 5, &ed);
		  }
		  
		  /* arc-altering of y in x */
		  if( (is_subtree_m(treeB, j, y)) && (treeA->n[x].size == 1) ){
		    w = score_sp(gapA, treeA->n[x].labelX, treeB->n[y].labelX, treeB->n[y].labelY,
				 treeA->nb_seq, treeB->nb_seq, pairedA, treeB->n[y].paired) 
		      + get_table(x-1, y-1);

		    update(w, 7, &ed);
		  }
		  
		  if(table[x][y] < ed.weight){
		    update_edge(&alignment_graph[x][y], ed);
		    table[x][y] = ed.weight;
		  }

		  /*** edge going from vertex (x,y) to vertex (i-1,j-2) ***/
		  if( (j>1) && is_subtree_m(treeA,i,x) && is_closedB[j-1][y] && (treeB->n[j-1].size == 1) ){
		    
		    ed.targetX = i-1;
		    ed.targetY = j-2; 
		    
		    /* arc-altering of x in j-1 */
		    ed.weight = score_sp(treeA->n[x].labelX, treeA->n[x].labelY, treeB->n[j-1].labelX, gapB,
					 treeA->nb_seq, treeB->nb_seq, treeA->n[x].paired, pairedB) 
		      + get_table(x-1,y);

		    ed.edit_operation = 4;

		    /* arc-breaking of arc x in (j-1,y)*/
		    if(treeB->n[y].size == 1 ) {
		      w = score_sp(treeA->n[x].labelX, treeA->n[x].labelY, treeB->n[j-1].labelX, treeB->n[y].labelX,
				   treeA->nb_seq, treeB->nb_seq, treeA->n[x].paired, pairedB) + get_table(x-1, y-1);
		      update(w, 2, &ed);
		    }
		    update_edge(&alignment_graph[x][y], ed);
		  }

		  /*** edge going from vertex (x,y) to vertex (i-2,j-1) ***/

		  if( (i>1) && is_subtree_m(treeB, j, y) && is_closedA[i-1][x]  && (treeA->n[i-1].size == 1) ){
		    
		    ed.targetX = i-2; 
		    ed.targetY = j-1;
		    
		    /* arc-altering of arc y in i-1 */
		    ed.weight = score_sp(treeA->n[i-1].labelX, gapA, treeB->n[y].labelX, treeB->n[y].labelY,
					 treeA->nb_seq, treeB->nb_seq, pairedA, treeB->n[y].paired) 
		      + get_table(x, y-1);

		    ed.edit_operation = 8;

		    /* arc-breaking of arc  y in (i-1,x)*/
		    if(treeA->n[x].size == 1){
		      w = score_sp(treeA->n[i-1].labelX, treeA->n[x].labelX, treeB->n[y].labelX, treeB->n[y].labelY,
				   treeA->nb_seq, treeB->nb_seq, pairedA, treeB->n[y].paired) + get_table(x-1, y-1);
		      update(w, 6, &ed);
		    }
 		    push_edge(&alignment_graph[x][y], ed);
		  }
		}
	      }
	    }
	  }/* end zone I */
	  
	  /*********************/
	  /*     zone II       */
	  /*********************/
	  
	  for(x=k+1; x<=m; x++){
	    if(is_closedA[i][x]){
	      for(y=j; y<=l; y++){
		if(is_closedB[j][y] && constraint[x][y] == ALLOWED ){
		  computeTable_Cell_m(treeA, treeB, x, y, i-1, j-1);
		  
		  /*** edge going from vertex (x,y) to vertex (i-1,j-1) ***/
		  
		  ed.targetX = i-1; 
		  ed.targetY = j-1;	      
		  ed.weight = table[x][y];
		  
		  /* arc-removing of y in B*/
		  if( is_subtree_m(treeB, j, y) ){
		    w = score_sp(gapA, gapA, treeB->n[y].labelX, treeB->n[y].labelY, 
				 treeA->nb_seq, treeB->nb_seq, pairedA, treeB->n[y].paired) + get_table(x, y-1);
		    update(w, 5, &ed);
		  }
		  
		  /* arc-altering of arc y in x */
		  if( (is_subtree_m(treeB, j, y)) && (treeA->n[x].size == 1) ){
		    w = score_sp(gapA, treeA->n[x].labelX, treeB->n[y].labelX, treeB->n[y].labelY,
				 treeA->nb_seq, treeB->nb_seq, pairedA, treeB->n[y].paired) 
		      + get_table(x-1, y-1);
		    
		    update(w, 7, &ed);
		  }
		  
		  if(ed.weight>table[x][y]){
		    table[x][y] = ed.weight;
		    update_edge(&alignment_graph[x][y], ed);
		  }
		  
		  /*** edge going from vertex (x,y) to vertex (i-2,j-1) ***/
		  if( (i>1) && is_subtree_m(treeB, j, y) && is_closedA[i-1][x]  && (treeA->n[i-1].size == 1) ){
		    
		    ed.targetX = i-2; 
		    ed.targetY = j-1;
		    
		    /* arc-altering of arc y in i-1 */	
		    ed.weight = score_sp(treeA->n[i-1].labelX, gapA, treeB->n[y].labelX, treeB->n[y].labelY, 
					 treeA->nb_seq, treeB->nb_seq, pairedA, treeB->n[y].paired) 
		      + get_table(x, y-1);

		    ed.edit_operation = 8;

		    /* arc-breaking of arc y in (i-1,x)*/
		    if(treeA->n[x].size == 1 ){
		      w = score_sp(treeA->n[i-1].labelX, treeA->n[x].labelX, treeB->n[y].labelX, treeB->n[y].labelY, 
				   treeA->nb_seq, treeB->nb_seq, pairedA, treeB->n[y].paired) + get_table(x-1, y-1);
		      update(w, 6, &ed);
		    }
		    push_edge(&alignment_graph[x][y], ed);
		  }
		}
	      }
	    }
	  }/* end zone II */
	  
	  /************************/
	  /*      zone III        */
	  /************************/
	  for(x=i; x<=k; x++) {
	    if( is_closedA[i][x]){
	      for(y=l+1; y<=n; y++){
		if(is_closedB[j][y] && constraint[x][y] == ALLOWED ){
		  computeTable_Cell_m(treeA, treeB, x, y, i-1, j-1);
		  
		  /*** edge going from vertex (x,y) to vertex (i-1,j-1) ***/
		  
		  ed.targetX = i-1; 
		  ed.targetY = j-1; 
		  ed.weight = table[x][y];
		  
		  /* arc-removing of arc x in A*/
		  if( is_subtree_m(treeA, i, x) ){
		    w = score_sp(treeA->n[x].labelX, treeA->n[x].labelY, gapB, gapB,
				 treeA->nb_seq, treeB->nb_seq, treeA->n[x].paired, pairedB) + get_table(x-1, y);
		    update(w, 1, &ed);
		  }
		  
		  /* arc-altering of arc x in y, 3' match */
		  if( (is_subtree_m(treeA, i, x)) && (treeB->n[y].size == 1) ) {
		    w = score_sp(treeA->n[x].labelX, treeA->n[x].labelY, gapB, treeB->n[y].labelX,
				 treeA->nb_seq, treeB->nb_seq, treeA->n[x].paired, pairedB) 
		      + get_table(x-1, y-1);

		    update(w, 3, &ed); 
		  }
		  
		  if(ed.weight > table[x][y]){
		    update_edge(&alignment_graph[x][y], ed);
		    table[x][y] = ed.weight;
		  }
		  
		  /*** edge going from vertex (x,y) to vertex (i-1,j-2) ***/
		  
		  if( (j > 1) &&  is_subtree_m(treeA, i, x) && is_closedB[j-1][y] && (treeB->n[j-1].size == 1) ){
		    
		    ed.targetX = i-1; 
		    ed.targetY = j-2;
		    
		    /* arc-altering of arc x in j-1, 5'match */
		    ed.weight = score_sp(treeA->n[x].labelX, treeA->n[x].labelY, treeB->n[j-1].labelX, gapB,
					 treeA->nb_seq, treeB->nb_seq, treeA->n[x].paired, pairedB) 
		      + get_table(x-1, y);

		    ed.edit_operation = 4;

		    /* arc-breaking of x in (j-1,y)*/		      
		    if(treeB->n[y].size == 1 ) {
		      w = score_sp(treeA->n[x].labelX, treeA->n[x].labelY, treeB->n[j-1].labelX, treeB->n[y].labelX,
				   treeA->nb_seq, treeB->nb_seq, treeA->n[x].paired, pairedB) + get_table(x-1, y-1);
		      update(w, 2, &ed); 
		    }
		    update_edge(&alignment_graph[x][y], ed);
		  }
		}
	      }
	    }
	  }/* end zone III */
	}
      } /* end node j in treeB */
    }
  }/* end node i in treeA */
  
  free(ancestorA);
  free(ancestorB);
  free(fatherA);
  free(fatherB);
  free(rightmostA);
  free(rightmostB);
  free(tree_closedA);
  free(tree_closedB);
  free(pairedA);
  free(pairedB);
  free(gapA);
  free(gapB);
}


static void build_optimal_path_m(postorder_tree_m *treeA, postorder_tree_m *treeB, vertex **alignment_graph, 
			       int i, int j, int k, int l, int positionA, int positionB, six_stack *p, int **supertree, int **table)
{
  int x,y, old_values; 
  int d; 
  int finished = 0;
  edge ed;   
   
  x = k; 
  y = l;
  
  
  while ( x+y > i+j ){

    finished = 0;
    /*******************/
    /* implicit edges  */
    /*******************/

    old_values = x+y; 

    if ( (x > i) && (y > j) && (treeA->n[x].size == 1) && (treeB->n[y].size == 1) && 
	 (table[x][y] == base_mismatch_m(treeA->n[x].labelX, treeB->n[y].labelX, treeA->nb_seq, treeB->nb_seq) 
	  + table[x-1][y-1]) ){
      supertree[treeA->n[x].positionX][treeB->n[y].positionX]=1;
      x--; y--; 
      finished = 1;
    }
    else if ( (x > i) && (table[x][y] == table[x-treeA->n[x].size][y] + full_deletionA[x]) ){
      for (d=x; d>x-treeA->n[x].size; d--){
	  supertree[treeA->n[d].positionX][max2(positionB,treeB->n[y].positionY)]=-1;
	  supertree[treeA->n[d].positionY][max2(positionB,treeB->n[y].positionY)]=-1;
      }
      x = x-treeA->n[x].size;
      finished = 1;
    }
    else if ( (y > j) && (table[x][y] == table[x][y-treeB->n[y].size] + full_deletionB[y]) ){
      for (d=y; d>y-treeB->n[y].size; d--){
	  supertree[max2(positionA,treeA->n[x].positionY)][treeB->n[d].positionX]=-2;
	  supertree[max2(positionA,treeA->n[x].positionY)][treeB->n[d].positionY]=-2;
	}
      y = y-treeB->n[y].size; 
      finished = 1;
    }
    
    /*******************/
    /* explicit edges  */
    /*******************/
    for (d=0;  (!finished) && (d <= alignment_graph[x][y].current_edge); d++){
      get_edge(&alignment_graph[x][y], &ed, d);
      if ( (ed.targetX >=i ) && (ed.targetY >= j) && (table[x][y] == ed.weight + get_table(ed.targetX, ed.targetY)) ){
	finished = 1;
	switch (ed.edit_operation) {
	case 0 :  /* arc_(mis)match */
	  supertree[treeA->n[x].positionX][treeB->n[y].positionX]=1; 
	  supertree[treeA->n[x].positionY][treeB->n[y].positionY]=1; 
	  six_push(p, ed.targetX, ed.targetY, treeA->n[x].positionX, treeB->n[y].positionX, x-1, y-1); 
	  break;
	case 5 :  /* arc_insertion */
	  supertree[max2(positionA,treeA->n[ed.targetX].positionY)][treeB->n[y].positionX]=-2;
	  supertree[treeA->n[x].positionY][treeB->n[y].positionY]=-2;
	  six_push(p, ed.targetX, ed.targetY, max2(positionA,treeA->n[ed.targetX].positionY),  treeB->n[y].positionX, x, y-1);
	  break; 
	case 1 :  /* arc_deletion */ 
	  supertree[treeA->n[x].positionX][max2(positionB,treeB->n[ed.targetY].positionY)]=-1;
	  supertree[treeA->n[x].positionY][treeB->n[y].positionY]=-1;
	  six_push(p, ed.targetX, ed.targetY, treeA->n[x].positionX, max2(positionB,treeB->n[ed.targetY].positionY), x-1, y);
	  break;
	case 6 : /* arc_breaking in B */
	  supertree[treeA->n[ed.targetX+1].positionX][treeB->n[y].positionX]=1;
	  supertree[treeA->n[x].positionX][treeB->n[y].positionY]=1;
	  six_push(p, ed.targetX+1, ed.targetY, treeA->n[ed.targetX+1].positionX,treeB->n[y].positionX,  x-1, y-1); 
	  break;
	case 2 : /* arc_breaking in A */
	  supertree[treeA->n[x].positionX][treeB->n[ed.targetY+1].positionX]=1;
	  supertree[treeA->n[x].positionY][treeB->n[y].positionX]=1;
	  six_push(p, ed.targetX, ed.targetY+1,treeA->n[x].positionX,treeB->n[ed.targetY+1].positionX, x-1, y-1);
	  break;
	case 3 :  /* arc_altering in A with 3' match */
	  supertree[treeA->n[x].positionY][treeB->n[y].positionX]=1;
	  supertree[treeA->n[x].positionX][max2(positionB,treeB->n[ed.targetY].positionY)]=-1;
	  six_push(p, ed.targetX, ed.targetY,treeA->n[x].positionX,max2(positionB,treeB->n[ed.targetY].positionX),  x-1, y-1);
	  break;
	case 7 : /* arc_altering in B with 3' match */
	  supertree[treeA->n[x].positionX][treeB->n[y].positionY]=1;
	  supertree[max2(positionA,treeA->n[ed.targetX].positionY)][treeB->n[y].positionX]=-2;
	  six_push(p, ed.targetX, ed.targetY, max2(positionA,treeA->n[ed.targetX].positionY),treeB->n[y].positionX, x-1, y-1); 
	  break;
	case 4 :  /* arc_altering in A with 5' match */
	  supertree[treeA->n[x].positionX][treeB->n[ed.targetY+1].positionX]=1; 
	  supertree[treeA->n[x].positionY][treeB->n[y].positionY]=-1;
	  six_push(p, ed.targetX, ed.targetY+1,treeA->n[x].positionX,max2(positionB,treeB->n[ed.targetY+1].positionX), x-1, y); 
	  break;
	case 8 : /* arc_altering in B with 5' match */
	  supertree[treeA->n[ed.targetX+1].positionX][treeB->n[y].positionX]=1;
	  supertree[treeA->n[x].positionY][treeB->n[y].positionY]=-2; 
	  six_push(p, ed.targetX+1, ed.targetY,max2(positionA,treeA->n[ed.targetX+1].positionX),treeB->n[y].positionX, x, y-1);
	  break;
	default :
	  printf("error \n");fflush(stdout);
	  break;
	} /* end of switch */
	x = ed.targetX; 
	y = ed.targetY;
      }
    }
    
    if( old_values == x+y ){
      printf("alignment is not possible \n");
      exit(0);
    }
  }
}

/* Construction of the optimal mapping: 
   the mapping is a set of pairs of nodes of treeA x treeB. 
   It is stored in a stack. */
static int trace_back_alignment_mapping_m(postorder_tree_m *treeA, postorder_tree_m *treeB, vertex **alignment_graph,
					int ** supertree, int **table)
{
  int score = 0;
  int i, j, k, l,positionA, positionB;
  six_stack *p;
  int old_i = 0;
  int old_j = 0;
  
  p = malloc(sizeof(six_stack));
  init_six_stack(p, (treeA->size + treeB->size) * 2);
  computeTable_m(treeA, treeB,  0, 0,treeA->size-1, treeB->size - 1);
  build_optimal_path_m(treeA, treeB, alignment_graph, 0, 0,treeA->size - 1, treeB->size - 1, 0,0,p, supertree, table);
  score = table[treeA->size - 1][treeB->size - 1];
  while(!is_empty_six_stack(p)){
    six_pop(p, &i, &j, &positionA, &positionB, &k, &l);
    if ( (i != old_i) || (j != old_j) ){
      old_i = i; old_j = j;
      computeTable_m(treeA, treeB, i, j, k, l);
    }
    build_optimal_path_m(treeA, treeB, alignment_graph, i, j, k,l,positionA, positionB, p, supertree, table);
  }
  free_six_stack(p);
  return score;
}

static int optimal_arc_alignment_m(RNA_sequence_m *rnaA, RNA_sequence_m *rnaB,  RNA_sequence_m *align,
				   int do_mapping, int EQ, int SIMILAR){
  int i, j;
  int score;
  int ** supertree;
  postorder_tree_m *treeA = NULL;
  postorder_tree_m *treeB = NULL;

  treeA = RNAsequence2postorder_m(rnaA);
  treeB = RNAsequence2postorder_m(rnaB);

  constraint = (char **)calloc(treeA->size, sizeof(char *) );
  for(i=0; i<treeA->size; i++)
    constraint[i] = (char *)calloc( treeB->size, sizeof(char));

  table = (int **)calloc( treeA->size, sizeof(int *) );
  for(i=0; i<treeA->size; i++)
    table[i] = (int *)calloc( treeB->size, sizeof(int) );
  
  alignment_graph= (vertex**)malloc(sizeof(vertex *) * treeA->size);
  for(i=0; i<treeA->size; i++)
    alignment_graph[i] = (vertex *)malloc(sizeof(vertex) * treeB->size);
  
  full_deletionA = (int*)calloc( treeA->size, sizeof(int) );
  full_deletionB = (int*)calloc( treeB->size, sizeof(int) );
  /* deletions in A */
  for(i=1; i<treeA->size; i++)
    full_deletionA[i] = tree_full_deletion_m(treeA, i, treeB->nb_seq);
  /* insertions in B */
  for(i=1; i<treeB->size; i++)
    full_deletionB[i] = tree_full_deletion_m(treeB, i,treeA->nb_seq);

  no_constraint(constraint, treeA->size, treeB->size);

  if( EQ ){
    full_constraint(constraint, treeA->size, treeB->size);
    eq_constraint(constraint, treeA, treeB, rnaA->eq, rnaB->eq, rnaA->s[0].size, rnaB->s[0].size, 1);
  }

  if( SIMILAR )
    diagonal_constraint(constraint, treeA->size, treeB->size, (int)((double)( rnaA->s[0].size)/100 * 30) );
  

  build_alignment_graph_m(treeA, treeB);

  if(do_mapping){
    supertree = (int **)malloc( sizeof(int*)*(rnaA->s[0].size+1) );
    for (i=0; i<=rnaA->s[0].size; i++)
      supertree[i]= (int *)malloc( sizeof(int)*(rnaB->s[0].size+1) );
    for (i=0; i<=rnaA->s[0].size; i++)
       for (j=0; j<=rnaB->s[0].size; j++)
	 supertree[i][j]=0;
    score = trace_back_alignment_mapping_m(treeA, treeB, alignment_graph, supertree, table);
    create_alignment_RNA_sequences_m(rnaA, rnaB, supertree, align, 1, rnaA->s[0].size-1, 1, rnaB->s[0].size-1);
  
  }
  else{
    computeTable_m(treeA, treeB, 0, 0, treeA->size-1, treeB->size-1);
    score = table[treeA->size-1][treeB->size-1];
  }

  for(i=0; i<treeA->size; i++){ 
    free(table[i]);
    free(constraint[i]);
  }

  free(table);
  free(constraint);
  free(full_deletionA);
  free(full_deletionB);

  for(i=1; i<treeA->size; i++){
    for(j=1; j<treeB->size; j++)
      free(alignment_graph[i][j].elements);
    free(alignment_graph[i]);
  }
  free(alignment_graph[0]);
  free(alignment_graph);
  
  for(i=1; i<treeA->size; i++)
    free(is_closedA[i]);
  free(is_closedA);

  for(i=1; i<treeB->size; i++)
    free(is_closedB[i]);
  free(is_closedB);

  free_postorder_tree_m(treeA);
  free_postorder_tree_m(treeB);
  free(treeA);
  free(treeB);

  return score;
}

int optimal_multiple_arc_alignment_mapping(RNA_sequence_m *rnaA, RNA_sequence_m *rnaB,  
				    RNA_sequence_m *align, int EQ, int SIMILAR){
  return optimal_arc_alignment_m(rnaA, rnaB, align, 1, EQ, SIMILAR);
}

int optimal_multiple_arc_alignment_score(RNA_sequence_m *rnaA, RNA_sequence_m *rnaB, 
				  int EQ, int SIMILAR){
  RNA_sequence_m *align = NULL; 
  return optimal_arc_alignment_m(rnaA, rnaB, align, 0, EQ, SIMILAR);
}

void multiple_alignment(RNA_sequence_m **list, int nb_seq, int EQ, int SIMILAR, 
			int FASTA, int CLUSTAL, char *nameFileOut ){

  RNA_sequence_m *alignment;
  int score=0, i, j;
  int **tab;
  int max_length=0;
  
  for (i=0; i<nb_seq;i++){
    max_length+=list[i]->s[0].size; 
  }
  
  tab = (int **)malloc( sizeof(int *) * nb_seq );
  for(i=0; i<nb_seq; i++)
    tab[i] = (int *)malloc( sizeof(int) * (i) );
  
  for(i=0; i < nb_seq; i++)
    for(j=0; j < i; j++)
      tab[i][j] = optimal_multiple_arc_alignment_score(list[i], list[j], EQ, SIMILAR);

  alignment = clustering(tab, nb_seq, list, &score, EQ, SIMILAR);
  
  printf("\nScore : %d", score);  
  /*printf("\nDensity|Covariations: %d\t %d\n", alignment_density(max_length, nb_seq, alignment), covariations(alignment)); */

  if( !FASTA )
    if( !CLUSTAL )
      display_alignment_m(alignment, nameFileOut);
    else
      clustal_w_output(alignment, nameFileOut);
  else
      multi_fasta(alignment, nameFileOut);
  
  /* free memory */
  for(i=0; i<nb_seq; i++)
    free(tab[i]);
  free(tab);
  free_RNA_sequence_m(alignment);
  free(alignment);
}
