/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or */
/* (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "../headers/score.h"

int **paired_matrix;
int **unpaired_matrix;
int use_matrix_score;

void free_matrices(){

  int i;

  for(i=0; i<PAIRED_SIZE; i++){
    free(paired_matrix[i]);
    if(i<UNPAIRED_SIZE)
      free(unpaired_matrix[i]);
  }
  free(paired_matrix);
  free(unpaired_matrix);
}

void get_score(char * score_file,char * nameFileMatrix, int* BASE_MAT, int* BASE_MIS, int* BASE_DEL, int *ARC_MAT, int * ARC_MISMAT, int* ARC_MIS, int* ARC_REM, int* ARC_ALT_MAT, int* ARC_ALT_MIS, int* ARC_BRE_MAT,int* ARC_BRE_MISMAT,  int* ARC_BRE_MIS, int *GAP_OPEN, int USE_MATRIX_FILE, int ARC, int FULL_EDIT){
  
  FILE * fp;
  *GAP_OPEN = 0;
  
  fp=fopen(score_file, "rt");

  if (USE_MATRIX_FILE){
    get_matrices(nameFileMatrix);
    fp=fopen(score_file, "rt");
    if( !fp ){
      printf("Warning: Invalid file name (%s). ", score_file);
      printf("Using default scoring system.\n");
      fp=fopen("matrices.txt","rt");
    }
    fscanf(fp, "%d", BASE_DEL);
    fscanf(fp, "%d", ARC_REM);
    if (ARC==1){
      fscanf(fp, "%d", ARC_ALT_MAT); 
      fscanf(fp, "%d", ARC_BRE_MAT); 
      fscanf(fp, "%d", GAP_OPEN); 
    }
  }  
  else{
    fp=fopen(score_file, "rt");
    if( !fp ){
      printf("Warning: Invalid file name (%s). ", score_file);
      printf("Using default scoring system.\n");
      fp=fopen("score.txt","rt");
    }
    fscanf(fp, "%d", BASE_MAT); 
    fscanf(fp, "%d", BASE_MIS);  
    fscanf(fp, "%d", BASE_DEL);
    fscanf(fp, "%d", ARC_MAT);
    fscanf(fp, "%d", ARC_MISMAT);
    fscanf(fp, "%d", ARC_MIS);
    fscanf(fp, "%d", ARC_REM);
    if (ARC==1){
      fscanf(fp, "%d", ARC_ALT_MAT); 
      fscanf(fp, "%d", ARC_ALT_MIS); 
      fscanf(fp, "%d", ARC_BRE_MAT); 
      fscanf(fp, "%d", ARC_BRE_MISMAT);
      fscanf(fp, "%d", ARC_BRE_MIS);
      fscanf(fp, "%d", GAP_OPEN); 
    }
  }  
  
  fclose(fp);
  if (FULL_EDIT){
    *ARC_REM=*ARC_REM-2*(*BASE_DEL);
    *ARC_MAT=*ARC_MAT-2*(*BASE_MAT);
  }  
  use_matrix_score = USE_MATRIX_FILE;
}


void print_score(char * nameFileMatrix, int base_mat, int base_mis, int base_del, int arc_mat, int arc_mismat, int arc_mis, int arc_rem, int arc_alt_mat, int arc_alt_mis, int arc_bre_mat, int arc_bre_mismat, int arc_bre_mis, int gap_open, int USE_MATRIX_SCORE, int ARC, int FULL_EDIT){

 printf("Scoring system: ");
 if (USE_MATRIX_SCORE){
   if (FULL_EDIT)
     printf (" %s with base-deletion %d, arc-removing %d\n", nameFileMatrix, base_del, arc_rem);
   else if ((ARC)&&(gap_open==0))
     printf (" %s with base-deletion %d, arc-removing %d, arc-breaking penalty %d, arc-altering penalty %d\n", nameFileMatrix, base_del, arc_rem, arc_bre_mat, arc_alt_mat);
   else if (ARC)
      printf (" %s with base-deletion %d, arc-removing %d, arc-breaking penalty %d, arc-altering penalty %d, module creation %d\n", nameFileMatrix, base_del, arc_rem, arc_bre_mat, arc_alt_mat, gap_open);
   else
     printf (" %s with base-deletion %d, arc-removing %d\n", nameFileMatrix, base_del, arc_rem);
 }

 else{
   if (FULL_EDIT)
     printf("base-match %d, base-mismatch %d, base-deletion %d, arc-match %d, arc-mismatch(1) %d, arc-mismatch(2) %d, arc-removing %d, arc-altering(1) %d, arc-altering(2) %d, arc-breaking(1) %d, arc-breaking(2) %d, arc-breaking(3) %d\n", base_mat, base_mis, base_del, arc_mat+2*base_mat, arc_mat+base_mat+base_mis, arc_mat+2*base_mis, arc_rem+2*base_del, arc_rem+base_del+base_mat, arc_rem+base_del+base_mis, arc_rem+2*base_mat, arc_rem+base_mat+base_mis, arc_rem+2*base_mis);
   else if ((ARC) && (gap_open==0))
     printf("base-match %d, base-mismatch %d, base-deletion %d, arc-match %d, arc-mismatch(1) %d, arc-mismatch(2) %d, arc-removing %d, arc-altering(1) %d, arc-altering(2) %d, arc-breaking(1) %d, arc-breaking(2) %d, arc-breaking(3) %d\n", base_mat, base_mis, base_del, arc_mat, arc_mismat, arc_mis, arc_rem, arc_alt_mat, arc_alt_mis, arc_bre_mat, arc_bre_mismat, arc_bre_mis);
   else if (ARC)
     printf("base-match %d, base-mismatch %d, base-deletion %d, arc-match %d, arc-mismatch(1) %d, arc-mismatch(2) %d, arc-removing %d, arc-altering(1) %d, arc-altering(2) %d, arc-breaking(1) %d, arc-breaking(2) %d, arc-breaking(3) %d, module creation %d\n", base_mat, base_mis, base_del, arc_mat, arc_mismat, arc_mis, arc_rem, arc_alt_mat, arc_alt_mis, arc_bre_mat, arc_bre_mismat, arc_bre_mis, gap_open);
   else
     printf("base-match %d, base-mismatch %d, base-deletion %d, arc-match %d, arc-mismatch(1) %d, arc-mismatch(2) %d, arc-removing %d\n", base_mat, base_mis, base_del, arc_mat, arc_mismat, arc_mis, arc_rem);
 }  
 printf("\n");

}  



int parse_double(char *cp){
  
  int precision = 2;
  int x = 0;
  int minus = 1;
  int entiere = 0 , decimale = 0;
  int i, k;
  int puis = 0;
  
  if( cp[0] == '-'){
    minus = -1;
    cp = cp + 1;
  }
   
  puis = 0;
  while( cp[puis] != '.' )
    puis++;
  
  k = 0;
  for(i=puis-1; i>=0; i--){
    entiere = entiere + pow(10, i) * ((int)cp[k++] - 48);
  }

  cp = cp + puis + 1;
  
  k = 0;
  for(i=precision-1; i>=0; i--)
    decimale = decimale + pow(10, i) * ((int)cp[k++] - 48);

  x = entiere * 100 + decimale;
  x = x * minus;

  cp = cp + precision;

  return x;
}


void get_matrices(char *input_file){

  int k=0;
  int l=0;
  int i;
  FILE *fp;
  char linebuf[256];
  char fullbuf[16384];
  int fullbuf_used = 0;
  char *cp, *end_mat_pos;

  paired_matrix = (int **)malloc( sizeof(int *) * PAIRED_SIZE);
  for(i=0;i<PAIRED_SIZE; i++)
    paired_matrix[i] = (int *)malloc( sizeof(int) * (i+1) );

  unpaired_matrix = (int **)malloc( sizeof(int *) * UNPAIRED_SIZE);
  for(i=0;i<UNPAIRED_SIZE; i++)
    unpaired_matrix[i] = (int *)malloc( sizeof(int) * (i+1) );

  fp = fopen(input_file, "r");
  if(!fp){ 
    printf("invalid file: %s", input_file); 
    exit(0);
  }
  
  while(fgets (linebuf, 255, fp)) {
    strncpy(fullbuf+fullbuf_used, linebuf, 16384-fullbuf_used-1);
    fullbuf_used += strlen(linebuf);
    if(fullbuf_used >= 16384) {
      printf("ERROR: Matrix file bigger than 16kb\n");
      exit(0);
    }
  }

  /*-------- UNPAIRED MATRIX -----------*/
  /* First, find RIBO */
  cp = strstr (fullbuf, "RIBO");
  for(i = 0; cp[i] && !isspace(cp[i]); i++);   /* Find space after RIBO */
  cp = cp + i;

  /* Now, find the first A */
  cp = strchr (cp, 'A');
  while(*cp != '\n' && cp-fullbuf < fullbuf_used) cp++;
  
  /* Find next A */
  while(*cp != 'A' && (cp-fullbuf) < fullbuf_used) cp++;

  /* Take numbers until we hit the H: */
  end_mat_pos = strstr (cp, "H:");
  for(i=0; cp - fullbuf < end_mat_pos-fullbuf; i++) {
    while(!isdigit(*cp) && *cp != '-' && *cp != '.' &&
	   cp-fullbuf < fullbuf_used && cp != end_mat_pos) { 
      cp++;
    }
    if(cp == end_mat_pos)
      break;
    if(cp-fullbuf < fullbuf_used) {
      unpaired_matrix[k][l] = parse_double(cp);
      if(k == l){
	l=0;
	k++;
      }
      else
	l++;

      while((isdigit (*cp) || *cp == '-' || *cp == '.') &&
	     (cp-fullbuf <fullbuf_used)) {
	cp++;
      }
    }
  }
  /* Skip the H: */
  cp += 2;
  /* Now, go past the E: */
  cp = strstr (cp, "E:") + 2;

  /*-------- PAIRED MATRIX -----------*/
  k=0;
  l=0;
  /* Now, find the first A */
  cp = strchr(cp, 'A');
  while(*cp != '\n') cp++;
  /* Find next A */
  while(*cp != 'A' && (cp-fullbuf) < fullbuf_used) cp++;

  /* Take numbers until we hit the H: */
  end_mat_pos = strstr (cp, "H:");
  for(i=0; cp - fullbuf < end_mat_pos-fullbuf; i++) {
    while(!isdigit(*cp) && *cp != '-' && *cp != '.' &&
	   cp-fullbuf < fullbuf_used && cp != end_mat_pos) { 
      cp++;
    }
    if(cp == end_mat_pos)
      break;
    if(cp-fullbuf < fullbuf_used) {
      paired_matrix[k][l] = parse_double(cp);
      if(k == l){
	l=0;
	k++;
      }
      else
	l++;

      while((isdigit (*cp) || *cp == '-' || *cp == '.') &&
	     (cp-fullbuf <fullbuf_used)) {
	cp++;
      }
    }
  }
  fclose(fp);
}


/*score for a substitution */
int substitution(postorder_tree * treeA,postorder_tree * treeB,int i,int j){

  if(treeA->n[i].labelX==treeB->n[j].labelX) /* the two first elements are identical */
    if (treeA->n[i].labelY==0)
      if(treeB->n[j].labelY==0) return  ID; /* base-match*/ 
      else return sub_leaf_node; 
    else /* i is not a single base */
      if (treeB->n[j].labelY==0) return sub_leaf_node;
      else if (treeA->n[i].labelY==treeB->n[j].labelY) return ID /* arc-match*/;
      else return SUB ;
  else /* the two first elements are different */
    if ((treeA->n[i].labelX=='X')&& (treeB->n[j].labelX!='X')) return sub_leaf_node;
 else
   if (treeA->n[i].labelY==0) 
     if (treeB->n[j].labelY==0) return SUB;
     else return sub_leaf_node;
   else if (treeB->n[j].labelY==0) return sub_leaf_node;
   else return SUB; 
}


int arc_removing(char a, char b){
  return arc_rem;
}

int base_mismatch(char a, char b){

  int a1 = a;
  int b1 = b;

  /*==== Score RIBOSUM ====*/
  if(use_matrix_score){
    if(a1>=b1){
      return unpaired_matrix[a1][b1];
    }
    else
      return unpaired_matrix[b1][a1];
  }

 /*=== Other Score ===*/
  if (a==b) 
    return base_mat ; /* base-match */
  else 
    return base_mis; /*base-mismatch*/
}

int base_deletion(char a){
  return base_del;
}

int arc_mismatch(char a, char b, char c, char d){
 
  int a1 = a;
  int b1 = b;
  int c1 = c;
  int d1 = d;


  /*==== Score RIBOSUM ====*/
  if(use_matrix_score){
    if (a=='X') return 20;

    if(4*a1+b1 >= 4*c1+d1)
      return paired_matrix[4*a1 + b1][4*c1 + d1];
    else
      return paired_matrix[4*c1 + d1][4*a1 + b1];
  }

 /*=== Other Score ===*/ 
 if (a==c){
   if (b==d)
     return arc_mat; 
   else 
     return arc_mismat;
  }
  else {
    if (b==d) 
      return arc_mismat;
    else 
      return arc_mis;
  }
}


int arc_breaking(char a, char b, char c, char d){
 
 /*==== Score RIBOSUM ====*/
  if(use_matrix_score){
    return arc_mismatch(a,b,c,d)+arc_bre_mat;
  }

 /*=== Other Score ===*/
 if (a==c) {
    if (b==d)
      return arc_bre_mat; 
    else 
      return arc_bre_mismat;
  }
  else {
    if (b==d)
      return arc_bre_mismat;
    else
      return arc_bre_mis;
  }
}

int arc_altering1(char a, char b, char c){

   /*==== Score RIBOSUM ====*/
  if(use_matrix_score){
    return base_mismatch(a,c)+arc_alt_mat;
  }

  /*=== Other Score ===*/ 
 if (a==c)
    return arc_alt_mat; 
  else 
    return arc_alt_mis;
} 

int  arc_altering2(char a, char b, char c){
 
  /*==== Score RIBOSUM ====*/
  if(use_matrix_score){
    return base_mismatch(b,c)+arc_alt_mat;
  }

 /*=== Other Score ===*/
 if (b==c)
   return arc_alt_mat; 
 else 
  return arc_alt_mis; 
} 

/* score for a deletion or an insertion */
int indel(postorder_tree * tree, int i){
  if (tree->n[i].size==1)
    return base_deletion(tree->n[i].labelX);
  else
    return arc_removing(tree->n[i].labelX, tree->n[i].labelY);
}

