/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or */
/* (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "../headers/score_m.h"

int base_mismatch_m(char *a, char *b, int nb_seqA, int nb_seqB){
  int i, j, val = 0;
  int a1, b1;

  assert( a != NULL && b != NULL );

  for(i=0; i<nb_seqA; i++){
    for(j=i+1; j<nb_seqA; j++){
      a1 = a[i];
      b1 = a[j];
      if( a1 != 4 && b1 != 4 )
	val += base_mismatch(a[i], a[j]);
      else if( a1 != 4 && b1 == 4 )
	val += base_deletion( a[i] );
      else if( a1 == 4 && b1 != 4 )
	val += base_deletion( a[j] );
    }
    for(j=0; j<nb_seqB; j++){
      a1 = a[i];
      b1 = b[j];
      if( a1 != 4 && b1 != 4 )
	val += base_mismatch(a[i], b[j]);
      else if( a1 != 4 && b1 == 4 )
	val += base_deletion( a[i] );
      else if( a1 == 4 && b1 != 4 )
	val += base_deletion( b[j] );
    }
  }

  for(i=0; i<nb_seqB; i++){
    for(j=i+1; j<nb_seqB; j++){
      a1 = b[i];
      b1 = b[j];
      if( a1 != 4 && b1 != 4 )
	val += base_mismatch(b[i], b[j]);
      else if( a1 != 4 && b1 == 4 )
	val += base_deletion( b[i] );
      else if( a1 == 4 && b1 != 4 )
	val += base_deletion( b[j] );
    }
  }
    
  return val;
}

int base_deletion_m(char *a, int nb_seqA, int nb_seqB){
  int i, j, val = 0;
  int a1, a2;

  assert( a!= NULL );

  for(i=0; i<nb_seqA; i++){
    a1 = a[i];
    for(j=i+1; j<nb_seqA; j++){
      a2 = a[j];
      if( a1 != 4 && a2 != 4)
	val += base_mismatch(a[i], a[j]);

      else if( a1 != 4 && a2 == 4)
	val += base_deletion(a[i]);

      else if( a1 == 4 && a2 != 4)
	val += base_deletion(a[j]);
    }

    if( a1 != 4)
      val += base_deletion(a[i]) * nb_seqB;    
  }
  
  return val;
}

static int base_operation(char a1, char a2){

  int a1i = a1;
  int a2i = a2;

  if( a1i != 4 && a2i != 4 )
    return base_mismatch(a1, a2);
  else if( a1i != 4 && a2i == 4 )
    return base_deletion( a1 );
  else if( a1i == 4 && a2i != 4 )
    return base_deletion( a2 );
  else
    return 0;

}

/* pair a1,b1 of tree A against pair a2,b2 of tree B */
static int operation(char a1, char b1, char a2, char b2, int pairedA, int pairedB){
  int a1i = a1;
  int a2i = a2;
  int b1i = b1;
  int b2i = b2;

  /* arc removing */
  if( a1i != 4 && b1i != 4 && 
      a2i == 4 && b2i == 4 &&
      pairedA == 1 && pairedB == 0)
    return arc_removing(a1, b1);
  
  else if( a1i == 4 && b1i == 4 && 
	   a2i != 4 && b2i != 4 &&
	   pairedB == 1 && pairedA == 0)
    return arc_removing(a2, b2);

  /* arc mismatch */
  else if( a1i != 4 && b1i != 4 && 
	   a2i != 4 && b2i != 4 &&
	   pairedB == 1 && pairedA == 1)
    return arc_mismatch(a1, b1, a2, b2);

  /* arc altering */
  else if( a1i != 4 && b1i != 4 && 
	   a2i == 4 && b2i != 4 &&
	   pairedA == 1)
    return arc_altering2(a1, b1, b2);
    
  else if( a1i != 4 && b1i != 4 && 
	   a2i != 4 && b2i == 4 &&
	   pairedA == 1)
    return arc_altering1(a1, b1, a2);
  
  else if( a1i == 4 && b1i != 4 && 
	   a2i != 4 && b2i != 4 &&
	   pairedB == 1)
    return arc_altering2(a2, b2, b1);
  
  else if( a1i != 4 && b1i == 4 && 
	   a2i != 4 && b2i != 4 &&
	   pairedB == 1)
    return arc_altering1(a2, b2, a1);
  
  /* arc breaking */
  else if( a1i != 4 && b1i != 4 && a2i != 4 && b2i != 4 && 
	   ( (pairedA == 1 && pairedB == 0) || (pairedA == 0 && pairedB == 1) ) )
    return arc_breaking(a1, b1, a2, b2);
  
/*   else if( a1i != 4 && b1i != 4 && a2i != 4 && b2i != 4 && */
/* 	   pairedA == 0 && pairedB == 1 ) */
/*     return arc_breaking(a1, b1, a2, b2); */

  /* base deletion */
  else if( a1i != 4 && b1i != 4 &&
	   a2i == 4 && b2i == 4 &&
	   pairedA == 0)
    return base_deletion(a1) + base_deletion(b1);

  else if( a1i == 4 && b1i == 4 &&
	   a2i != 4 && b2i != 4 &&
      pairedB == 0)
    return base_deletion(a2) + base_deletion(b2);

  else if( a1i == 4 && a2i == 4 && b1i == 4 && b2i == 4 )
    return 0;

  else if( pairedA == 0 && pairedB == 0)
    return  base_operation(a1, a2) + base_operation(b1, b2);

  printf("%i:%i => %c, %c, %c, %c\n", pairedA, pairedB, a1, a2, b1, b2);
  return 0;
}

/*
 * a1: column X of the first node 
 * b1: column Y of the first node
 *
 * a2: column X of the second node
 * b2: column Y of the second node
 */
int score_sp(char *a1, char *b1, char *a2, char *b2, int nb_seqA, int nb_seqB, int *pairedA, int *pairedB){

  int i, j, val = 0;

  assert( a1 != NULL && b1 != NULL);

  for(i=0; i<nb_seqA; i++){
    for(j=i+1; j<nb_seqA; j++){
      val += operation(a1[i], b1[i], a1[j], b1[j], pairedA[i], pairedA[j]);
    }

    if( a2 != NULL && b2 != NULL){
      for(j=0; j<nb_seqB; j++){
	val += operation(a1[i], b1[i], a2[j], b2[j], pairedA[i], pairedB[j]);
      }
    }
  }

  if( a2 != NULL && b2 != NULL){
    for(i=0; i<nb_seqB; i++)
      for(j=i+1; j<nb_seqB; j++){
	val += operation(a2[i], b2[i], a2[j], b2[j], pairedB[i], pairedB[j]);
      }
  }

  return val;
}
