/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or */
/* (at your option) any later version. */

/* This program is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/* GNU General Public License for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "../headers/score.h"
#include "../headers/stack.h"
#include "../headers/RNA_structure.h"
#include "../headers/tree_alignment.h"
#include "../headers/graph.h"
#include "../headers/matrix.h"
#include "../headers/constraint.h"

static vertex** alignment_graph;
static int ** table;
static int * full_deletionA, * full_deletionB;
static char ** is_closedA;
static char ** is_closedB; 
static char ** constraint=NULL;




/* right_sibling[i]=x : x is the right sibling of  i, if it exists, -1 otherwise 
 */
static void right_sibling(postorder_tree * tree, int* tab){
 int i ;  
 for (i=1;i<tree->size; i++){
   tab[i]=-1;
 }
 for (i=1;i<tree->size; i++){
   tab[i-tree->n[i].size]=i;
 }
}


/* if no father : -1 */
static void father(postorder_tree * tree, int * tab ){
  int i ,j; 
  for (i=1;i<tree->size; i++){
    tab[i]=-1;
  }
 for (i=1;i<tree->size; i++){
   if (tree->n[i].size>1) 
     for (j=i-1; j>i-tree->n[i].size; j=j-tree->n[j].size)
       tab[j]=i;
 }
}

/* number of closed subforests ending at position i, 
 * including i 
 */
static void tree_closed(postorder_tree * tree, short int * tab_closed){
  int i; 
  int j; 
  int * fatherT; 
  fatherT=(int*)malloc(sizeof(int)*tree->size); 
  father(tree,fatherT);
  
  for (i=1; i<tree->size;i++){
    j=i-tree->n[i].size;
    if ((j>0) && ((fatherT[i]==-1) || (fatherT[i]==fatherT[j])))
      tab_closed[i]=tab_closed[j]+1;
    else 
      tab_closed[i]=1;
  }
}

static void is_closed_forest(char ** tab, postorder_tree * tree){
  int x=0;
  int * child; 
  int y, z,t;
  int i,j;
  
  child=(int* ) malloc(sizeof(int)*tree->size);

  for (i=1; i<tree->size; i++)
    for (j=1; j<tree->size; j++)
      tab[i][j]=0;

  for (i=1; i<tree->size; i++){
    if (tree->n[i].size>1){ /* internal node of tree */
      x=0;
      for (j=i-1;  j>i-tree->n[i].size; j=j-tree->n[j].size){ /* looking for children */
	child[x]=j; x++;
      }
      for (j=0; j<x;j++){
	t=child[j];
	for (y=0;y<=j; y++){
	  z=child[y];
	  tab[t-tree->n[t].size+1][z]=1;
	  tab[z][t-tree->n[t].size+1]=1;
	}
      }
    }/* end internal node */
  } 
  x=0;
 
  for (j=tree->size-1;  j>0; j=j-tree->n[j].size){
    child[x]=j; x++;
   }
  for (j=0; j<x;j++){
    t=child[j];
    for (y=0;y<=j; y++){
      z=child[y];
      tab[t-tree->n[t].size+1][z]=1;
      tab[z][t-tree->n[t].size+1]=1;
    }
  }  
  free(child);
}


/* rightmost_sibling[i]=x : x is the rightmost sibling of i, if it exists, 
 * i otherwise */
static void rightmost_sibling(postorder_tree * tree, int* tab){
  int * tab2;
  int i; 
  tab2=(int *) malloc(sizeof(int)*tree->size);
  right_sibling(tree, tab2);
  for (i=tree->size-1; i>0; i--){
    if (tab2[i]!=-1) 
      tab[i]=tab[tab2[i]];
    else 
      tab[i]=i;
  } 
  free(tab2);
}


static int tree_full_deletion(postorder_tree * tree, int i){
  int w=0;
  int x;
  for (x=i; x>i-tree->n[i].size; x--)
    if (tree->n[x].size==1)
      w+=base_deletion(tree->n[x].labelX); 
    else
      w+=arc_removing(tree->n[x].labelX, tree->n[x].labelY);
  return w; 
}


static int get_table(int x, int y){
     if (constraint[x][y]==ALLOWED)
    return table[x][y];
   return -1000000;
}

static void computeTable_Cell(postorder_tree * treeA, postorder_tree * treeB,int x,int y, int i, int j){
  int  d, max_d,w;
  edge ed;
  int w_max=-10000;
  
  /*** we look at all edges coming from (x,y) ***/
  
  /* implicit edges  */
  
  /* full deletion */
  if (x>i)
    w_max=get_table(x-treeA->n[x].size,y)+full_deletionA[x];
  /* full insertion */
  if (y>j){
    w=get_table(x,y-treeB->n[y].size)+full_deletionB[y];
    if (w>w_max) {w_max=w;}
  }

  if ((treeA->n[x].size==1) && (treeB->n[y].size==1)){
    /* substitutions entre bases libres */
    if ((x>i)&& (y>j)) {  
      w=get_table(x-1,y-1) + base_mismatch(treeA->n[x].labelX,treeB->n[y].labelX);
      if (w>w_max) {w_max=w;}
    }
  }
  else{
    /* explicit edges  */
    max_d = alignment_graph[x][y].current_edge +1;
    for (d=0; d<max_d; d++){
      get_edge(&alignment_graph[x][y], &ed, d);
      if ((ed.targetX>=i) && (ed.targetY>=j)){
	w=ed.weight+get_table(ed.targetX,ed.targetY);
	if (w>w_max) {w_max=w;} 
      }
    }
  }
  
  table[x][y]=w_max;
}


static void computeTable(postorder_tree * treeA, postorder_tree * treeB,int i,int j, int k, int l){
  int x,y;
  table[i][j]=0;
  
  for (x=i+1; x<=k; x++)
     if  (constraint[x][j]==ALLOWED) table[x][j]=full_deletionA[x] + get_table(x-treeA->n[x].size,j); /*a revoir */
  for (y=j+1; y<=l; y++)
     if (constraint[i][y]==ALLOWED) table[i][y]=full_deletionB[y] + get_table(i,y-treeB->n[y].size); /*a revoir */

  for(x=i+1; x<=k; x++)
    if(is_closedA[i+1][x])
      for(y=j+1; y<=l; y++)
	if(is_closedB[j+1][y])
	  if (constraint[x][y]==ALLOWED)
	  computeTable_Cell(treeA,treeB,x,y,i,j);
}


static void build_alignment_graph(postorder_tree * treeA, postorder_tree * treeB){
  int i,j,k,l,m,n,x,y;
  short int * tree_closedA;
  short int * tree_closedB; 
  int * rightmostA;
  int * rightmostB;
  int * ancestorA;
  int * ancestorB;
  int * fatherA;
  int * fatherB;
  edge ed; 
  int w;
  int degree=0; 
  
  
  tree_closedA=(short int*)malloc(sizeof(short int)*treeA->size);
  tree_closedB=(short int*)malloc(sizeof(short int)*treeB->size); 
  tree_closed(treeA,tree_closedA);
  tree_closed(treeB,tree_closedB);
  
  ancestorA=(int*)malloc(sizeof(int)*treeA->size);
  ancestorB=(int*)malloc(sizeof(int)*treeB->size); 
  ancestor(treeA,ancestorA);
  ancestor(treeB,ancestorB);
 
  rightmostA=(int*)malloc(sizeof(int)*treeA->size); 
  rightmostB=(int*)malloc(sizeof(int)*treeB->size);
  rightmost_sibling(treeA,rightmostA); 
  rightmost_sibling(treeB,rightmostB);
 
  fatherA=(int*)malloc(sizeof(int)*treeA->size);
  fatherB=(int*)malloc(sizeof(int)*treeB->size); 
  father(treeA,fatherA);
  father(treeB,fatherB);

  is_closedA= (char**) malloc(sizeof(char*)*treeA->size);
  for (i=1; i<treeA->size; i++){
    is_closedA[i]=(char*) malloc(sizeof(char)*treeA->size);
  }
  is_closedB= (char**) malloc(sizeof(char*)*treeB->size);
  for (i=1; i<treeB->size; i++){
    is_closedB[i]=(char*) malloc(sizeof(char)*treeB->size);
  }
  is_closed_forest(is_closedA, treeA);
  is_closed_forest(is_closedB, treeB);
       
  /**********************************/
  /*    skeleton of the graph      */
  /**********************************/ 
  
   for (i=1; i<treeA->size; i++){
    for (j=1; j<treeB->size; j++){
      if (constraint[i][j]==ALLOWED){
        if ((treeA->n[i].size>1) &&  (treeB->n[j].size>1)) degree=2*(tree_closedB[j]+tree_closedA[i]);
      else if (treeA->n[i].size>1) degree=2* tree_closedB[j]; 
      else if (treeB->n[j].size>1) degree=2* tree_closedA[i];
      else degree=0; 
      init_edge_stack(&alignment_graph[i][j],degree);
      }
      else  init_edge_stack(&alignment_graph[i][j],0); /* a retirer prudemment */
    }
  }
  
  /*for (i=1; i<treeA->size; i++){
      for (j=1; j<treeB->size; j++){
	if ((treeA->n[i].size>1) || (treeB->n[j].size>1)) degree=2*(2+ tree_closedB[j]+tree_closedA[i]);
	else degree=0;
	init_edge_stack(&alignment_graph[i][j],degree);
      }
      }*/

  /*******************************/ 
  /*     weight of the edges     */
  /*******************************/
  
  for (i=treeA->size-1; i>=1; i--){
    if (treeA->n[i].size==1) {
      for (j=treeB->size-1; j>=1; j--){
	if   ((constraint[i-1][j-1]==ALLOWED) && (treeB->n[j].size==1)) {
	  
	  k=ancestorA[i]; 
	  l=ancestorB[j];
	  m=rightmostA[k];
	  n=rightmostB[l];
	  
	  /* table[x][y] : cost of an optimal path from (x,y) to (i-1,j-1).  
	   * table and alignment_graph are computed simultaneously */
	  
	  table[i-1][j-1]=0;
	  for (x=i; x<=m; x++)
	     if (constraint[x][j-1]==ALLOWED) table[x][j-1]=full_deletionA[x]+get_table(x-treeA->n[x].size,j-1); /*a revoir */
 	  for (y=j; y<=n; y++)
	    if (constraint[i-1][y]==ALLOWED) table[i-1][y]=full_deletionB[y]+get_table(i-1,y-treeB->n[y].size); /*a revoir */  
	  
	  /*** We compute the weight of all edges going to vertex (i-1,j-1) ***/

	  ed.targetX= i-1; 
	  ed.targetY= j-1;

	  /*       zone I       */
	  for (x=i; x<=k; x++){
	    if (is_closedA[i][x]){
	      for (y=j; y<=l; y++){  
		if ((is_closedB[j][y]) &&   (constraint[x][y]==ALLOWED)){
		  
		  computeTable_Cell(treeA, treeB,x,y,i-1,j-1);
		  ed.weight=get_table(x,y);  
		  
		  /* arc_mismatch between arcs x in A and y in B*/
		  if ((is_subtree(treeA,i,x)) && (is_subtree(treeB,j,y))) { 		
		    w=arc_mismatch(treeA->n[x].labelX,treeA->n[x].labelY,treeB->n[y].labelX,treeB->n[y].labelY)+get_table(x-1,y-1);
		    update(w,0,&ed); 
		  }
		  
		  /* arc-removing of x in A */      
		  if (is_subtree(treeA,i,x)) {
		    w=arc_removing(treeA->n[x].labelX, treeA->n[x].labelY)+get_table(x-1,y); 
		    update(w,1,&ed);
		  }
		  
		  /* arc-removing of y in B */
		  if (is_subtree(treeB,j,y)) {
		    w=arc_removing(treeB->n[y].labelX, treeB->n[y].labelY)+ get_table(x,y-1);
		    update(w,5,&ed);
		  }
		  
		 
		  if (get_table(x,y)<ed.weight){
		    update_edge(&alignment_graph[x][y], ed);
		    table[x][y]=ed.weight;
		  }
		}
	      }
	    }  
	  }/* end zone I */
	  
	  /*     zone II       */
	  for (x=k+1; x<=m; x++){
	    if (is_closedA[i][x]){
	      for (y=j; y<=l; y++){
		if ((is_closedB[j][y])&& (constraint[x][y]==ALLOWED)){
		  computeTable_Cell(treeA, treeB,x,y,i-1,j-1);
		  ed.weight=get_table(x,y);
		  
		  /* arc-removing of y in B*/
		  if (is_subtree(treeB,j,y)){
		    w=arc_removing(treeB->n[y].labelX, treeB->n[y].labelY)+ get_table(x,y-1);
		    update(w,5,&ed);
		  }
		  		  
		  if (ed.weight>get_table(x,y)){
		    table[x][y]=ed.weight;
		    update_edge(&alignment_graph[x][y], ed);
		  }	      
		}
	      }
	    }
	  }/* end zone II */
	  
	  /*      zone III        */
	  for (x=i; x<=k; x++) {
	    if ( is_closedA[i][x]){
	      for (y=l+1; y<=n; y++){
		if ((is_closedB[j][y]) && (constraint[x][y]==ALLOWED)){
		  computeTable_Cell(treeA, treeB,x,y,i-1,j-1);
		  ed.weight=get_table(x,y);
		  
		  /* arc-removing of arc x in A*/
		  if (is_subtree(treeA,i,x)){
		    w=arc_removing(treeA->n[x].labelX, treeA->n[x].labelY)+get_table(x-1,y);
		    update(w,1,&ed); 	
		  }
		  
		  if (ed.weight>get_table(x,y)){
		    update_edge(&alignment_graph[x][y], ed);
		    table[x][y]=ed.weight;
		  }
		}
	      }
	    }
	  }/* end zone III */
	}
      } /* end node j in treeB */
    }
  }/* end node i in treeA */
  
  free(ancestorA);
  free(ancestorB);
  free(fatherA);
  free(fatherB);
  free(rightmostA);
  free(rightmostB);
  free(tree_closedA);
  free(tree_closedB);
  /* ajouter is_closed */
}

static void build_optimal_path(postorder_tree *treeA, postorder_tree *treeB, vertex **alignment_graph, 
			       int i, int j, int k, int l, int positionA, int positionB, six_stack *p, int **supertree, int **table)
{
  int x,y, old_values; 
  int d; 
  int finished = 0;
  edge ed;   
   
  x = k; 
  y = l;
  
  
  while ( x+y > i+j ){

    finished = 0;
    /*******************/
    /* implicit edges  */
    /*******************/

    old_values = x+y; 

    if ( (x > i) && (y > j) && (treeA->n[x].size == 1) && (treeB->n[y].size == 1) && 
		(table[x][y]==base_mismatch(treeA->n[x].labelX, treeB->n[y].labelX)+get_table(x-1,y-1))){  
      supertree[treeA->n[x].positionX][treeB->n[y].positionX]=1;
      x--; y--; 
      finished = 1;
    }
    else if ( (x > i) && (table[x][y] == table[x-treeA->n[x].size][y] + full_deletionA[x]) ){
      for (d=x; d>x-treeA->n[x].size; d--){
	   if (treeA->n[x].labelX!='X') supertree[treeA->n[d].positionX][max2(positionB,treeB->n[y].positionY)]=-1;
	   if (treeA->n[x].labelX!='X') supertree[treeA->n[d].positionY][max2(positionB,treeB->n[y].positionY)]=-1;
      }
      x = x-treeA->n[x].size;
      finished = 1;
    }
    else if ( (y > j) && (table[x][y] == table[x][y-treeB->n[y].size] + full_deletionB[y]) ){
      for (d=y; d>y-treeB->n[y].size; d--){
	   if (treeB->n[y].labelX!='X') supertree[max2(positionA,treeA->n[x].positionY)][treeB->n[d].positionX]=-2;
	   if (treeB->n[y].labelX!='X') supertree[max2(positionA,treeA->n[x].positionY)][treeB->n[d].positionY]=-2;
	}
      y = y-treeB->n[y].size; 
      finished = 1;
    }
    
    /*******************/
    /* explicit edges  */
    /*******************/
    for (d=0;  (!finished) && (d <= alignment_graph[x][y].current_edge); d++){
      get_edge(&alignment_graph[x][y], &ed, d);
      if ( (ed.targetX >=i ) && (ed.targetY >= j) && (table[x][y] == ed.weight + get_table(ed.targetX, ed.targetY)) ){
	finished = 1; 
	switch (ed.edit_operation) {
	case 0 :  /* arc_(mis)match */
	   if (treeA->n[x].labelX!='X') supertree[treeA->n[x].positionX][treeB->n[y].positionX]=1; 
	   if (treeA->n[x].labelX!='X') supertree[treeA->n[x].positionY][treeB->n[y].positionY]=1; 
	  six_push(p, ed.targetX, ed.targetY, treeA->n[x].positionX, treeB->n[y].positionX, x-1, y-1); 
	  break;
	case 5 :  /* arc_insertion */
	   if (treeB->n[y].labelX!='X') supertree[max2(positionA,treeA->n[ed.targetX].positionY)][treeB->n[y].positionX]=-2;
	   if (treeB->n[y].labelX!='X') supertree[treeA->n[x].positionY][treeB->n[y].positionY]=-2;
	  six_push(p, ed.targetX, ed.targetY, max2(positionA,treeA->n[ed.targetX].positionY),  treeB->n[y].positionX, x, y-1);
	  break; 
	case 1 :  /* arc_deletion */ 
	   if (treeA->n[x].labelX!='X') supertree[treeA->n[x].positionX][max2(positionB,treeB->n[ed.targetY].positionY)]=-1;
	   if (treeA->n[x].labelX!='X') supertree[treeA->n[x].positionY][treeB->n[y].positionY]=-1;
	  six_push(p, ed.targetX, ed.targetY, treeA->n[x].positionX, max2(positionB,treeB->n[ed.targetY].positionY), x-1, y);
	  break;
	default :
	  printf("error \n");fflush(stdout);
	  break;
	} /* end of switch */
	x = ed.targetX; 
	y = ed.targetY;
      }
    }
    
    if( old_values == x+y ){
      printf("alignment is not possible \n");
      exit(0);
    }
  }
}


/* Construction of the optimal mapping: 
   the mapping is a set of pairs of nodes of treeA x treeB. 
   It is stored in a stack. */
static int trace_back_alignment_mapping(postorder_tree * treeA, postorder_tree * treeB,  vertex ** alignment_graph,
					int ** supertree, int**table){
  int score;
  int i, j, k, l,positionA, positionB;
  six_stack *  p;
  int old_i=0;
  int old_j=0;
  
  p = malloc(sizeof(six_stack));
  init_six_stack(p, (treeA->size + treeB->size) * 2);
  computeTable(treeA, treeB,  0, 0,treeA->size-1, treeB->size - 1);
  build_optimal_path(treeA, treeB, alignment_graph, 0, 0,treeA->size - 1, treeB->size - 1, 0,0,p, supertree, table);
  score = table[treeA->size - 1][treeB->size - 1];
  while(!is_empty_six_stack(p)){
    six_pop(p, &i, &j, &positionA, &positionB, &k, &l);
    if ( (i != old_i) || (j != old_j) ){
      old_i = i; old_j = j;
      computeTable(treeA, treeB, i, j, k, l);
    }
    build_optimal_path(treeA, treeB, alignment_graph, i, j, k,l,positionA, positionB, p, supertree, table);
  }
  free_six_stack(p);
  return score;
}



static int optimal_tree_alignment(RNA_sequence * rnaA, RNA_sequence * rnaB,  RNA_sequence* alignA, RNA_sequence* alignB, int full_edit, int similar, int primary, int do_mapping){
  int i,j;
  int score;
  int ** supertree; 
  postorder_tree * treeA;
  postorder_tree * treeB;


  treeA=(postorder_tree * ) malloc(sizeof(postorder_tree));
  treeB=(postorder_tree * ) malloc(sizeof(postorder_tree));
  
   if (full_edit) {
    RNAsequence2triplet_postorder(rnaA,treeA);
    RNAsequence2triplet_postorder(rnaB,treeB);
  }
  else{ 
    RNA_sequence2postorder(rnaA,treeA);
    RNA_sequence2postorder(rnaB,treeB);
  }

  
  table=(int**)malloc(sizeof(int*)*treeA->size);
  for (i=0; i<treeA->size; i++)
    table[i]=(int*) malloc(sizeof(int)*treeB->size);
    
  alignment_graph= (vertex**) malloc(sizeof(vertex*)*treeA->size);
  for (i=0; i<treeA->size; i++)
    alignment_graph[i]=(vertex*) malloc(sizeof(vertex)*treeB->size);
  
   
  full_deletionA=(int*)malloc(sizeof(int)*treeA->size);
  full_deletionB=(int*)malloc(sizeof(int)*treeB->size); 
  /* deletions in A */
  for (i=1; i<treeA->size; i++)
    full_deletionA[i]=tree_full_deletion(treeA, i);   
  /* insertions in B */
  for (i=1; i<treeB->size; i++)
    full_deletionB[i]= tree_full_deletion(treeB, i); 
  

  constraint=new_matrix_char(treeA,treeB);	
  no_constraint(constraint,treeA->size,treeB->size);	
  if (similar==1)
    diagonal_constraint(constraint,treeA->size,treeB->size, treeA->size/5);		
  if(primary==1)
    primary_sequence_constraint(constraint,treeA,treeB,rnaA,rnaB);


  build_alignment_graph(treeA, treeB);
  /* display_alignment_graph(alignment_graph, treeA->size-1, treeB->size-1); */



  if (do_mapping==1){  
    supertree = (int **)malloc( sizeof(int*)*(rnaA->size+1) );
    for (i=0; i<=rnaA->size; i++)
      supertree[i]= (int *)malloc( sizeof(int)*(rnaB->size+1) );
    for (i=0; i<=rnaA->size; i++)
      for (j=0; j<=rnaB->size; j++)
	supertree[i][j]=0;
    score= trace_back_alignment_mapping(treeA, treeB, alignment_graph, supertree,table);
    create_alignment_RNA_sequences(rnaA, rnaB, supertree, alignA,  alignB, 1, rnaA->size-1, 1, rnaB->size-1);
  } 
  else{
    computeTable(treeA, treeB,0,0,treeA->size-1, treeB->size-1);
    score=  table[treeA->size-1][treeB->size-1];
  }
  
  for (i=0; i<treeA->size; i++) free(table[i]);
  free(table);  
  
  free(full_deletionA);
  free(full_deletionB);
  free_tree(treeA);
  free_tree(treeB);

  return score;
}


/* computes the score of the optimal alignment of two trees */
/* without building the optimal mapping */
int optimal_tree_alignment_score(RNA_sequence * rnaA, RNA_sequence * rnaB, int full_edit, int similar, int primary){
  RNA_sequence * alignA=NULL; 
  RNA_sequence * alignB=NULL;
  return optimal_tree_alignment(rnaA, rnaB, alignA, alignB, full_edit,similar,primary,0);
}


int optimal_tree_alignment_mapping(RNA_sequence * rnaA, RNA_sequence * rnaB,  RNA_sequence* alignA, RNA_sequence* alignB, int full_edit, int similar, int primary){
  return optimal_tree_alignment(rnaA, rnaB, alignA, alignB, full_edit,similar, primary,1);
}


